;; Script for asynchronous Prolog process
;; private slow project since 2010-2016
;; Motivation: introduction Prolog power
;; into Emacs lisp.
;; Author: Kuniaki Mukai

(setq log t)	;; debug flag. t / nil
(setq send-receive-log  (get-buffer-create "*send-receive-log*"))
(setq log-string-length-max 100)	;; initial substring of lenght 100.
;; (setq log-string-length-max nil) ;; full string.

(defvar nullstop "\000\n")	;; nullstop
(defvar null "\000")		;; nullbyte code

;;
(defvar handle-list nil)	;; obsolete
(defvar edit-command-set nil)	;; obsolete
(defvar dir-set  nil)		;; obsolete
(defvar recent-command		;; command history
  (expand-file-name
   "~/.emacs.d/Recent-Command.txt"))
(defvar time-stamp		;; for time-stamped memo. handle([time, memo]).
  (expand-file-name
   "~/Dropbox/notes/time-stamp/"))

;; input/output log
(defun log-buffer (b obj)
     (save-excursion
       (set-buffer send-receive-log)
       (goto-char (point-max))
       (insert b)
       (insert (let ((print-escape-newlines t))
		 (if log-string-length-max
			(truncate-string (prin1-to-string obj)
					 log-string-length-max)
		   (prin1-to-string obj))))
       (insert "\n")))

;; input log.
(defun receive-log (x)
  (when log (log-buffer "receive: "  x)))

;; output log.
(defun send-log (x)
  (when log (log-buffer "send: " x)))

;; truncate string.
(defun truncate-string (s n)
  (let  ((len (length s)))
    (if (< len  n)  s
      (concat (substring s 0 n) "....\""))))


;; Start a PAC (Prolog) process.
(defun prolog-process-start (path-to-prolog-server message)
  (interactive)
  (let ((process-connection-type nil))
	(prolog-process-kill)
	(setq CONFIG  (list 'parse))
	(setq prolog-proc
        (start-process
              "PAC"
              "prolog-process-buffer"
              path-to-prolog-server))
	(set-process-filter prolog-proc 'prolog-process-filter)
	(set-process-coding-system prolog-proc 'utf-8 'utf-8)
	(message message)
	)
  )

;; kill the prolog process.
(defun prolog-process-kill ()
   (let (err)
     (condition-case err
	     (progn (kill-process (process-name prolog-proc))
		    (message "PAC process killed. "))
         (error nil))))

;; process filter
(defun prolog-process-filter (p s)
  (setq prolog-proc p)
  (receive-log s)
  (state-transition s))

;;	About `end-of-data'  marker.
;;
;; Nullstop codes ('\000\n') is reserved as
;; "end-of-data" codes from Prolog process.
;; It is still strange for the author of this codes that
;; that long texts from the process often are segmented in pieces.
;; Hence putting a special reserved marker at the end as
;; "end-of-datae" marker, there seems no other way to recover
;; the original data string from the pieces.  Thus
;; the result is undefined if the nullstop appears
;; in the middle of input data. However only nullstop which
;; appears at the end of string  passed to the filter is interpreted
;; as the end of data. Hence unless the data happens to be cut at the right
;; end of nullstop, the nullstop is read as data correctly.
;; Accordingly, the user might want to forget safely about the "nullstop"
;; as far as ordinal text editing are concerned.


;; Definition of process filter.
;;
;; Three states model;  "parse" as initial state, "text", and "script."
;; CONFIG is a list, and first element is one of the three atoms:
;; parse, text, or script.
;; The length of CONFIG always is 1 both for parse and text state.

(defun state-transition (string)
  (let (err)
    (condition-case err
	(apply  (car CONFIG) (list string))
	(error (message
		(concat "** PAC state transition error ** " err))
	       (sit-for 2) (start-prolog))
     )))

(setq text-key "$TEXT$ ")
(setq text-key-len 7)
(setq script-key "$SCRIPT$ ")
(setq script-key-len 9)

(defun parse (string)
  (let (x)
  (cond ((string-prefix-p text-key string)
		 (setq CONFIG '(text))
		 (text (substring string text-key-len)))
		((string-prefix-p script-key string)
		 (setq CONFIG '(script))
		 (script (substring string script-key-len)))
		(t
		 (setq x (insert-scan-keyword string))
		 (cond (x (parse x))
			   (t nil))))))

;; naive string scanning
;; (insert-scan-keyword "ab script cde")
(defun insert-scan-keyword (s)
	(let ((i  (string-match text-key s 0)))
	  (cond (i  (insert (substring s 0 i))
		    (substring s i))
		(t  (setq i (string-match script-key s 0))
		    (cond (i (insert (substring s 0 i))
			     (substring s i))
			  (t (insert s) nil))))))


;; text state: simply inserting the string to the buffer.
;;
;; Insert string into the buffer omitting the null codes.
;; (text "abc efg")

(defun text (string)
  (let ((totallen (length string)) bodylen)
	(setq bodylen (- totallen 2))
	(cond ((and (>= totallen 2)
		    (string= (substring string bodylen) nullstop))
		    (insert (substring string 0 bodylen))
		    (setq CONFIG '(parse)))
		(t   (insert string)
		     (setq CONFIG '(text))))))

;; script state:
;;
;; Push segmented data in CONFIG, but
;; when the final segment has come, it recovers
;; the original data  from that kept in CONFIG
;; as a series of lisp s-expressions, and then
;; eval these s-expressions in order.

(defun script (string)
  (let ((totallen (length string)) bodylen)
	(setq bodylen (- totallen 2))
	(cond ((and  (>= totallen 2)
		     (string= (substring string bodylen) nullstop))
		   (progn-string
		       (apply 'concat
			   (reverse (cons (substring string 0 bodylen)
					  (cdr CONFIG)))))
		   (setq CONFIG '(parse)))
	      (t (setq CONFIG (cons 'script
				    (cons string
					  (cdr CONFIG))))))))
;; example.  (progn-string "(insert \"abc\")")
(defun progn-string (string)
      (let ((s string) c err err0)
	   (condition-case err
		(while (not (string= s  ""))
		       (condition-case err0
			  (setq c (read-from-string s))
			  (error (setq c nil) (setq s "")))
		       (when c (setq s (substring s (cdr c)))
			       (eval-cons (car c))))
		(error  (message-simplified err )))
	  ))

(defun message-simplified  (e)
  (if (stringp e) (message e)
    (message (prin1-to-string e))))

(defun message-extra  (e)
  (if (stringp e) (message e)
    (message (prin1-to-string e))))

;;
(defun send-string (x)
  (send-log  x)
  (process-send-string prolog-proc x))

;;
(defun send-nullstop ()  (send-string "\000\n"))
(defun send-with-nullstop (x) (send-string x) (send-nullstop))
(defun send-done nil (send-with-nullstop "done"))
(defun eval-cons (e) (when (consp e) (eval e)))

;; basic modes for sending to PAC process
(defun respond (e to-string)
  (if to-string (send-with-nullstop (prin1-to-string e))
    (send-with-nullstop (if (numberp e) (prin1-to-string e) e))
    )
  )

;; for compatibility
(defun reply (e) (respond e t))

(defun report (e) (eval (cons 'progn e))
        (send-with-nullstop  "done"))

(defun prolog-query (query) (interactive)
	(send-with-nullstop   query))

(defun ask-handle () (interactive)
  (let (x)
    (setq x (read-from-minibuffer "handle: "))
    (setq edit-command-set (put-at-head x edit-command-set))
    (put-edit-command-top x)
    x))

;; PAC process may request "stop-me."
(defun stop-pac ()
  (interactive)
  (send-with-nullstop "stop")
 )

(defun ask-command () (interactive)
   (let (x)
     (setq x (completing-read "command: "  edit-command-set))
     (setq edit-command-set (put-at-head x edit-command-set))
     (put-edit-command-top x)
     x))

(defun prompt (hi  candidate)
  (let (x)
    (setq x (completing-read  (concat hi "[" (car (eval candidate)) "]: ")
			      (eval candidate)))
    (set candidate (put-at-head x (eval candidate)))
    x))

(defun put-at-head (x y) (cons x (remove-elem x y nil)))

(defun remove-elem (x y z)
  (if (eq y nil) (reverse z)
    (if (equal  x (car y)) (append (reverse z) (cdr y))
      (remove-elem x (cdr y) (cons (car y) z)))))

(defun put-at-top (command)
  "Delete the line that exactly containes command in buffer,
and  insert command at beggining of buffer"
  (goto-char (point-min))
  (when (re-search-forward (concat "^" command "$") nil t)
	(beginning-of-line)
	(kill-line)
	(delete-blank-lines))
  (goto-char (point-min))
  (insert command)
  (insert "\n")
  )

(defun get-lines ()
  (let ((x nil) y z)
    (while (not (eobp))
      (setq y (point))
      (end-of-line)
      (setq z (buffer-substring-no-properties y (point)))
      (unless (string= z "") (setq x (cons z x)))
      (forward-line))
    (reverse x)))

(defun put-lines (x)
  (while x
    (insert (car x))
    (insert "\n")
    (setq x (cdr x))))

(defun initialize-edit-command-set ()
  (save-excursion
    (set-buffer (find-file-noselect
		 (expand-file-name "~/.emacs.d/Recent-Command.txt")))
    (goto-char  (point-min))
    (setq edit-command-set  (get-lines))
    ))

(defun put-edit-command-top (command)
  (save-excursion
    (set-buffer (find-file-noselect
		 (expand-file-name "~/.emacs.d/Recent-Command.txt")))
    (goto-char  (point-min))
    (put-at-top command)
    (basic-save-buffer)
    ))

(defun ask-term (prompt) (interactive)
       (completing-read  (concat prompt  ": ") nil ))

(defun get-region ()
  (buffer-substring-no-properties
   (if (mark t) (mark) (point)) (point)))

(defun send-region (&optional x y)
	(unless x  (setq x (if (mark t) (mark) (point))))
	(unless y  (setq y (point)))
	(set-mark x)
	(goto-char y)
	(process-send-region prolog-proc x y)
	(send-nullstop)
)

(defun region-bound ()
    (list (if (integerp (mark t)) (mark) (point))  (point)))

;; [2014/03/17]
(defun replace-in-line (l m x)
  "Replace substring in the current line "
	(let ((p  (+ (mark)  l)))
		(kill-region p (+ p m))
		(goto-char p)
		(insert x)))

;; [2013/09/18]
;; mark the current paragraph region.
(defun mark-paragraph-region ()
	(let ((x (point)) p)
		(setq p (search-backward "\n\n" nil t))
		(if p  (set-mark (+ p 2)) (set-mark (point-min)))
		(goto-char x)
		(setq p (search-forward "\n\n" nil t))
		(if p (goto-char  (- p 2)) (goto-char (point-max))))
)

;; global set key for PAC process
(defun global-set-key-kbd (P)
  	 (global-set-key (kbd (car P))
		  (list 'lambda nil '(interactive)
			(list 'prolog-query (cdr  P)))))

(mapcar 'global-set-key-kbd'(
  ( "C-<return>"   . "ejockey:handle([prolog])")
  ( "C-c p"   . "ejockey:handle([prolog, region])")
  ( "C-c P"   . "ejockey:handle([prolog, paragraph])")
  ( "C-c u"   . "ejockey:handle([html,tag,code])")
  ( "s-<return>"   . "ejockey:handle([parse, bind, context, append])")
  ( "s-M-<return>" . "ejockey:handle([parse, bind, context])")
  ( "S-<return>"   . "ejockey:handle([shell, dakuten])")
  ( "M-<return>"   .  "ejockey:handle([eval, markup, text, append])")
  ( "S-M-<return>" .  "ejockey:handle([eval, markup, text])")
  ( "S-s-RET" . "ejockey:handle([note, \"~/Dropbox/notes/snippets.txt\", []])")
  ( "M-s-√"        . "ejockey:handle([clipboard])")
  ( "s-<down>"     . "ejockey:handle([edit, pred])")
  ( "s-<up>"	   . "ejockey:handle([edit, pred, *])")
  ( "s-<right>"	   . "ejockey:handle([compile, pac, generic])")
  ( "s-<left>"	   . "ejockey:handle([compile, pac, region])")
  ( "C-l"	   . "elisp:edit_emacs_L(elisp:ask_handle)")
  ( "s-L"	   . "ejockey:handle([set,context,module])")
  ( "C-x C-;"        . "ejockey:handle([comment, region])")
  ( "C-x C-:"        . "ejockey:handle([uncomment, region])")
  ))

;; (defun my-resize-margins ()
;;   (let ((margin-size (/ (- (frame-width) 80) 2)))
;;     (set-window-margins nil margin-size margin-size)))

;; (add-hook 'window-configuration-change-hook #'my-resize-margins)
;; (my-resize-margins)

(defun my-horizontal-recenter ()
  "make the point horizontally centered in the window"
  (interactive)
  (let ((mid (/ (window-width) 2))
        (line-len (save-excursion (end-of-line) (current-column)))
        (cur (current-column)))
    (if (< mid cur)
        (set-window-hscroll (selected-window)
                            (- cur mid)))))


;;		/***************************************
;;		*     Emacs - Prolog  communication    *
;;		***************************************/

(setq   START-PAC	"~/lib/swipl/pack/pac/prolog/pac/start-pac.pl"
	SWIPL-PROLOG	"/usr/local/bin/swipl"
	QSAVED-PAC		"~/lib/swipl/pack/pac/etc/swipl"
	DUMP-PAC		"~/lib/swipl/pack/pac/shell/dump-pac"
	QSAVED-EH		"~/lib/swipl/pack/pac/etc/emacshandler"
	DUMP-EH			"~/lib/swipl/pack/pac/shell/dump-eh"
	PAC-RESTORE     "~/lib/swipl/pack/pac/etc/pac-restore.pl"
	)

(defun start-qsaved-emacshandler () (interactive)
  "open communation via qsaved emacs-handler"
  (prolog-process-start
	(expand-file-name QSAVED-EH)
	(concat DUMP-EH " starts..." )
	)
   (send-with-nullstop "set_dired, ejockey:handle([handle,list])")
 )

(defun restart-emacshandler ()
  (prolog-process-start
   (expand-file-name QSAVED-EH)
	"*** Exception ***  Restart PAC...")
  )

(global-set-key  (kbd "C-c f p")
	    (lambda nil (interactive)
		(find-file-existing "~/lib/swipl/pack/pac/elisp/prolog-process.el")))

(global-set-key  (kbd "C-c f P")
	    (lambda nil (interactive)
		(find-file-existing "~/lib/swipl/pack/pac/")))

(global-set-key (kbd "s-p")  'start-qsaved-emacshandler)

(global-set-key (kbd "M-s-π")
				(lambda () (interactive)
				  (shell-command
				    (concat (expand-file-name DUMP-EH)
							" >> /dev/null 2>&1"))
					(message "pac/etc/dump-eh done")
				(start-qsaved-emacshandler)))

(global-set-key (kbd "C-c q")(lambda () (interactive)
	  "consult qsaved ediprolog mode for PAC top"
	  (when ediprolog-process (ediprolog-dwim 0)) ;; kill the current
	  (setq ediprolog-program
		(expand-file-name QSAVED-PAC))
	  (find-file (expand-file-name PAC-RESTORE))
 	  (command-execute 'ediprolog-consult)
	  (switch-to-buffer (other-buffer (current-buffer) t))))

(global-set-key (kbd "C-c Q")(lambda () (interactive)
          "** Dump pac, and consult qsaved ediprolog mode for PAC top. **"
	  (shell-command
	   (concat (expand-file-name DUMP-PAC)
				" >> /dev/null 2>&1"))
	  (message "pac/etc/dump-pac done")
	  (when ediprolog-process (ediprolog-dwim 0)) ;; kill the current
	  (setq ediprolog-program
		(expand-file-name QSAVED-PAC))
	  (find-file (expand-file-name PAC-RESTORE))
 	  (command-execute 'ediprolog-consult)
	  (message (concat QSAVED-PAC " starts..." ))
	  (switch-to-buffer (other-buffer (current-buffer) t))))

(global-set-key (kbd "C-c t")(lambda () (interactive)
  	 "** consult Plain SWI-Prolog **"
	  (when ediprolog-process (ediprolog-dwim 0)) ;; kill the current
	  (setq ediprolog-program
		(expand-file-name SWIPL-PROLOG))
	  (find-file (expand-file-name  START-PAC))
	  (command-execute 'ediprolog-consult)
	  (switch-to-buffer (other-buffer (current-buffer) t))))

(initialize-edit-command-set)
