:- module(expand_dict, [
				expand_dict/3,
%				btree_build/2,
				btree_to_dict/2,
				expand_dict_head/3,
				expand_dict_basic/2,
				expand_dict_goal/2,
				expand_dict_clause/2,
				expand_dict_dcg_rule/2
			   ]).

:- use_module('odict-attr').
:- use_module(reduce).
:- use_module([
		library(lists),
		library(sort),
		library(ordsets)]).

% for short.
put_attr(V, A):- put_attr(V, cil, A).
get_attr(V, A):- get_attr(V, cil, A).

% The term expansion is designed so that the expanded dict terms
% have  no calls of unify/2. It may have calls of role/3, though.
% btree unification is hooked in cil module.

%	Syntax of iterm and dterm.
%
% term		:=	<prolog term>
% atom		:=  <prolog atom>
% variable  :=  <prolog variable>
% key		:=  <prolog ground term>
%
% iterm 	:=	{}		(void iterm)
%			|	{ key:dterm, key:dterm, ..., key:dterm }
%
% dterm		:=	term
%			|	atom(dterm, ..., dterm)
%			|	iterm
%			|	rterm
%
% rterm		:=  variable.kterm.kterm....kterm
%
% kterm		:=  term.

% Examples of iterm
%   {a:1,  b:{c:2, d:3} }
%   {a:X,  b:f(X, X)}
%   {a:X,  b:X.c.d}
%   {a:f(X.a, Y.b)}
%   {a:f(X.A.B, Y.B)}

	/**************************************
	*        term expansion for Dict      *
	**************************************/

non_clause(:-(_)).
non_clause(?-(_)).

is_clause(X):- \+ non_clause(X).

%
expand_dict(X, Y):- is_clause(X),
					expand_dict_clause(X, Y).

% ?- [util('ptq-fragment')].
% ?- module(ptqfrag).
% ?- run_samples.

% ?- module('expand_dict').
% ?- btree_build({A:Y}, Z).
% ?- btree_build({a:Y, a:Y}, Z).
% ?- enable_odict.
% ?- enable_pac_query.
% ?- X = {a:1, B}.		% not dict.
% ?- X = {a:1, a:2}.	% dict.
% ?- X={c:1, d: Y.d}, Y={c:X.c, d:2}, C= Y.c, D=X.d.
% ?- C= Y.c, D=X.d, X={c:1, d: Y.d}, Y={c:X.c, d:2}.
% ?- expand_dict_goal( (#=(role(X,a), 1), V = role(X,a)), R).

% ?- disable_odict.
% ?- disable_pac_query.
% ?- trace, btree_build({k:1}, S0), btree_build({j:2}, R0), S=x(S0,_), R=x(R0,_),  unify(a(S,R), a(M, M)), role(k, M, V), role(j, M, U).

btree_build(X, X):- (var(X); atomic(X)), !.
btree_build({X}, Y):- !,
	( var(X) -> throw(btree_build('variable dict found.'))
	;  	dict_to_balanced_btree(X, Y)
	).
btree_build(X, Y):- X=..[F|As],
		maplist(btree_build, As, Bs),
		Y=..[F|Bs].

%  './2' and './3' are reserved for the dict in SWI-7.
%  './2' is untuchable for the user in a direct way.
period_term(X):- functor(X, (.), 2), !.
period_term(role(_,_)).

%
period_args(X, A, B):- X=..[(.), A, B].
period_args(role(A, B), A, B).

% % anti_subst(+X, -A, -R) is det.
%  True if apply the assoc to A makes X.
%  Roughly it is an inverse operation of the substitution.
%
% ?- anti_subst(a({b:1}), B, R).
% ?- anti_subst(a({b:{c:1}}), B, R).
% ?- anti_subst(a({b:{C:1}}), B, R).
% ?- anti_subst(a({b:2, b:1}), B, R).

%
anti_subst(X, A, R):- anti_subst(X, A, R, []),
 		  maplist(check_item, R).
%
anti_subst(X, A, R, R):- (var(X); atomic(X)), !, A=X.
anti_subst({X}, A, R, Q):- !,
		( var(X)	-> throw(btree_build('unexpected variable found'))
		;	anti_subst(X, X0, R, R0),
			R0= [A={X0}|Q]
		).
anti_subst(X, A, [A=X|R], R):- period_term(X), !. % X= .(_,_)
anti_subst(X, A, R, R0):- X=..[F|Xs],
			anti_subst_list(Xs, As, R, R0),
			A=..[F|As].
%
anti_subst_list([],[],R,R).
anti_subst_list([X|Xs],[Y|Ys],R, R0):-
			anti_subst(X, Y, R, R1),
			anti_subst_list(Xs, Ys, R1, R0).

%
check_item(_={X}):- !,
	 ( once( (check_ground_key(X),
			  check_duplicate_key(X, [], _))) -> true
	 ; throw(btree_build(non_ground_or_duplicate_key({X})))
	 ).
check_item(_=_).   % X.a.b etc
%
check_ground_key(X):- var(X), !,
					  throw(btree_build('unexpected variable found')).
check_ground_key(X:_):- ground(X).
check_ground_key((X,Y)):- check_ground_key(X),
		check_ground_key(Y).
%
check_duplicate_key(X:_, Ks, [X|Ks]):- \+ memberchk(X, Ks).
check_duplicate_key((X,Y), L, M):-  check_duplicate_key(X, L, N),
		check_duplicate_key(Y, N, M).

% ?- dict_to_balanced_btree((b:2, c:3, a:1), B).
dict_to_balanced_btree(X, Y):- dict_to_list(X, X0, []),
							   list_to_btree(X0, Y).

%
dict_to_list(X, _, _):- var(X), !,
 		throw(btree_build('unexpected variable found')).
dict_to_list((X,Y), P, Q):- !, dict_to_list(X, P, P0),
							dict_to_list(Y, P0, Q).
dict_to_list(X, [X|P], P).

% ?- module(expand_dict).
% ?- list_to_btree([a:1, b:2, c:3, d:4, e:5], T).

list_to_btree(X, Y):- sort(X, X0),
	length(X0, N),
	list_to_btree(X0, N, Y).

%
list_to_btree([], _, _).  % for open dict
list_to_btree(X, N, t(K, U, L0, R0)):-
		  J is N//2,
		  length(L, J),
		  append(L, [Pair|R], X),
		  pair(Pair, K, U),
		  list_to_btree(L, J, L0),
		  J0 is N - J - 1,
		  list_to_btree(R, J0, R0).
%
pair(A-B, A, B).
pair(A=B, A, B).
pair(A:B, A, B).

	/*************************************
	*        convert btree to list/dict  *
	*************************************/

% ?- module(expand_dict).
% ?- btree_to_dict({}, X).
% ?- btree_to_dict({a:1, b:2}, X).
% ?- trace, btree_to_dict({a:f(1)}, X).
% ?- btree_to_dict({a:f({a:1})}, X).
% ?- btree_to_dict({a:f({a:A})}, X).
%
is_btree(t(_,_,_,_)).
is_btree({}).

%
list_to_comma([], {}):-!.
list_to_comma(X, {Y}):- list_to_comma_(X, Y).
%
list_to_comma_([X], X):-!.
list_to_comma_([X, Y|Z], (X, U)):- list_to_comma_([Y|Z], U).

% ?- module(expand_dict).

%% btree_to_dict(+X, -Y) is det.
%  Y is a dict form of an internal binary tree X.

btree_to_dict(X, Y):- map_btree(X, Y0), list_to_comma(Y0, Y).

%% skelton(+X, -Y) is det.
%   True if Y is a btree X in the form of the list with
%	all leaves of X being removed.
%
%  ?- skelton(t(a, b, t(c, t(d, E,_,_), _, t(f,g, _, _)), _), X).
%  ?- skelton(t(a, t(c, t(d, E,_,_), _,_), _, _), X).
%  ?- skelton(t(a, t(c, 1, _, _), _, _), X).
%  ?- skelton(t(a, b, t(c, d, _,_), _), X).

skelton(X, Y):- map_btree_to_list(skelton_pair, X, Y, []).

% Form options for pairing key-value, supposed to be
% passed to map_btree_to_list.

default_pair(terminal, K, V, K-leaf(V)).
default_pair(nonterminal, K, V, K-V).
%
ambiguous_pair(_, K, V, K-V).
%
skelton_pair(terminal, K, _, K).
skelton_pair(nonterminal, K, V, K-V).

% Amiguous mapping  a btree to its list form.
map_btree(X, Y):- map_btree_to_list(ambiguous_pair, X, Y, []).

%% map_btree_to_list(+M, +X, -Y, +Z) is det.
%  Y is a list form of an internal binary tree X,
%  with a key-value formed by option M.

map_btree_to_list(M, X, Y, Z):- var(X), !,
				 ( get_attr(X, X0)
				 -> (X0 = btree(Btree)
					->	map_btree_to_list(M, Btree, Y, Z)
					;	Y = Z
						)
				 ;	Y = Z
				 ).
map_btree_to_list(_, {}, Y, Y):- !.
map_btree_to_list(M, t(K, V, L, R), P, Q):-
				map_btree_to_list(M, L, P, P0),
				map_btree_arg(M, K, V, Pair),
				P0=[Pair|P1],
				map_btree_to_list(M, R, P1, Q).

%% map_btree_arg(+M, +K, +V, -Pair) is det.
%  Pair is unified with a key-value formed by option M,
%	where V is converted to be the value.

map_btree_arg(M, K, V, Pair):- attvar(V), !,
				get_attr(V, V0),
				(	is_btree(V0)
				->	map_btree_to_list(M, V0, P, []),
					call(M, nonterminal, K, P, Pair)
				;	call(M, terminal, K, V, Pair)
				).
map_btree_arg(M, K, V, Pair):- var(V), !,
			call(M, terminal, K, V, Pair).
map_btree_arg(M, K, V, Pair):- is_btree(V), !,
			map_btree_to_list(M, V, U, []),
			call(M, nonterminal, K, U, Pair).
map_btree_arg(M, K, V, Pair):-
			(	atomic(V) -> call(M, terminal, K, V, Pair)
			;	is_list(V) -> maplist(map_btree_arg(M), V, U, []),
		 				 call(M, terminal, K, U, Pair)
			;	V =..[F|Vs],
				maplist(map_btree_list(M), Vs, Us, []),
				U =..[F|Us],
				call(M, terminal, K, U, Pair)
			).

% ?- expand_dict:expand_dict({a:1}, Y, G).

expand_dict(X, Y, G):-
		btree_build(X, X1),
		region_constr_of_leaves(X1, total, Put_attrs, []),
		G = (	put_attr(X0, cil, btree(X1)),
				maplist(call, Put_attrs),
				cil:(Y=X0)
			).

	/*************************************************
	*      Expand Feature Structure Unification      *
	*************************************************/

% ?- make_body_unify([a = A.b], X).
% ?- make_body_unify([a = A.b.c], X).

% Note:   cil:(X=Y)  is for sending X=Y to attr_unify_hook/2
% defined in the cil module.

make_head_unify([], true).
make_head_unify([E], X):- !,
		make_head_unify_one(E, X).
make_head_unify([E|R], (X, R0)):-
		make_head_unify_one(E, X),
		make_head_unify(R, R0).
%
make_head_unify_one(A = B, cil:(A=B)):- (var(B); atomic(B)), !.
make_head_unify_one(A = {B}, (put_attr(B0, cil, btree(B1)),
							  maplist(call, Put_attrs),
							  cil:(A=B0)) ):- !,
		btree_build({B}, B1),
		region_constr_of_leaves(B1, total, Put_attrs, []).
make_head_unify_one(A = P, G):- period_term(P), !,
		flatten_period(P, L, []),
		L=[X|L0],
		expand_dict_role(L0, X, A, G).
make_head_unify_one(A = P, cil:(A=P)).

%
make_body_unify([], true).
make_body_unify([E], X):- !,
		make_body_unify_one(E, X).
make_body_unify([E|R], (X, R0)):-
		make_body_unify_one(E, X),
		make_body_unify(R, R0).
%
make_body_unify_one(A = B, true):- (var(B); atomic(B)), !, A = B.
make_body_unify_one(A = {B}, (	put_attr(B0, btree(B1)),
								maplist(call, Put_attrs),
								cil:(A=B0)
							 )):- !,
		btree_build({B}, B1),
		region_constr_of_leaves(B1, total, Put_attrs, []).
make_body_unify_one(A = P, G):- period_term(P), !,
		flatten_period(P, L, []),
		L = [X|L0],
		expand_dict_role(L0, X, A, G).
make_body_unify_one(A = P, true):- A = P.

%
expand_dict_role([], X, A, cil:(A=X)):-!.
expand_dict_role([R|P], X, A, (role(R, X, Y), G) ):-
		expand_dict_role(P, Y, A, G).
%
flatten_period(P, [P|L], L):- var(P), !.
flatten_period(P, Q, R):- period_args(P, A, B), !,
		flatten_period(A, Q, Q0),
		flatten_period(B, Q0, R).
flatten_period(P, [P|Q], Q).

% ?- expand_dict_goal((true, true), X).
expand_dict_goal(X, Y):-
		once(expand_dict_to_front(X, Y0)),
		once(slim_goal(Y0, Y)).
%
expand_dict_to_front((X,Y), (X0, Y0)):-
	expand_dict_to_front(X, X0),
	expand_dict_to_front(Y, Y0).
expand_dict_to_front(X;Y, X0; Y0):-
	expand_dict_to_front(X, X0),
	expand_dict_to_front(Y, Y0).
expand_dict_to_front(X->Y, X0->Y0):-
	expand_dict_to_front(X, X0),
	expand_dict_to_front(Y, Y0).
expand_dict_to_front(not(X), \+(X0)):-
	expand_dict_to_front(X, X0).
expand_dict_to_front(X, Y):- X=..[phrase, P|R], !,
	expand_dict_dcg_rule_body(P, Q),
	Y =..[phrase, Q|R].
expand_dict_to_front(L=R, G):-
	anti_subst(L=R, L0=R0, U0),
	make_body_unify(U0, U),
	( U==[]	->  G = (cil:(L0=R0))
	;  G = (U, cil:(L0=R0))
    ).
expand_dict_to_front(X, Y):- expand_dict_basic(X, Y).


 	/****************************************
	*   Expand CIL clauses, DCG rules,		*
	*	and queries.						*
	****************************************/

% ?- expand_dict_clause(m: ((a:-b), c:-d), R).
%@ R = m:((a:-b), c:-d).

expand_dict_clause(:-(H, B), :-(NewH, NewB)):-!,
		expand_dict_head(H, NewH, U),
	    expand_dict_goal(B, NewB0),
		slim_goal((U, NewB0), NewB).
expand_dict_clause(M:A, M:B):-!,
		expand_dict_clause(A, B).
expand_dict_clause(X-->X0, Y):-
		expand_dict_dcg_rule(X-->X0, Y),
		!.
expand_dict_clause(X, H:-Eqs):-
         expand_dict_head(X, H, Eqs).

% ?- expand_dict:expand_dict_head(a:-b, R, Eqs).
expand_dict_head(H, NewH, Eqs):-
        anti_subst(H, NewH, U),
		make_head_unify(U, Eqs).

% ?- expand_dict:expand_dict_basic(p({a:1}), R).
expand_dict_basic(X, (U, X0)):-
	anti_subst(X, X0, U0),
	make_body_unify(U0, U).

% ?- module(expand_dict).
% ?- expand_dict_dcg_rule(a({j:1})-->b, R).
% ?- expand_dict_dcg_rule(a({j:1})-->b({k:2}), R).
% ?- expand_dict_dcg_rule(a({j:{i:1}})-->b({k:{l:2}}), R).
% ?- expand_dict_dcg_rule(a-->b({k:{l:2}}), R).

expand_dict_dcg_rule(H --> B, (NewH--> {SlimU}, NewB)):-
        anti_subst(H, NewH, Eqs),
		make_head_unify(Eqs, U),
		slim_goal(U, SlimU),
	    expand_dict_dcg_rule_body(B, B0),
		slim_goal(B0, NewB).
%
expand_dict_dcg_rule_body(A, A):-
		(var(A); atomic(A); is_list(A); string(A)), !.
expand_dict_dcg_rule_body((A, B), (A0, B0)):- !,
		expand_dict_dcg_rule_body(A, A0),
		expand_dict_dcg_rule_body(B, B0).
expand_dict_dcg_rule_body(X;Y, X0; Y0):-
		expand_dict_dcg_rule_body(X, X0),
		expand_dict_dcg_rule_body(Y, Y0).
expand_dict_dcg_rule_body(X|Y, X0|Y0):-
		expand_dict_dcg_rule_body(X, X0),
		expand_dict_dcg_rule_body(Y, Y0).
expand_dict_dcg_rule_body({A}, {B}):- !,
		expand_dict_goal(A, B).
expand_dict_dcg_rule_body(A, ({U}, A0)):-
		anti_subst(A, A0, U0),
		make_body_unify(U0, U).
