:- module(pac_regex, [parse_regex/2,
		      regex_unix_macro/2]).

% [2015/02/17]

regex_unix_macro($, @( $ )).
regex_unix_macro(^, []).

% ?- parse_regex:parse_regex("a", X).
% ?- parse_regex:parse_regex("a|b", X).
% ?- parse_regex:parse_regex("(abc)*", X).
% ?- parse_regex:parse_regex("([% \t]|(%@*)", X).
% ?- parse_regex:parse_regex("([% \t]|(%@*)|(\\?-))*", X).
% ?- parse_regex:parse_regex("[^\n]*$", X).

			/*********************
			*     parse regex    *
			*********************/
% %
% dcg_d((.), [A|B], B, [A]).
% dcg_d([A], [A|B], B, [A]).
% dcg_d({G},  A,    A, {G}).
% dcg_d(X, A, B, X0):-  pacx:complete_args(X, [A, B], X0).

% ?- pac_regex:parse_regex("ab", R).
% ?- parse_regex("a\\*",R).
% ?- parse_regex("a\\\*",R).  % <== intentional Syntex error
% ?- parse_regex("a\\\\*",R).
% ?- parse_regex("a\\\\\\*",R).
% ?- parse_regex("[abc\\]]*",R).
% ?- parse_regex("[abc\\\\]]*",R).
% ?- parse_regex("a", R).
% ?- parse_regex("(a)", R).
% ?- parse_regex("(abc)", R).
% ?- parse_regex("[abc]", R).
% ?- parse_regex("[abc]*",R).
% ?- parse_regex("[^abc]", R).
% ?- parse_regex("(.*)", R).
% ?- parse_regex("(a*)", R).
% ?- parse_regex("a*b", R).
% ?- parse_regex(".", R).
% ?- parse_regex("[^a-zA-Z]",R).
% ?- parse_regex("a|b|c", R).
% ?- parse_regex("(a|b|c)**", R).
% ?- parse_regex("(\\(*[a]|1)",R).
% ?- parse_regex("abc",R).
% ?- parse_regex("[a]",R).
% ?- parse_regex("[ab]",R).
% ?- parse_regex("[a-b]",R).
% ?- parse_regex(".",R).
% ?- parse_regex("\\.",R).
% ?- parse_regex("a",R).
% ?- parse_regex("[^\n]",R).
% ?- parse_regex("$",R).

%%	parse_regex(+X:<nonempty string>, -Y) is det.
parse_regex(X, Y) :- string_chars(X, X0),
	once(parse_regex([], [], [], Y0, X0, [])),
	Y0 = [Y].
%
parse_regex(A, B, A0, B0) --> token(T), !,
	{ once(push_pop_stacks(T, A, B, A1, B1)) },
	parse_regex(A1, B1, A0, B0).
parse_regex(A, B, A0, B0) --> % end of the regex
	{ close_block(A, B, A0, B1),
	  fold_block_reversely(B1, B0) }.

% standard push/pop actions from  operator-precedence grammars
push_pop_stacks(')', A, X, B, Y) :-  close_block(A, X, B, Z),	% block close
	fold_block_reversely(Z,  Y).
push_pop_stacks('(', A, X, ['('|A], ['('|X]).	% block open
push_pop_stacks({}(I), A, [T|X],  A, [T^I|X]).	% repeat spec
push_pop_stacks(F, A, X, A, Y) :-  unary_opr(F),
	apply_unary_opr(F, X, Y).
push_pop_stacks(F, A, X, [F|B], Y) :- binary_opr(F),
	once(sweep_higher_opr(F, A, X, B, Y)).
push_pop_stacks(T, A, X, A, [T|X]).

%
token('C'([C])) --> [(\)], [C0], {char_code(C0, C)}.  % escape charcter
token(X) --> [ A ], { regex_unix_macro(A, X) }.
token(T) --> [T], {memberchk(T, ['(',')', '*', '+', '!', '|', '.', '?'])}.
token(out(D))	--> ['[', ^],  char_class(D0, []), {chars_interval(D0, D)}.
token(dot(D))	--> ['['], char_class(D0, []), {chars_interval(D0, D)}.
token({}(I))	--> ['{'], repeat_spec(I).
token('C'([C]))	--> [C0], {char_code(C0, C)}.

%
sweep_higher_opr(_, [], X, [], X).
sweep_higher_opr(_, ['('|X], Y, ['('|X], Y).
sweep_higher_opr(F, [G|A], X, B, Y):- higher_priority(G, F),
	apply_binary_opr(G, X, Z),
	sweep_higher_opr(F, A, Z, B, Y).
sweep_higher_opr(F, A, X, B, Y):- push_pop_stacks(F, A, X, B, Y).

%
close_block([], X, [], X).
close_block(['('|A], X, A, X).
close_block([F|A], X, B, Y):-
	apply_binary_opr(F, X, Z),
	close_block(A, Z, B, Y).

% ?- parse_regex:fold_block_reversely([a, c, '(', b], X).

fold_block_reversely([X|Y], [Z|U]):- fold_block_reversely(X, Y, Z, U).
fold_block_reversely([], []).

%
fold_block_reversely(X, [], X, []).
fold_block_reversely(X, ['('|Y], X, Y).
fold_block_reversely(X, [Y|Z], U, V):- fold_block_reversely(Y+X, Z, U, V).

%
unary_opr(*).
unary_opr(?).
unary_opr(!).
unary_opr(+).

%
binary_opr(&).
binary_opr('|').

% only for binary operators
higher_priority(&, &).
higher_priority(&, '|').
higher_priority('|', '|').

%
apply_unary_opr(*, [X|Y], [*(X)|Y]).
apply_unary_opr(+, [X|Y], [+(X)|Y]).
apply_unary_opr(?, [X|Y], [?(X)|Y]).
apply_unary_opr(!, [X|Y], [!(X)|Y]).

%
apply_binary_opr('|', [X, Y|Z], ['|'(Y, X)|Z]).
apply_binary_opr('&', [X, Y|Z], ['&'(Y, X)|Z]).

% ?- parse_regex:repeat_spec(X, ['1', '2', ',', '3', '4', '}'], Y).
%@ X = 12-34,
% ?- parse_regex:repeat_spec(X, ['1', '2', ',','}'], Y).
%@ X = >=(12),
repeat_spec(X) --> number_chars(J0), [','],   { J0\==[] },
	 number_chars(K0), ['}'],
	{ 	number_chars(J, J0),
		( K0 \== []
		->	number_chars(K, K0),
			X = J-K
		;	X = (>=(J))
		)
	}.

%
number_chars([D|Ds]) --> [D], {char_type(D, digit)},
	number_chars(Ds).
number_chars([]) --> [].

%
char_class([\(C)|Y], Z)	--> [\], [C], char_class(Y,Z).
char_class([A|X], Y)	--> [A, A], char_class(X, Y).  % "idempotent" law
char_class(X, X)		--> [']'].
char_class([C|Y], Z)	--> [C], char_class(Y,Z).

%
chars_interval([],[]).
chars_interval([X, -, Y|R], [C - D|S]):- drop_escape(X, X0),
	drop_escape(Y, Y0),
	char_code(X0, C),
	char_code(Y0, D),
	chars_interval(R, S).
chars_interval([X|R], [C|S]):- drop_escape(X, X0),
	char_code(X0, C),
	chars_interval(R, S).

%
drop_escape(\(X), X).
drop_escape(X, X).
