:- module(reduce, [attach_prefix/3, complete_args/3,
		   optimize_call_N/2,
		   slim_goal/2,
		   subtree/4, subtree/5,
		   reduce/3, reduce/4, reduce_one/4,
		   term_rewrite/3, term_rewrite/4]).

% reduce/[3,4]  by rules
% ?- reduce:reduce(math:desugar, -(a-b), X).
%@ X = -1* (a+ -1*b).
% ?- reduce:term_write(math:desugar, [],  -(+(a)-b), X).
%@ X = -1* (a+ -1*b).
% ?- reduce:term_write(math:desugar, (-);(+),  -(+(a)-b), X).
% ?- reduce:term_write(math:desugar, {-,+},  -(a), X).

		/******************************************************************
		*     [2016/03/20]                                                *
		*     Regular expressions over signatures (S-Regex).              *
		*                                                                 *
		*     <S-Regex> -->                                               *
		*	  []                                                          *
		*     |	<atom>                                                    *
		*     |	<S-Regex>; <S-Regex>                                      *
		*     |	<S-Regex>, <S-Regex>                                      *
		*     |	\+ <S-Regex>                                              *
		*     |	<S-Regex>/N                                               *
		*     |	[N, N, ..., N]		(non-empty list)                      *
		*                                                                 *
		*     where <atom> is a prollog atom and N a positive integer.    *
		******************************************************************/

:- meta_predicate reduce_one(2, 4, ?, ?).
reduce_one(R, P, X, Y):- call(P, X, Y, S, S0),
	nonvar(S),				% [2010/03/05]
	call(R, S, S0).

:- meta_predicate reduce(2, ?, ?),
				reduce(2, 4, ?, ?).
:- meta_predicate term_rewrite(2, ?, ?),
				term_rewrite(2, ?, ?, ?).

reduce(R, A, B):-  reduce(R, subtree([]), A, B).
%
reduce(R, P, X, Y):- reduce_one(R, P, X, X0), !,
					 reduce(R, P, X0, Y).
reduce(_, _, X, X).

%
term_rewrite(R, X, Y):- term_rewrite(R, [], X,Y).
%
term_rewrite(R, E, X, Y):- reduce_one(R, subtree(E), X, X0), !,
						   term_rewrite(R, E, X0, Y).
term_rewrite(_, _, X, X).

%%% The subtree relation is the transtive closure of child relation.
% subtree/[2,3,4,5]
% ?- reduce:subtree(a+b, X).
% ?- reduce:subtree(+, a*b+ c*d, X).
% ?- reduce:subtree(\+ (+), a*b+ c*d, X).
%@ X = a*b+c*d ;
% ?- reduce:subtree(\+ (-), a*b+ c*d, X).
% ?- reduce:subtree((+); (*)/2, a*b + c*d, X).
% ?- reduce:subtree(\+ ((+); []/2), a*b + c*d, X).
% ?- reduce:subtree({+, *}, a*b + c*d, X).
% ?- reduce:subtree(f(a,b), T, S, S0).
% ?- reduce:subtree(f, f(a, f(b,c)), X, Y, Z).
% ?- reduce:subtree(([], \+g),  f(a,g(b)), X, Y, Z).
% ?- reduce:subtree(\+g,  f(a,g(b)), X, Y, Z).

subtree(T, S):- child([], T, S).

subtree(_, T, T).
subtree(Sgn, T, S):- child(Sgn, T, U), subtree(Sgn, U, S).

child(S, X, Y):- compound(X),
	functor(X, F, N),
	between(1, N, I),
	sgn_index_chk(S, F, N, I),
	arg(I, X, Y).

% ?- reduce:subtree((f,[1,2]) ; (g, [1]), f(b,g(a)), X, Y, Z).
subtree(T, T0, S, S0):- subtree([], T, T0, S, S0).

subtree(_, T, T0, T, T0).
subtree(Sgn, T, T0, S, S0):- child(Sgn, T, T0, U, U0),
	subtree(Sgn, U, U0, S, S0).

% Get a child (backtrackable)
% ?- reduce:child([], a+b, X0, Y, Y0).
% ?- reduce:child(*, a+b, X0, Y, Y0).

% ?- reduce:child((+)/1;(+)/2, a+b, X0, Y, Y0).
% ?- reduce:child((+)/1;(+)/2, a+b, X0, Y, Y0).
% ?- reduce:child((+)/3, a+b, X0, Y, Y0).

child(T, T0, A, A0):- child([], T, T0, A, A0).

child(S, X, X0, Y, Y0):- compound(X),
	functor(X, F, N),
	functor(X0, F, N),
	between(1, N, I),
	sgn_index_chk(S, F, N, I),
	arg(I, X, Y),
	copy_args(N, I, Y0, X, X0).

:- meta_predicate(sgn_index_chk(+,+,+,+)).
%  sng_index_chk(+Sgn, +F, +N,  +I): det.
%	True if (F, N, I) is an instance of Signature Sgn.

% ?- reduce:sgn_index_chk(*, f, 3, 2).
% ?- reduce:sgn_index_chk([], f, 3, 2).
% ?- reduce:sgn_index_chk([], *, 3, 2).
% ?- reduce:sgn_index_chk(f, f, 3, 2).
% ?- reduce:sgn_index_chk(f, g, 3, 2).
% ?- reduce:sgn_index_chk((f/3,[1,3]), f, 3, 1).
% ?- reduce:sgn_index_chk(((g/2;f/3), [1,3]), f, 3, 1).
% ?- reduce:sgn_index_chk(\+ [], f, 3, 1).
% ?- reduce:sgn_index_chk(\+ ((*)/2), f, 3, 1).
% ?- reduce:sgn_index_chk(\+ ([]/2), f, 3, 1).
% ?- reduce:sgn_index_chk((*)/2, *, 2, 1).
% ?- reduce:sgn_index_chk([]/2, *, 2, 1).

sgn_index_chk(X,Y,Z,U):- once(sgn_index(X,Y,Z,U)).

:- meta_predicate(sgn_index(+,+,+,+)).

% '*' is a wild-card
% The null list [] is '*'-functor
sgn_index([], _, _, _).
sgn_index(F, G, _, _):- atom(F), G==F.
sgn_index(A;B, F, N, I):-
	(	sgn_index(A, F, N, I)
	;	sgn_index(B, F, N, I)).
sgn_index((A,B), F, N, I):-
	sgn_index(A, F, N, I),
	sgn_index(B, F, N, I).
sgn_index(\+(A), F, N, I):- \+ sgn_index(A, F, N, I).
sgn_index(S/N, F, N, I):- sgn_index(S, F, N, I).
sgn_index([I|_], _, _, I).
sgn_index([_|L], _, _, I):- memberchk(I, L).
sgn_index({A}, F, N, I):- sgn_index_brace(A, F, N, I).

%
sgn_index_brace((A, B), F, N, I):- sgn_index_brace(A, F, N, I), !
	; 	sgn_index_brace(B, F, N, I).
sgn_index_brace(A, F, N, I):- sgn_index(A, F, N, I).

% copy_args(J, I, X, T, T0) :-  T and T0 have the same children
% up to the J-th child except that the I-th child of T0 is X.
% ?- copy_args(3, 2, X, f(a,b,c), f(A,B,C)).

copy_args(0, _, _, _, _):-!.
copy_args(J, I, A, T, T0):-
	(   J==I
	->  arg(J, T0, A)
	;   arg(J, T, X),
	    arg(J, T0, X)
	),
	J1 is J-1,
	copy_args(J1, I, A, T, T0).
%
term_complete([], G, G):-!.
term_complete(A, G, G0):- G =.. H,
	append(H, A, H0),
	G0 =.. H0.

% ?- complete_args(f(a), [b,c], X).
% ?- complete_args(user:m:F,[a,b], R).
% ?- complete_args([]:[]:F,[a,b], R).
% ?- complete_args([]:M:[]:F,[a,b], R).

complete_args(X,  As,  Y):- var(X), !,
	Y=..[call, X|As].
complete_args(M:X, As, Y):- !,
	(	M == []
	-> 	complete_args(X, As, Y)
	;	complete_args(X, As, Y0),
		Y = M:Y0
	).

complete_args(X, [], X):- !.
complete_args(G, As, G0):- atom(G), !,
 	G0 =.. [G|As].
complete_args(G, As, G0):-
	G =.. H,
 	append(H, As, H0),
 	G0 =.. H0.

		/***************************
		*     Simplifying Goals    *
		***************************/

% some tiny.
%  C is safely supposed to be without a module prefix.
canonical_prefix_form(M:C, M, C).
canonical_prefix_form(C, [], C).

% ?-  prefix_var(_:_).
prefix_var(X):- var(X),!.
prefix_var(_:X):- prefix_var(X).

% ?-  prefix_nonvar(_:_).
prefix_nonvar(X):- nonvar(X), !,
				( X = _:Y  -> prefix_nonvar(Y)
				; true
				).

% ?- module(reduce).
% Goal simplification for Prolog *operational* semantics !!
% ?- slim_goal((true, (fail-> X)), R).
% ?- slim_goal((true, (true-> X)), R).
% ?- slim_goal((A,true), R).
% ?- slim_goal((A,false), R).
% ?- slim_goal(true, R).
% ?- listing(slim_goal).


slim_goal(X, Y):- boole_rule(X, X0), !, slim_goal(X0, Y).
slim_goal(X, X).

% Pseudo truth tables for goal simplification.
boole_rule(A, call(A))		 :- var(A).
boole_rule((A, B), (A0, B))	 :- boole_rule(A, A0).
boole_rule((A, B), (A, B0))	 :- boole_rule(B, B0).
boole_rule((A; B), (A0; B))	 :- boole_rule(A, A0).
boole_rule((A; B), (A; B0))	 :- boole_rule(B, B0).
boole_rule((A -> B), (A0->B)) :- boole_rule(A, A0).
boole_rule((A -> B), (A->B0)) :- boole_rule(B, B0).
boole_rule(\+(A), \+(A0))	 :- boole_rule(A, A0).
boole_rule(M:A, M:B)			 :- boole_rule(A, B).
boole_rule(once(A), once(B))	 :- boole_rule(A, B).
boole_rule(once(A), A)		 :- ( A == true; A == false ).
boole_rule((true, A), A).
boole_rule((A, true), A).
boole_rule((false, _), false).
% boole_rule((_, false), false).		% Disabled.
boole_rule((true; true), true).
boole_rule((A; false), A).
boole_rule((false; A), A).
boole_rule((false -> _), false).
boole_rule((true -> A), A).
boole_rule(\+(false), true).
boole_rule(\+(true), false).
boole_rule(X=Y, true)		 :- X == Y.
boole_rule(X=Y, false)		 :- X \== Y, ground((X,Y)).
boole_rule(X==Y, true)		 :- X == Y.
boole_rule(X==Y, false)		 :- X \== Y, ground((X,Y)).
boole_rule(nonvar(X), true)	 :- nonvar(X).
boole_rule(var(X), false)	 :- nonvar(X).
boole_rule(fail, false).
boole_rule(_:(M:A), M:A).
boole_rule(_:E, E)			 :- prefix_cancellable(E).
boole_rule(M:A, A)			 :- M == [].
boole_rule(M:E, ME)			 :- propagate_prefix(M, E, ME).
boole_rule(X, Y)				 :- slim_call(X, Y).

%
propagate_prefix(M, X, X):- M==[], !.
propagate_prefix(M, once(X), once(M:X)).
propagate_prefix(M, (X, Y), (M:X, M:Y)).
propagate_prefix(M, (X; Y), (M:X; M:Y)).
propagate_prefix(M, (X -> Y), (M:X -> M:Y)).
propagate_prefix(M, \+(X), \+(M:X)).

%
% prefix_cancellable(=(_,_)).  % attributes unify_hook
prefix_cancellable(!).
prefix_cancellable(true).
prefix_cancellable(false).

%
optimize_call_N(X, Y):- slim_call(X, Y).
optimize_call_N(X, X).

%
slim_call(X, Y) :-	canonical_prefix_form(X, M, X0),
					reducible_call(X0),
					slim_call(X0, M, Y).

%
reducible_call(X) :- X=..[F, A|_],
				  (F == call; F == apply),
				  prefix_nonvar(A).

% ?- slim_call(apply(a(b), [c,d]), [], R).
% ?- slim_call(apply(F, [a,b]), [], R).
% ?- slim_call(call(call(p(a), b)), m, X).
% ?- slim_call(a, [], Y).
% ?- slim_call(A, M, Y).
% ?- slim_call(call(f, a, b), m, Y).
% ?- slim_call(call(call(f), a, b), m, Y).
% ?- slim_call(apply(call(f), [a,b]), m, Y).

slim_call(X, M, call(Y)):- var(X), !,
	attach_prefix(M, X, Y).
slim_call(M:X, _, Y):-!, slim_call(X, M, Y).
slim_call(X, M, Y):- X =..[call, F|As], !,
	slim_call(F, M, F0),
	complete_args(F0, As, Y).
slim_call(apply(F, As), M, Y):-
	slim_call(F, M, F0),
	complete_args(F0, As, Y).
slim_call(apply(F), M, Y):-
	slim_call(F, M, Y).
slim_call(X, M, Y):- attach_prefix(M, X, Y).

% [2015/03/04] various attach-prefixes integrated.
%%  attach_prefix(M, X, Y) is det.
%	Y is unified with X attached prefix M.

attach_prefix(M, X, Y):- var(X), !,
		( var(M), Y = M:X
		; M == [], Y = X
		; Y = M:X
		),
		!.
attach_prefix(_, X, X)  :- integer(X), !.
attach_prefix(_, X, X)  :- string(X), !.
attach_prefix(_, X, X)	:- X = (_:_), !.
attach_prefix(M, X, X)	:- M == [], !.
attach_prefix(M, X, M:X).

% Let E be a system of equations:
%	m1 = t1
%	m2 = t2
%	...
%	mr  = tr
% The following conditions are equivalent.
% 1.  E has a solution such that no variable x that appears in some
%	ti is not affected by the unification.
% 2.  The list [m1,...,mr] subsumes the list [t1,...,tr]

% ?- reduce_subsumes(f(a, b), f(a,  Q), R).
%@ R = (a==a,b==Q,true).
% ?- reduce_subsumes(f(a, b), f(a,  Q), R), call(R).
%@ false.
% ?- reduce_subsumes(f(a, B), f(a,  b), R), call(R).
%@ B = b,

reduce_subsumes(X, Y, G):- collect_subsumes(X, Y, G),
	check_subsumes_throw(G).

reduce_subsumes(N, X, Y, G):- collect_subsumes(N, X, Y,true,G),
	check_subsumes_throw(G).

check_subsumes_throw(G):- unzip_equality(G, L, R),
	(	subsumes_term(L, R)
	->	true
	;	throw(error('reduce_subsumes rule'))
	).

%
collect_subsumes(X, Y, X = Y):- var(X), !.
collect_subsumes(X, Y, X == Y):- atomic(X), !.
collect_subsumes(X, Y, (nonvar(Y), Y=Y0, G)):- var(Y), !,
	functor(X, F, N),
	functor(Y0, F, N),
	collect_subsumes(N, X, Y0, true, G).
collect_subsumes(X, Y, G):- functor(X, F, N),
	functor(Y, F, N),
	!,
	collect_subsumes(N, X, Y, true, G).

collect_subsumes(0, _, _, G, G):-!.
collect_subsumes(J, X, T, G0, G):- arg(J, X, A),
	arg(J, T, B),
	collect_subsumes(A, B, H),
	K is J - 1,
	collect_subsumes(K, X, T, (H, G0), G).

% ?- unzip_equality((a=b, c=d), P, Q).
%@ P = [c,a],
%@ Q = [d,b].

unzip_equality(E, P, Q):- unzip_equal_dlist(E, []-P, []-Q), !.

unzip_equal_dlist((E,F), P-Q, R-S):- unzip_equal_dlist(E, P-T, R-U),
	unzip_equal_dlist(F, T-Q, U-S).
unzip_equal_dlist(X=Y, A-[X|A], B-[Y|B]).
unzip_equal_dlist(_, P-P, Q-Q).
