:- module(automaton, [automatondemo/0]).
% :- set_prolog_flag(unknown, fail).
:- set_prolog_flag(double_quotes, codes).

% ?- gv(a).
%% Generating automata for regular expressio
%%     by Kuniaki Mukai  (2003/11/27, 2003/12/16, 2004/12/04, 2006/09/30)
%%
:- op(500, yfx, &).

:- dynamic initial/1, final/1.
:- dynamic gensym_count/1.
:- dynamic next/3.
:- retractall(gensym_count(_)), assert(gensym_count(0)).

gensym(X, Y):- retract(gensym_count(C)),
	C0 is C+1, assert(gensym_count(C0)),
	atomic_list_concat([X, C0], Y).

dir_path(deldel). % directory for temporary files  .pdf .png .bb  etc
file_name(deldel). % file name without extension.

%% For demo.
automatondemo:-prompt(A,''), automatondemo(A,_).

automatondemo(N):-sample(N,X),format("~w~n",[X]), gv(X).

automatondemo(_,N):-sample(N,X),format("~w~n",[X]), gv(X), get0(_), fail.
automatondemo(A,_):-prompt(_,A).

%% Sample Regular expression
sample(1,epsilon).
sample(2,empty).
sample(3,a).
sample(4,a ^ b).
sample(5,a ^ b ^c).
sample(6,a + b).
sample(7,a + a + a).
sample(8,&(a,b)).
sample(9,&(*(a),*(a+b))).
sample(10,*(a) + *(b)).
sample(11,*(a) ^ b).
sample(12,*(a)^ *(b)).
sample(13,*(a)^ *(a)).
sample(14,*(*(a))).
sample(15,*(a) - *(a)).
sample(16,*(a) - epsilon).
sample(17, (a+ *(b))^ (*(a)+b)).
sample(18, *(*(a)^b^ *(a)^ b ^ *(a))).
sample(19, &(*(*(a)^b^ *(a)^ b ^ *(a)), *(*(b)^a^ *(b)^ a ^ *(b)))).
sample(20, *(a+ *(b^ *(a)^b))).    % Parity check automaton
sample(21, &(*(a+ *(b^ *(a)^b)),   % Doubly parity check automaton
	     *(b+ *(a^ *(b)^a)))).
%% for showing automaton in graphics using GraphViz.
gv(R) :- graphviz(R).

graphviz(R) :- 	symbols(R,I,L),!,graphviz(R,I,L).

graphviz(R,[],_):- !, graphviz(R,[a]).
graphviz(R,I,L):- length(I,N), N =< 3, L=<30, !, graphviz(R,I).
graphviz(_,_,_):- format(
	"<p> Too long expression or more than 3 input symbols</p>",[]).

graphviz(R,I) :- reg2aut(R,Aut, I), autviz(Aut).

av(A) :- autviz(A).
av(A,M) :- write(M), nl, autviz(A), get0(_).

autviz(Aut):-
	DOT = 'DOTTEMP.dot',
	PDF = 'DOTTEMP.PDF',
	file(DOT, write, automaton:aut2dot(Aut)),
	atomic_list_concat(['dot -Tpdf -o ', PDF, ' ', DOT], DotCom),
	shell(DotCom),
%	atomic_list_concat(['open ', '-a Graphviz ', DOT], Comm),
	atomic_list_concat(['open ', PDF], Comm),
	shell(Comm).

% sample runs:
% ?- match([0,1,(.),1,0], (0+1)^ *(0+1) ^ (epsilon+ (.) ^ *(0+1))).
% ?- match([1,2,3],*(-(0))).
% ?- match([1,1], *(1) + *(0)).
% ?- automaton((*(0+1))-(*(0)^*(1))).

% for compatibility
new_atom(X):- gensym(q,X).

% ex.  symbols(a+b, X) ==> X=[a,b].

symbols(X,Y):- eval(symbols, X, (Y,_)).
symbols(X,Y,L):- eval(symbols, X, (Y,L)).

symbols(epsilon, ([],1), [], [], true).
symbols(empty, ([],1), [], [], true).
symbols(*(X), (Z,L), [X], [(Z,L0)],  L is L0+1).
symbols(-(X), (Z,L), [X], [(Z,L0)],  L is L0+1).
symbols(F, (Z,L), [X, Y], [(Z1,L1), (Z2,L2)],
	(union(Z1,Z2,Z), L is L1+L2+1)) :-
	F=..[Op,X,Y], member(Op,[(^),(+),(-),(&)]).
symbols(A, ([A],1), [], [],true).

% ex. automaton(a+b) ==> will display the created automaton.

automaton(R):-
	symbols(R,I),
	reg2aut(R,I,am(S,M,Ini,Fin)),
	format("Regular expression = ~w~n",[R]),
	format("Initial state = ~w~n",[Ini]),
	format("List of final states = ~w~n",[Fin]),
	format("List of states = ~w~n",[S]),
	format("State transitions:~n"),
	(member(mv(A,B,C),M),
	 format("    ~w ----- ~w -----> ~w~n",[A,B,C]),fail
	;          % or
	 true).

edge_in_dot(X,A,Y):- format("~w -> ~w [label = ~w];~n",[X,Y,A]).

aut2dot(am(_,M,Ini,Fin)):-
     format("digraph g {~n",[]),
     format("rankdir=LR;~n",[]),
     format("dummy [shape = none, label = """"];~n",[]),
     final_in_dot(Fin),
     move_in_dot([mv(dummy,'""',Ini)|M]),
     format("}~n",[]).

final_in_dot([]).
final_in_dot([X|Y]):-format("~w [shape = doublecircle];~n", [X]),
	final_in_dot(Y).

move_in_dot([]).
move_in_dot([mv(X,A,Y)|Z]):-edge_in_dot(X,A,Y), move_in_dot(Z).

% match([a],a) ==> yes
% match([a],b) ==> undefined symbol in the input.
% match([a,a], *(a)) ==> yes
% match([a,b],(-b)^(-a))  ==> yes

match(Y,X):-
	make_and_assert(X,initial,final,next),
	symbols(X,I),
	(subset(Y,I) -> accept(Y,initial,final,next)
	  ;
	format("undefined symbol in the input: ~w~n",[Y])).

% 'accept' is a predicate to test whether input is accepted by automaton
% (<initial sate>, <finail states>, <state transition function>)
%
% accept(X,S,F,E)  if the automaton (?, ?, E, S, F) accepts the input X.
%

accept(X,S,F,E):-
	T=..[S,U],
	call(T),
	accept1(X,U,F,E).

accept1([],S,F,_):-!, applyfinal(F,S).
accept1([X|Y],S,F,E):-
	applynext(E,S,X,U),
	accept1(Y,U,F,E).

applyfinal(F,S):-T=..[F,S], call(T).

applynext(E,S,X,U):-T=..[E,S,X,U], call(T).

% main predicate
% generete an automaton for the given regular expression.
% reg2aut(<regular expression>, <input symbols>, <Automaton>).
reg2aut(X,Am,I):-
       compose_am(X,Am1,I),
       minimize_am(Am1,Am2),
       removeredundant(Am2,Am3),
       refresh_am(Am3,Am).
%%       av(Am, I=X).

removeredundant(am(_,E,St,F),am(D,E1,St,F1)):-
       sptree(St,E,D),
       intersection(F,D,F1),
       filterlinks(D,E,E1).

filterlinks(_,[],[]).
filterlinks(D,[mv(X,_,_)|R],Z):- (\+ member(X,D)),!,filterlinks(D,R,Z).
filterlinks(D,[H|R],[H|Z]):-filterlinks(D,R,Z).

refresh_am(X,Y):- refresh_am(X,0,Y).

refresh_am(am(S,D,St,F),First,am(S1,D1,St1,F1)):-
	makeassoc(First,S,Map),
	assoclist(S,Map,S1),
	mapmove(D,Map,D1),
	member(St-St1,Map),
	assoclist(F,Map,F1).

% make non-determinitic moves into deterministic ones
%
% ?- closure(([a,b],[mv(p,a,p),mv(p,a,q),mv(p,b,p)]), [[p]], X,Y).
%
% X = [[p, q], [p]]
% Y = [mv([p, q],a,[p, q]), mv([p,q],b,[p]), mv([p],a,[p,q]),
%     mv([p], b, [p])]
%

closure(E,X,Y,Z):-closure(E,X,[],[],Y,Z).

closure(_,[],X,Y,X,Y).
closure(E,[S|R],X,Y,Z,U):-member(S,X),!,closure(E,R,X,Y,Z,U).
closure(E,[S|R],X,Y,Z,U):-
	successors(E,S,L,M),
	append(L,R,LR),
	append(M,Y,MY),
	closure(E,LR,[S|X],MY,Z,U).

successors((I,E),X,L,M):-
	setof(mv(X,A,Y),
	      (member(A,I),setof(Z,K^(member(K,X),member(mv(K,A,Z),E)),Y)),M),
	setof(U,X1^B^member(mv(X1,B,U),M),L).

% ex. makeassoc([a,b,c],X) => X = [a-0, b-1, c-2]

makeassoc(X,Y):-makeassoc(0,X,Y).

makeassoc(_,[],[]).
makeassoc(N,[X|Y],[X-N|R]):-N1 is N+1, makeassoc(N1,Y,R).

assoclist([],_,[]).
assoclist([X|Y],Map,[N|B]):-member(X-N,Map),assoclist(Y,Map,B).

mapmove([],_,[]):-!.
mapmove([mv(X,L,Y)|Z], Map, [mv(X1,L,Y1)|Z1]):-
	member(X-X1,Map),
	member(Y-Y1,Map),
	mapmove(Z,Map,Z1).

make_and_assert(X,State,Final,Next):-
	symbols(X,I),
	reg2aut(X,I,am(_,M,Ini,Fin)),
	assert_automaton(State,Ini,Final,Fin,Next,M).

assert_automaton(Sn,S,Fn,F,An,A):-
	abolish(Sn/1),
	abolish(Fn/1),
	abolish(An/3),
	T=..[Sn,S],
	asserta(T),
	assertlist(Fn,F),
	assertarrowlist(An,A).

assertlist(_,[]):-!.
assertlist(F,[X|Y]):-
	T=..[F,X],
	asserta(T),
	assertlist(F,Y).

assertarrowlist(_,[]):-!.
assertarrowlist(A, [mv(X,L,Y)|Z]):-
	T=..[A,X,L,Y],
	asserta(T),
	assertarrowlist(A,Z).

final(am(_,_,_,X),X).


% ex. compose_am(a,X,[a]) =>
% X =am([q1,q2,q3],[mv(q1,a,q2),mv(q21,a,q3),mv(q3,a,q3)],q1,[q2])


compose_am(empty, am([S],Trans,S,[]),I):-!,
	new_atom(S),
	setof(mv(S,A,S),member(A,I),Trans).
compose_am(epsilon,am([S,T],Trans,S,[S]),I):-!,
	new_atom(S),
	new_atom(T),
	setof(mv(Q,A,T),(member(Q,[S,T]),member(A,I)),Trans).
compose_am(Z,am(S,D,St,F),I):-
	Z =.. [Op,X,Y],
	member(Op,[(+),(-),(&)]),
	!,
	reg2aut(X,AX,I),
	reg2aut(Y,AY,I),
	product(AX,AY,[S,D,St]),
	final(AX,FX),
	final(AY,FY),
	Fexp =..[Op,FX,FY],
	makefinal(Fexp,S,F).
compose_am(^(X,Y),am(S,E,[St1],F),I):-!,
	reg2aut(X, am(S1,E1,St1,F1),I),
	reg2aut(Y, AmY,I),
	length(S1,L),
	L1 is L+1,
	refresh_am(AmY,L1,am(_,E2,St2,F2)),
	cartesian(F1,[St2],E_links),
	tcl(E_links, E_tcl),
	append(E1,E2,E3),
	extend_by_emove(E_tcl, E3, E4),
	closure((I,E4),[[St1]],S,E),
        extend_set_by_emove(F2,E_tcl,F3),
	(setof(P,Q^(member(P,S),member(Q,P),member(Q,F3)),F),!; F=[]).
compose_am(*(X),am(S,R,[NewSt],Fin), I):-!,
	reg2aut(X,am(_,E,St,F),I),
	new_atom(NewSt),
	cartesian(F,[NewSt],Emoves),
	tcl([(NewSt,St)|Emoves],TclE),
	extend_by_emove(TclE,E,D),
	closure((I,D),[[NewSt]],S,R),
	extend_set_by_emove([NewSt],TclE,Fin1),
	(setof(P,Q^(member(P,S),member(Q,P),member(Q,Fin1)),Fin),!; Fin=[]).
compose_am(-(X),am(S,R,St,Fin), I):-!,	reg2aut(X,am(S,R,St,F),I),
	subtract(S,F,Fin).
compose_am(A, am([P,Q,R],[mv(P,A,Q)|Trans],P,[Q]),I):-
	new_atom(P),
	new_atom(Q),
	new_atom(R),
	setof(mv(S,B,R),(member(S,[P,Q,R]),member(B,I)),Moves),
	remove(mv(P,A,R),Moves,Trans).

extend_by_emove(X,Y,Z):- maplist(insert_link_label(epsilon),X,X1),
	add_join(X1,Y,Y1),
	add_join(Y1,X1,Z1),
 	remove_emove(Z1,Z).

insert_link_label(X,(Y,Z), mv(Y,X,Z)).

extend_set_by_emove(X,Y,Z):-choose((A,B),Y,Y1),member(B,X),!,
	extend_set_by_emove([A|X],Y1,Z).
extend_set_by_emove(X,_,Y):-sort(X,Y).

remove_emove([],[]).
remove_emove([mv(_,epsilon,_)|X],Y):-!,remove_emove(X,Y).
remove_emove([X|Y],[X|Z]):-remove_emove(Y,Z).

makefinal(FX + FY,S,F):-
	setof((P,Q),(member((P,Q),S),(member(P,FX);member(Q,FY))),F),!.
makefinal(&(FX,FY),S,F):-
	setof((P,Q),(member((P,Q),S),(member(P,FX),member(Q,FY))),F),!.
makefinal(-(FX,FY),S,F):-
	setof((P,Q),(member((P,Q),S),(member(P,FX), (\+ member(Q,FY)))),F),!.
makefinal(_,_,[]).

product(am(X1,Y1,Z1,_),am(X2,Y2,Z2,_),[X3,Y3,(Z1,Z2)]):-
	cartesian(X1,X2,X3),
	setof(mv((P1,P2),A,(Q1,Q2)),
	      (member(mv(P1,A,Q1),Y1), member(mv(P2,A,Q2),Y2)),Y3).

%% For testing the pathexpression predicate defined below.

reg2reg(R,R1):-symbols(R,I), reg2aut(R,I,A),aut2reg(A,R1).

%% Computing regular expressions from automata.

aut2reg(am(S,E,St,F), R):-aut2reg(St,F,E,S,R1),reduce(R1,R).

aut2reg(_,[],_,_,empty).
aut2reg(X,[Y|Z],E,S,P+Q):- pathexp(X,Y,E,S,P), aut2reg(X,Z,E,S,Q).

% pathexp(X,Y,E,N,P)

pathexp(X,X,E,N,*(C)):-!,remove(X,N,N1),cycle(X,E,N1,C).
pathexp(X,Y,E,N,*(C) ^ Q):-
	remove(X,N,N1),
	cycle(X,E,N1,C),
	pathexp1(X,Y,E,E,N1,Q).

%% path1(X,Y,E,F,N,P)==> P is a  path expression denoting the set
%%   of paths from X to Y  the node of which on but last  on the
%%   is in the list N of nodes.

pathexp1(_,_,[],_,_,empty).
pathexp1(X,Y,[mv(X,A,Z)|R],E,N,(A^P)+Q):-
	member(Z,N),!,
	pathexp(Z,Y,E,N,P),
	pathexp1(X,Y,R,E,N,Q).
pathexp1(X,Y,[mv(X,A,Y)|R],E,N,A+Q):-!,pathexp1(X,Y,R,E,N,Q).
pathexp1(X,Y,[_|R],E,N,P):-pathexp1(X,Y,R,E,N,P).

cycle(X,E,N,P):-cycle(X,E,E,N,P).

cycle(_,[],_,_,empty).
cycle(X,[mv(X,A,X)|R],E,N,A+P):-!,cycle(X,R,E,N,P).
cycle(X,[mv(X,A,Y)|R],E,N,(A^P)+Q):-member(Y,N),!,
	pathexp1(Y,X,E,E,N,P),
	cycle(X,R,E,N,Q).
cycle(X,[_|R],E,N,P):-cycle(X,R,E,N,P).

% equations to simpify regular expressions.
equal(epsilon^X,X).
equal(X^epsilon,X).
equal(_^empty,empty).
equal(empty^_,empty).
equal(empty+X,X).
equal(X+empty,X).
equal(*(*(X)), *(X)).
equal(*(empty), epsilon).
equal(*(epsilon), epsilon).

% Applying the equations as far as they are applicable.
reduce(X,Y):-reduceone(X,X1),!,reduce(X1,Y).
reduce(X,X).

reduceone(X,X1):-equal(X,X1).
reduceone(X+Y,X1+Y):-reduceone(X,X1).
reduceone(X+Y,X+Y1):-reduceone(Y,Y1).
reduceone(X^Y,X1^Y):-reduceone(X,X1).
reduceone(X^Y,X^Y1):-reduceone(Y,Y1).
reduceone(&(X,Y),&(X1,Y)):-reduceone(X,X1).
reduceone(&(X,Y),&(X,Y1)):-reduceone(Y,Y1).
reduceone(*(X),*(X1)):-reduceone(X,X1).

%% States minimization. The idea (based on Hopcroft etal):
%%   first, make the maximum incompatible relation.  Then
%%   the complement of the relation is the equivalence
%%   relation what we want.
%%
%%   equivrel([1,2,3],[2,3],
%%   [mv(1,a,2),mv(1,b,3),mv(2,a,2),mv(2,b,2),mv(3,a,3),mv(3,b,3)], X)
%%  ==>  X=[(2,3)]
%% equivrel([1,2,3],[1],
%%  [mv(1,a,2),mv(1,b,3),mv(2,a,2),mv(2,b,2),mv(3,a,3),mv(3,b,3)], X)
%%    ==> X=[(2,3)]
%%

minimize_am(am(S,T,St,F),am(QS,QT,QSt,QF)):-
	equivrel(S,F,T,R),
	qstate(R,S,QS),
	qdelta(T,QT,QS),
	qfinal(F,QF,QS),
	qstart(St,QSt,QS).


equivrel(S,F,M,E):-
        subtract(S,F,S1),
	pairs(S,S,S2),
	pairs(S1,F,P),
	subtract(S2,P,S3),
	fill_table(S3,P,mv_triples(M),E).

mv_triples(M, X, A, Y):- member(mv(X,A,Y), M).

%%
%%  Table filling method in Prolog for states minizisation
%%  c.f. Hopcroft etal's book
%%  Amazing method !
%%  ?- fill_table([ (q10, q12)], [ (q11, q12), (q10, q11)],
%%     [mv(q10, a, q11), mv(q11, a, q12), mv(q12, a, q12)], E).
%%  =>   E=[].
%%

fill_table(R,P,M,E):-choose(X,R,R1),separable(X,M,P),!,
	fill_table(R1,[X|P],M,E).
fill_table(R,_,_,R).

separable((X,Y), M, P):- call(M, X, A,X1),
	call(M, Y, A, Y1),
	pair_member((X1,Y1),P).

pair_member((X,X),_):-!,fail.
pair_member((X,Y),A):- X @> Y, !, memberchk((Y,X),A).
pair_member(Z,A):- memberchk(Z,A).

choose(X,[X|R],R).
choose(X,[A|R],[A|S]):-choose(X,R,S).

%% Making Quotient Automaton from equivalence relation
%%
%% qstate(S,R,Q)  Q = S/R
%% ?- qstate([(a,b),(c,d)],[a,b,c,d],X).
%%   ==> X = [[a, b], [c, d]]

qstate(R,X,Y):-singleton(X,X1),qstate1(R,X1,Y).

qstate1([],X,X).
qstate1([(A,B)|R],X,Z):- mergeclass(A,B,X,X1),qstate1(R,X1,Z).

qdelta(X,Y,Q):-qdelta(X,[],Y,Q).
qdelta([],M,M,_).
qdelta([mv(U,A,V)|R],M,N,Q):-
	member(C,Q),member(U,C),
	member(D,Q),member(V,D),
	!,
	addnew(mv(C,A,D),M,M1),
	qdelta(R,M1,N,Q).

qfinal(X,Y,Q):-qfinal(X,[],Y,Q).

qfinal([],M,M,_).
qfinal([U|R],M,N,Q):-
	member(C,Q),member(U,C),
	!,
	addnew(C,M,M1),
	qfinal(R,M1,N,Q).

qstart(X,C,Q):-member(C,Q),member(X,C),!.

%%
%% Join graphs as relations.  (2004-12-2)
%%

add_join(X,Y,Z):-join(X,Y,Z1), append(X,Y,Z2), append(Z1,Z2,Z).

join(X,Y,Z):-join0(X,Y,Z-[]).

join0([],_,L-L).
join0([X|Y],Z,U-V):- join1(X,Z,U-W), join0(Y,Z,W-V).

join1(_,[],L-L).
join1(X,[Y|Z],[U|V]-W):-join2(X,Y,U),!,join1(X,Z,V-W).
join1(X,[_|Y],Z):-join1(X,Y,Z).

join2(mv(A,F,B),mv(B,G,C), mv(A,D,C)):- compose_label(F,G,D),!.

compose_label(epsilon,X,X).
compose_label(X,epsilon,X).
compose_label(X,Y,(X;Y)).

%% Computiong the  transitive closure Tcl of a graph E
%%
%% tcl: Transitve CLosure.
%% tcl([(a,b),(b,c),(c,a)],X)  ==>
%%  X = [ (a, a), (b, b), (c, b), (c, c), (b, a),
%%        (a, c), (a, b), (b, c), (c,a)]

tcl(E,Tcl):-tcl(E,E,Tcl).

tcl([],X,X):-!.
tcl(A,X,Y):-tcl(A,[],X,A1,X1),tcl(A1,X1,Y).

tcl([],A,X,A,X).
tcl([A|R],B,X,C,Y):-
	tcl(A,X,B,X,B1,X1),
	tcl(R,B1,X1,C,Y).

tcl(_,[],B,X,B,X).
tcl(A,[B|R],C,X,D,Y):-join_link(A,B,P),!,
	addnew(P,C,X,C1,X1),
	tcl(A,R,C1,X1,D,Y).
tcl(A,[_|R],C,X,D,Y):-tcl(A,R,C,X,D,Y).

join_link((A,B),(B,C),(A,C)).

%% sptree(R,E,N) computes the list N of all nodes
%% connencted to the root R in the graph E.
%% sptree(a,[mv(a,1,b), mv(c,1,a), mv(b,1,a)],X) ==> X = [b, a]
%%

sptree(Root,E,N):-sptree(E,[Root],[],N).

sptree(_,[],X,X).
sptree(E,[H|R],X,Y):-member(H,X),!,sptree(E,R,X,Y).
sptree(E,[H|R],X,Y):-sptree(H,E,R,E1,R1),sptree(E1,R1,[H|X],Y).

sptree(_,[],R,[],R).
sptree(H,[mv(H,_,X)|E],R,E1,[X|R1]):-!,sptree(H,E,R,E1,R1).
sptree(H,[A|E],R,[A|E1],R1):-sptree(H,E,R,E1,R1).

%%
%% Below this line are general purely set-theoretical predicates.
%%

addnew(X,D,D):-member(X,D),!.
addnew(X,D,[X|D]).

addnew(P,C,X,C,X):-member(P,X),!.
addnew(P,C,X,[P|C],[P|X]).

% singleton([a,b,c],X) ==> X=[[a], [b], [c]]
singleton([],[]).
singleton([X|Y],[[X]|Z]):-singleton(Y,Z).

% mergeclass(a,b, [[a,c],[d],[b,e]],X).
%   ==>  X = [[d], [a,c,b,e]]
mergeclass(_,_,[],[]).
mergeclass(A,B,[X|Y],Z):-member(A,X),!,mergeclass1(X,B,Y,Z).
mergeclass(A,B,[X|Y],Z):-member(B,X),!,mergeclass2(X,A,Y,Z).
mergeclass(A,B,[X|Y],[X|Z]):-mergeclass(A,B,Y,Z).

mergeclass1(X,A,Y,[X|Y]):-member(A,X),!.
mergeclass1(X,A,Y,Z):-mergeclass2(X,A,Y,Z).

mergeclass2(X,A,[Y|Z],[XY|Z]):-member(A,Y),!,append(X,Y,XY).
mergeclass2(X,A,[Y|Z],[Y|U]):-mergeclass2(X,A,Z,U).

%% cartesian([a,b,c],[1,2],X) => [ (a,1),(a,2),(b,1),(b,2),(c,1),(c,2)]

cartesian([],_,[]).
cartesian([A|B],C,D):-cartesian(B,C,E),
	cartesian1(A,C,F),
	append(F,E,D).

cartesian1(_,[],[]).
cartesian1(A,[B|C],[(A,B)|D]):-cartesian1(A,C,D).

%% Make "unordered" pairs (not in the sense of  orderd pairs)
%% from two lists.
%% pairs([a,x],[b,y],X). ==>
%%   X = [ (x, y), (b, x), (a, y), (a, b)]
pairs(X,Y,Z):-pairs(X,Y,[],Z).

pairs([],_,X,X).
pairs([A|B],C,D,E):-pairs1(A,C,D,E1), pairs(B,C,E1,E).

pairs1(_,[],X,X).
pairs1(A,[A|C],D,E):-!, pairs1(A,C,D,E).
pairs1(A,[B|C],D,E):-A @> B, !, addnew((B,A),D,D1),pairs1(A,C,D1,E).
pairs1(A,[B|C],D,E):-addnew((A,B),D,D1),pairs1(A,C,D1,E).

remove(_,[],[]):-!.
remove(X,[X|Y],Y):-!.
remove(X,[Y|Z],[Y|U]):-remove(X,Z,U).

%%%%%%%%  For Web

reg2html(F,D,C,R,[],_,T,H):- !, reg2html(F,D,C,R,[a],T,H).
reg2html(F,D,C,R,I,L,T,H):- length(I,N), N =< 3, L=<30, !,
	reg2html(F,D,C,R,I,T,H).
reg2html(_,_,_,_,_,_,_,
	 "<p> Too long expression 30 or more than 3 input symbols</p>").

reg2html(F,D,C,R,I,T,H) :-
        reg2aut(R,am(S,M,Ini,Fin),I),
	H1 = (format_codes("<p>Regular expression = ~w</p>",[R]) &
	format_codes("<p>Initial state = ~w</p>",[Ini]) &
	format_codes("<p>Final states = ~w</p>",[Fin]) &
	format_codes("<p>All states = ~w</p>",[S]) &
	format_codes("<p>State Transitions:</p>~n",[])),
	atomic_list_concat([D, '/', F, C, '.dot'], DotName),
	file(DotName, write, automaton:aut2dot(am(S,M,Ini,Fin))),
	formatForHtml(F,D,C,T,Format,Args),
	print_moves(M,M1),
	eval(H1 & format_codes(Format,Args) & M1, H).

formatForHtml(F,D,C,X,Format,[F,C,X]) :-!,
        img_frame(Format),
        atomic_list_concat([D, (/), F, C], Base),
	(   X == pdf
	->  Ext = ps2,
	    Com = ps2pdf("-sOutputFile="+ Base+ "."+ pdf,
			 Base+ "."+ Ext)
    	;   Ext = X,  Com = "DUMMY=1"
	),
        eh:sh(dot(-'T'(Ext), Base+ (.)+ dot, -o(Base+(.)+Ext)); Com).

img_frame(X) :- flatten([ "<p><div ",
	  "id='diagram' ",
	  "style='border : solid 2px #ff0000; ",
	  "width : 600px; ",
	  "height : 300px; ",
	  "overflow : auto; '><br/>",
	  "<img src=""~w~w.~w""/>",
	  "</div></p>~n"
	 ], X).

print_moves(M,H):-
	maplist(print_moves_x, M, T),
	flatten(T,T1),
	flatten(["<pre>\n",T1, "</pre>\n"], H).

print_moves_x(mv(X,A,Y), H):-
	format_codes("   ~w----~w--->~w~n",[X,A,Y], H).
