:- module(file,
	  [file/3, combine_file/2, clean_io/3,
%	   dsnap/3,
	   symbolic_link/2,
	   tmp_file_name/1,
	   cat_files_to_codes/3,
	   cat_files/2,
	   dict_of_files/1
	   ]).

%;; (setq module-query  "qcompile(util(file)), module(file).")
% ?- qcompile(util(file)), module(file).

:- use_module(pac(basic)).

:- meta_predicate file(+, ?, 0).
:- meta_predicate file(+, ?, 0, ?).
:- meta_predicate snap(1, ?).

		/**************
		*     file    *
		**************/

%%% standard I/O (read/write/append) files

file(X, Y, Z):- file(X,Y,Z,[encoding(utf8)]).

file(X, Y, Z, Ops):- once(file_(X,Y,Z,Ops)).

file_(F, M, A, Ops):-
        (	F = pipe(_) -> F0 = F
        ;	expand_file_search_path(F, F0)  -> true
		;	F0 = F
        ),
        open(F0, M, S, Ops),
        clean_io(S, M, A),
        close(S).
%
clean_io(F, read, A) :- !,
		current_input(Old),
		set_input(F),
        once(A),
		set_input(Old).
clean_io(F, _, A) :-
		current_output(Old),
		set_output(F),
        once(A),
		set_output(Old).

f_pipe(Data, Pipe) :-
		file(	pipe(Pipe),
				write,
                basic:smash(Data),
                [encoding(octet)]). % was utf8

combine_file(Ls, F) :-
	maplist(through_list_or_string, Ls, CombinedCodes),
	file(F, write, basic:smash(CombinedCodes)).

%
cat_files(G, F):- atomic(G), !,
	cat_files([G], F).
cat_files(Fs, F):-
	tmp_file_stream(text, TMPfile, Stream),
    close(Stream),
	maplist(expand_file_name, Fs, Gs),
	cat_files_rec(Gs, TMPfile),
	expand_file_name(F, [F0|_]),
	rename_file(TMPfile, F0).

cat_files_rec([], _).
cat_files_rec([X|Xs], T):- atomic(X), !,
	pshell(cat(X) >> T),
	cat_files_rec(Xs, T).
cat_files_rec([X|Xs], T):-
	cat_files_rec(X, T),
	cat_files_rec(Xs, T).

%
cat_files_to_codes([], C, C).
cat_files_to_codes([P|Q], C, C0):-
    	read_file_to_codes(P, C, [tail(C1), encoding(utf8)]),
	cat_files_to_codes(Q, C1, C0).

%
through_list_or_string(X, X):- (listp(X); string(X)), !.
through_list_or_string(X, Y):- call(X, Y).

%%	symbolic_link(+Target:string, +Link:string) is det.
%	Make a symbolic link Link to Target via a shell call.
symbolic_link(Target, Link) :- 	expand_file_name(Link, [Link0]),
	(	read_link(Link0,_,_)
	 ->	true
 	 ;	expand_file_name(Target, [Target0]),
		eh:sh(ln(-s, Target0, Link0))
	),
	!.

% tmp_file_name(File):- tmp_file_stream(text, File, Stream), close(Stream).
tmp_file_name(File):- tmp_file_stream(utf8, File, Stream), close(Stream).

push_to_file(M, F):- expand_file_name(F, [F1]),
	(	exists_file(F1)
	-> 	tmp_file_name(T1),
		file(T1, write, M),
		tmp_file_name(T2),
		pshell(cat(T1, F) + ' > ' +  T2),
		rename_file(T2, F1),
		delete_file(T1)
	;	file(F1, write, M)
	).

% ?- open_url('http://web.sfc.keio.ac.jp/~mukai/paccgi7/index.html', IN).
% sample coding
open_url(URL, In, Option) :-
        tmp_file_stream(text, File, Stream),
        close(Stream),
        process_create('/usr/bin/curl', ['-o', File, URL], []),
        open(File, read, In, Option),
        delete_file(File).  % Unix-only
%
open_url(URL, In) :- open_url(URL, In, []).

%%  dict_of_files(-D:<dict-like structure>) is det.
%   List files recursively under the working directory, and
%   unify D with the result in the form of a dict with
%	<directory name> - <dict> for subdirectories.
%
% ?- file:dict_of_files(D).

ignore_special((.)).
ignore_special((..)).
ignore_special(('.DS_Store')).
ignore_special(('.git')).

%
dict_of_files(X):- directory_files((.), Files),
		directory_files(Files, X, []).
%
directory_files([], X, X).
directory_files([F|R], X, Y):- ignore_special(F), !,
	   directory_files(R, X, Y).
directory_files([D|R], [D-Z|X], Y):- exists_directory(D), !,
		directory_files(D, Files),
		working_directory(_, D),
		directory_files(Files, Z, []),
		working_directory(_, (..)),
		directory_files(R, X, Y).
directory_files([F|R], [F|X], Y):- directory_files(R, X, Y).

		/**************
		*     snap    *
		**************/

user:snap(C)  :- getenv(snapshot, File),
	do_snap(File, basic:smash, [C,"\n"]).
%
user:snap(M, C):- user:snap(M >> C).
%
user:snap(M, X, X):- basic:smash(X, X0), user:snap(["\n", M, "\n", X0]).
%
user:dsnap(X, X):- user:snap(X).

%
do_snap(File, M, X):-
	current_output(Old),
	open(File, append, New, [encoding(utf8)]),
	set_output(New),
	call(M, X),
	set_output(Old),
	close(New).
