:-module(models,[]).
:- use_module(util(math)).
:- use_module(util(misc)).
:- use_module(pac(basic)).
:- use_module(pac(meta)).

%  models.pl   一階述語論理式のインタプリタ
%
%                        by  向井 国昭   1994年4月28日
%                                        2006年10月　　 　CGI化
%
% 制限： モデルの個体領域は, 有限個のアトムからなる： 関数記号は
% 扱わない。
%
% 使い方： まず個体領域と基本事実の集まりを次の例のように与える。
%   individual(taro).
%   individual(taro).
%   individual(hanako).
%   fact(love(taro,hanako)).
%   fact(love(hanako,hanako)).
%   fact(love(jiro,hanako)).
%
% このモデルで 論理式 たとえばall(x,some(y,love(x,y)))の真偽は判定するには
% 次のように実行する：
%
% | ?- models([], all(x,some(y,love(x,y))).
%
%
%  例：すぐ下に与えるモデルが次の論理式を充足するかいなかを
%       計算しよう。
%
% 用意された例題を動かすには次を実行すればよい:
%
% | ?- start.
%
% all(x,some(y, love(x,y))).
% some(x,all(y, love(x,y))).
% all(x,all(y, love(x,y))).
% some(y,some(x, love(x,y))).
% all(x,some(y, love(x,y))).
% some(x,all(y, love(x,y))).
% all(x,all(y, love(x,y))).
% some(y,some(x, love(x,y))).

% 被テスト論理式：
% formula(all(x,some(y, love(x,y)))).
% formula(some(x,all(y, love(x,y)))).
% formula(all(x,all(y, love(x,y)))).
% formula(some(y,some(x, love(x,y)))).
% formula(all(x,some(y, love(x,y)))).
% formula(some(x,all(y, love(x,y)))).
% formula(all(x,all(y, love(x,y)))).
% formula(some(y,some(x, love(x,y)))).

%% モデル例
% individual(taro).
% individual(jiro).
% individual(hanako).

% fact(love(taro,hanako)).
% fact(love(hanako,hanako)).
% fact(love(jiro,hanako)).
%% end of example model

start(F,Q,R):- split(F,F1), remove([],F1,F2),
	maplist(herbrand,F2,Facts),
	herbrand(Q,Query),
	individuals(Facts,Inds),
	(models(Inds,Facts,[],Query)-> R1=true; R1=false),
	term_smash0(["<font color=\"red\">", R1, "</font>\n"], R).

set_of_individuals(F,R):- split(F,F1), remove([],F1,F2),
	maplist(herbrand,F2,Facts),
	individuals(Facts,Inds),
	insert(",", Inds, Inds0),
	term_smash0(Inds0,R).

individuals(F,S):- maplist(atoms,F,F1),
	append(F1,F2),
	sort(F2,S).

atoms(X,[X]):-atomic(X),!.
atoms(X,Y):- is_list(X),!,
	maplist(atoms, X, Z),
	append(Z, Y).
atoms(X,Y):- X=..[_|A],
	maplist(atoms, A, B),
	append(B, Y).

% テストのメイン。
% start:-	printfact,
% 	(formula(S), write(S), test(S),nl,fail;true).

% printfact:-a
% 	fact(X),write(X),nl,fail;
% 	write('**** end of facts ****'),nl,nl,nl.

% test(S):- models([],S),!,write('  は真.').
% test(_):- write('  は偽.').
:- op(1100,xfx, <=>).

def_macro(false, not(true)).
% def_macro((X -> Y), or(not(X), Y)).
% def_macro((X <=> Y), and((X->Y), (Y->Y))).
def_macro(imply(X,Y), or(not(X), Y)).
def_macro(iff(X,Y), and(imply(X,Y), imply(Y,X))).
def_macro(all(X,P), not(some(X, not(P)))).

%%  models(Individuals, Facts, Assignments, Formula)
models(_,_,_,true) :-!.
models(X,Y,M,and(P,Q)) :-!, models(X,Y,M,P), models(X,Y,M,Q).
models(X,Y,M,or(P,Q))  :-!, (models(X,Y,M,P); models(X,Y,M,Q)).
models(X,Y,M,not(P))   :-!, \+ models(X,Y,M,P).  % negation of models(M,P)
models(I,F,M,some(X,P)):-!, member(A,I), models(I,F,[X=A|M],P).
models(I,F,M,P):- def_macro(P,P1), !, models(I,F,M,P1).
models(_,F,M,P):- substitute(P,M,Q), member(Q,F).

% % membership
% member(X,[X|_]).
% member(X,[_|Y]):-member(X,Y).

% substitute(X,M,Z)   apply the substitution M to X to get Y.
substitute(X,M,Y):-atomic(X), member(X=Y,M), !.
substitute(X,_,X):-atomic(X),!.
substitute(X,M,Y):-
        functor(X,F,N),
	functor(Y,F,N),
        substitute(N,X,M,Y).

substitute(0,_,_,_):-!.
substitute(J,X,M,Y):-
        arg(J,X,A),
        substitute(A,M,B),
        arg(J,Y,B),
        K is J-1,
        substitute(K,X,M,Y).

%%%%%%%% Yet Another DCG for `Tiny' PTQ.

run :- prompt(A,''),
	(sample(S), get0(_), format("~w.",[S]),
       call(models:S, X), get0(_), nl,
       format("~w~n~n",[X]), fail; prompt(_, A)).

sample(semantics(pn, [john], [male(j),male(b),female(m),unicorn(u), find(m,u),walk(j),walk(b),walk(m)])).
sample(semantics(np, [a, unicorn], [male(j),male(b),female(m),unicorn(u), find(m,u),walk(j),walk(b),walk(m)])).
sample(semantics(s, [a, unicorn, walk], [male(j),male(b),female(m),unicorn(u), find(m,u),walk(j),walk(b),walk(m)])).
sample(semantics(vp, [find, a, unicorn], [male(j),male(b),female(m),unicorn(u), find(m,u),walk(j),walk(b),walk(m)])).
sample(semantics(s, [john,find, a, unicorn], [male(j),male(b),female(m),unicorn(u), find(m,u),walk(j),walk(b),walk(m)])).
sample(semantics(pn, [john], [man(a),find(j,a),walk(j)])).
sample(semantics(pn, [john], [man(a),find(j,a),walk(j)])).
sample(semantics(tv, [find], [man(a),find(j,a),walk(j)])).
sample(semantics(itv, [walk], [man(a),find(j,a),walk(j)])).
sample(semantics(vp, [find, a, unicorn], [unicorn(a),find(j,a),walk(j)])).
sample(semantics(s, [john, walk], [man(a),find(j,a),walk(j)])).
sample(semantics(s, [every, man,  walk], [man(a),find(j,a),walk(j)])).
sample(semantics(s, [every, man,  walk], [man(j),find(j,a),walk(j)])).

s(member(VP,NP)) --> np(NP), vp(VP).

np(apply(currify(A), CN)) --> determiner(A), cn(CN).
np(PN) --> pn(PN).

vp(ITV)  --> itv(ITV).
vp(inverse(currify(TV), NP)) --> tv(TV), np(NP).

itv(ext(walk))-->[walk].

tv(ext(find)) --> [find].
tv(ext(kick)) --> [kick].

pn(filter(j)) --> [john].
pn(filter(b)) --> [bill].
pn(filter(m)) --> [mary].

cn(ext(unicorn)) --> [unicorn].
cn(ext(man)) -->   [man].
cn(ext(woman)) --> [woman].

determiner(q_sem(a)) --> [a].
determiner(q_sem(every)) --> [every].

%%%%% Interpreter

% ?- models:semantics(pn, [john], [man(a),find(j,a),walk(j)],X).
% ?- models:semantics(s, [john, walk], [man(a),find(j,a),walk(j)],X).
% ?- models:semantics(s, [every, man,  walk], [man(a),find(j,a),walk(j)],X).
% ?- models:semantics(s, [every, man,  walk], [man(j),find(j,a),walk(j)],X).
% ?- models:semantics(pn, [john], [man(x),find(j,x),walk(j)],X).
% ?- models:semantics(np, [a, man], [man(j), man(k)],  X).
% ?- models:semantics(np, [a, man], [man(j), man(k)],  X).

semantics(S, Facts, V)  :- semantics(s, S, Facts, V).
%
semantics(C, S, F, V):- call(C, E, S, []),
	individuals(F, Inds),
	sem(Inds, F, E, V).
semantics(_,_,_,'** Category  mismatch ?').
%
sem(_, Fac, ext(R), V) :- !, extension(Fac,R,V1), sort(V1,V).
sem(D, _, filter(S), V) :- !, math:principal_filter(D, S, V).
sem(D, _, q_sem(Q), V) :- !, q_sem(Q, D, V).
sem(D, F, E, V)   :- meta:mapterm_rec(models:sem(D,F), E, E1),
	call(E1,V).

%
extension([],_, []).
extension([A|B],R, [A1|B1]):- A=..[R|L], !,
	convert(L,A1),
	extension(B,R,B1).
extension([_|B],R, B1):- extension(B,R,B1).

convert([],    void):-!.
convert([X],   X):- !.
convert([X|Y], (X,Y1)):- convert(Y,Y1).

%
% apply(F,A,V):- member((B,V),F), subset(A, B), subset(B, A).
apply(F,A,V):- member(B-V,F), subset(A, B), subset(B, A).

%
member(X,Y,true) :- member(X,Y),!.
member(_,_,false).

%
currify --> math:rel_to_fun.

% ?- models:q_sem(a, [1,2,3], R).

q_sem(a, D, R):- !,
	(   powerset(D,PowD),
	    setof((X,Y), (member(X,PowD), member(Y,PowD), meet(X,Y)), R)
	->  true
	;   R=[]
	).
q_sem(every, D, R):- !,
	(   powerset(D,PowD),
	    setof((X,Y), (member(X,PowD), member(Y,PowD), subset(X,Y)), R)
	->  true
	;   R=[]
	).

meet(X,Y):- member(A,X), member(A,Y).

inverse(F, R, S):- inverse_image(F,R,S).
