:- module(elisp, [region/2, region_term/2,
		  overwrite/2, overwrite/3, paragraph/2,
		  line_get/1, global_set_kbd/1, global_unset_kbd/1, dired_mark_swipl/0,
		  prolog_symbol_at_point/1,
		  handle_lisp/3, book_bibitem/2, first_token_codes/2,
		  find_handle_call/4,
		  read_term_from_lisp/1,
		  read_mini_buffer/2,
  		  send_off/1,
		  call_lisp/2, call_lisp/1,
		  call_lisp_wait/1,
		  call_lisp_text/2,
		  lisp/1, lisp/2, lisp_to_list/2,
		  noreply/1,
		  get_string/2, set_string/2,
		  message/1,
		  insert_buffer/1,
		  insert/1,
		  set_mark_region/2,
		  neighbor_string/3, neighbor_string/5,
		  wait/1, wait/2,
		  read_codes/1, read_codes/2,
		  meta_handle/4, normal_keys/3
		  ]).

:- use_module(pac(basic)).
:- use_module(util(file)).
:- use_module(pac(reduce)).
:- use_module(util(polynomial)).
:- use_module(util(tex)).
:- use_module(util(obj)).
:- use_module(util(misc)).

% :- use_module(util('convert-dcg')).
% :- use_module(util('sed(dcg_ts')).

term_expansion --> pac:expand_pac.
:- use_module(pac('expand-pac')).
:- use_module(pac(op)).
% :- use_module(util(xml)).

% memo:  be careful to use sed/3   (module prefix needed) [2014/05/23]
%  Prolog-Emacs Interface
%  get buffer region
% :- use_module(pac('emacs-jockey')).

% ?- qcompile(util('prolog-elisp')), module(elisp).
:- nb_setval(command_history, [for_ascii(elisp:wrap($))]).
:- nb_setval(edit_command, (=)).
% :- use_module(library(edit)).

% help//2 is det.
%	list all command names.
help --> {shortcut_key(L)},
	 peek(L),
	 insert('\n'),
	 smash.

:- meta_predicate line(2,?,?).
:- meta_predicate for_ascii(2, ?, ?).

%% tex_cs_to_symbol(X:codes, Y:codes) is det.
%	Convert a tex control sequence codes to a symbol codes.
tex_cs_to_symbol --> sed(dcg_ts). % ts.

%% tex_symbol_to_cs(X:codes, Y:codes) is det.
%	Convert a symbol codes to a tex control sequence condes.
tex_symbol_to_cs --> sed(dcg_st). % st.

% %% read_terms(X:file, L:chars) is det.
% %	Read charater stream into a character list L.

% read_terms(X, L):- open_chars_stream(X, S), read_terms_(S, L).
% %
% read_terms_(S, []):- at_end_of_stream(S), !.
% read_terms_(S, [X|Xs]):-  read(S, X),  read_terms_(S, Xs).

preds --> clauses_preds.

%% clause_preds(X:stream, Y:list) is det.
%   Read  clauses  and show the heads of them.
clauses_preds -->  read_terms,
	maplist(clause_head),
	sort,
	insert("\n"),
	smash.

clause_head(X:- _, X):-!.
clause_head(X, X).

		/************************************
		*     command history management    *
		************************************/

%%	parse_commands(+X:codes, -Y:list) is det.
%	Parse lines as a list of Prolog queries.

parse_commands --> split,
	remove([]),
	maplist(herbrand).

%%	parse_merge_history(?X, -Y) is det.
%	Merge command history.

parse_merge_history --> region,
	parse_commands,
	merge_history,
	clear.
%%  merge_history(+Cs:list) is det.
%   Merge commands into the command history.

merge_history(Cs):- nb_getval(command_history, L),
	union(Cs, L, L0),
	nb_setval(command_history, L0).

%%  merge_history(+X:list, -Y:list) is det.
%   Merge commands in X into the command history.
%   Y = X.

merge_history(X, X):- merge_history(X).

%%  merge_history(F:phrase, ?X, ?Y) is det.
%   Parse region as commands, modify by F, and
%   merge into the command history.

merge_history(F) --> region,
	split,
	remove([]),
	maplist(phrase(F)),
	merge_history.

%%  show_history(-H:list) is det.
%   Unify H with the current history command.

show_history(H) :- nb_getval(command_history, L),
	insert('\n', L, L0),
	smashq(L0, H).

%%  show_history(?_, -H:list) is det.
%   Unify H with the current history command.

show_history(_, H) :- show_history(H).

%%  reset_history is det.
%   Reset the command history.
reset_history :- nb_setval(command_history, []).

%%  reset_history(?_, ?_) is det.
%   Reset the command history.
reset_history --> {reset_history}.

%%  set_history(?_, ?_) is det.
%   Reset the command history,  parse the region as commands,
%   and set the command history to it.
set_history --> reset_history, parse_merge_history.

%% show_command(-X:text) is det.
%  Unify X with the edit command.
show_command(X) :- nb_getval(edit_command, C),
	smashq(C, X).

%% show_command(_, -X:text) is det.
%  Unify X with the current command.
show_command(_, X):- show_command(X).

%% apply_command(X:term, Y:term) is det.
%  Unify X with the current comand history.
apply_command --> { nb_getval(edit_command, C) },
	phrase(C).

%
empty_history :- nb_getval(edit_command, []).

%
pop_history(X) :-  nb_getval(edit_command, [X|R]),
		nb_getval(edit_command, R).

%
push_history(X) :-  nb_getval(edit_command, C),
		    nb_setval(edit_command, [X|C]).

%
save_line_number :-  lisp('count-lines'(1, point()), I),
		     nb_setval(line_count, I).


% ?- apropos(compound_non_zero_arity).
% ?- compound_name_arguments(f(),  N, A).
%@ N = f,
%@ A = [].
% ?- compound_name_arguments(f,  N, A).

table_index(I) :-  lisp('count-lines'(1, point()), J),
		   nb_getval(line_count, K),
		   I is J-K.

%%	comma_left(+A, +B, -C) is det.
%	Unify C with a comma term (A, B).
comma_left(A, B, (A, B)).

%%	comma_right(+A, +B, -C) is det.
%	Unify C with a comma term  (B, A).
comma_right(A, B, (B, A)).

		/*******************************
		*     region object manager    *
		*******************************/

% [2011/06/09]
%  Save a phrase / a phrase with environment at  the region
%  to use it later.

%% initial_region_obj is det.
%	Initialize state_obj.
initial_region_obj :- nb_setval(state_obj, []).

%% region_obj(C:command) is det.
%	Update a state stored in state_obj by C.
region_obj(X):- nb_getval(state_obj, S),
	obj(X, S, S0),
	nb_setval(state_obj, S0).

%%	 get_phrase(?_, ?_) is det.
%	 Compile the goal at the buffer region,
%	 and save its handle to the state_obj.
get_phrase --> region,
	herbrand(_),
	call(pred([F, F0] :- let(F0, F))),
	current(G),
	{	region_obj( put( [command(G)] ) )  },
	clear.

%% apply_phrase(+X:term, -Y:term) is det.
%	Apply the saved phrase to  X to obtain Y, and
%	overwrite the region with Y.
apply_phrase --> region,
	{  region_obj( get( [command(G)] ) ) },
	phrase(G),
	overwrite.

%%  bind_phrase(+X:term, -Y:term) is det.
%   Apply the saved  bind_context command to X
%   to unify Y with the result.
bind_phrase --> region,
	{	region_obj( get( [command(G)] ) )	},
		current(X),
	{	bind_context(G, (X, []), (X0, _))	},
		peek(X0).

		/****************
		*     region    *
		****************/

%%	region(?A, -X) is det.
%	Unify Unify X with the content of the buffer region when
%	A == [], otherwise, with A.

region([], X) :- !, get_region(X).	% [2014/10/12], [2020/01/11]
region(A, A).

%%	get_region(-R:codes) is det.
%	Get the content of the buffer region, and
%	Unify R with it.
get_region(R) :-  lisp('send-region'()),
	read_codes(R).

%%	get_region(?_, -R:codes) is det.
%	Unify R with the content of the buffer region.
get_region  --> {get_region(R)}, peek(R).

%%	get_region(-X:codes, _, -X:codes) is det.
%	Unify X with the content of the buffer region.
get_region(X, _, X) :- get_region(X).		%

%%	region_term(-X:codes, _, -X:codes) is det.
%	Unify X with a term for the codes in the buffer region.
region_term(X, Y) :- region(X, X0),
					 term_codes(Y, X0).

%%	get_buffer_region(+X:int, Y:int, _, R:codes) is det.
%	Unify R with the content of the region boundary (X, Y).
get_buffer_region(X, Y) --> {get_buffer_region(X, Y, R)},
	peek(R).

%%	get_buffer_region(+X:int, Y:int,  R:codes) is det.
%	Unify R with the content of the region boundary (X, Y).
get_buffer_region(X, Y, R) :-
	lisp('send-region'(X, Y)),
	read_codes(R).

%%	region_bound(?_, X) is det.
%	Unify X with the boundary information of the region.
region_bound(_, X) :- region_bound(X).

%%	region_bound(X)  is det.
%	Shorthand for region_bound(_, X).
region_bound(X) :- call_lisp_exp('region-bound'(), X0),
	parse_lisp_form(X0, [X]).


%%	overwrite(+X:codes, ?_) is det.
%	Kill the region, and insert X to the buffer.
overwrite --> { kill_region },	smash_buffer.

%
overwrite([]) --> !, overwrite.
overwrite(_) --> [].

%%	kill_region is det.
%	Kill the region.
kill_region:-  call_lisp_wait('kill-region'(mark(), point())).

%%	kill_region(B:burrer_name) is det.
%	Kill the region of the buffer Buf.
kill_region(Buf):-  smash(Buf, QBuf),
	call_lisp_wait(progn(	'set-buffer'(QBuf),
			'kill-region'(mark(), point()))).

%% insert_buffer(+X:term) is det.
%	Insert X  into the current buffer.

insert_buffer([]):- !.
insert_buffer(X):- smash(["$TEXT$ ", X, "\000\\n"]),
	flush_output.
%
insert(X):- smash(X, Y), wait(insert(Y)).

%%	send_exp(L:script) is det.%	Send the script L to emacs-lisp process.
send_exp(L):- lisp(L).

%%	find_file(F:file) is det.
%	Send find_file command for F0 to emacs-lisp process.
find_file(F) :- expand_file_search_path(F, F0),
	lisp('find-file'(F0)).

%% find_pred(X:pred, Y, Y) is det.
%	Locate the predicate defiition of X.
find_pred(X)--> {find_pred(X)}.

%%	find_pred(+I) is det.
%	Locate the predicate indicator I = P/N.
find_pred(P/N):- functor(G, P, N),
	source_file(G, F),
	find_file(F).
find_pred(P):- find_pred(P/0).

%% eval_lisp_to_atom(+F:text, +S:text, -A:atom) is det.
%	Unify the value of evaluation of  S-expression (F S).
eval_lisp_to_atom(F, S, A):- lisp_atom(["(", F, " ", S, ")"], A).

%% eval_lisp_to_atom(+F:text, +X:text, +Y:text,  -A:atom) is det.
%	Unify the value of evaluation of  S-expression (F X Y).
eval_lisp_to_atom(F, X, Y, A):-
	lisp_atom(["(", F, " ", X, " ", Y, ")"], A).

%% atom_lisp_string(+A:atom, -S:string) is det.
%	convert the atom A to a lisp, and unify S with it.
atom_lisp_string(A, S):- atom_codes(A, C),
	quote(C, S).

%% read_file_name(+P:string, -F:string) is det.
%	Read a file name with prompt P,  and unify F with it.
read_file_name(P, F):- atom_lisp_string(P, P0),
	eval_lisp_to_atom('read-file-name', P0, F).

%% read_file_name(-F:string) is det.
%	Read a file name,  and unify F with it.
read_file_name(F):- read_file_name('file name: ', F).

%% completing_read(+P:string, +A:atoms, -R:term) is det.
%	Unify R with the result of completing-read with
%	prompt P and a candidate list A.
completing_read(Prompt, Atoms, R):-
	atom_lisp_string(Prompt, P0),
	maplist(atom_lisp_string, Atoms, As),
	call_lisp_text('completing-read'(P0, quote(As)), R).

% ?- elisp:read_codes_from_minibuffer("Hi :", R).
read_codes_from_minibuffer(Prompt,  Codes):-
	call_lisp_text('read-from-minibuffer'(Prompt), Codes).

% ?- elisp:read_term_from_minibuffer("Hi :", Term).
read_string_from_minibuffer(Prompt,  String):-
	read_codes_from_minibuffer(Prompt,  Codes),
	string_codes(String, Codes).

% ?- elisp:read_term_from_minibuffer("Hi :", Term, [variable_names(V)]).
read_term_from_minibuffer(Prompt,  Term, Options):-
	read_string_from_minibuffer(Prompt,  String),
	term_string(Term, String, Options).

% ?- elisp:read_term_from_minibuffer("Hi :", Term).
read_term_from_minibuffer(Prompt,  Term):-
	read_term_from_minibuffer(Prompt, Term, []).

%%  search_pred(P:pred) is det
%	Search predicate P.
search_pred(Pred):- search_swipl(["^", Pred]).

%% search_swipl(X:text) is det.
%	Search X in dired mode.
search_swipl(X):- smash(X, C),
	dired_mark_swipl,
	lisp('dired-do-search'(C)).

%% dired_mark_swipl is det.
%	Mark Prolog files in the pacpl7 directory in dired mode.

dired_mark_swipl :- getenv(home, Home),
    string_concat(Home,"/local/lib/pacpl7", PAC_dir),
	dired_mark_regex(PAC_dir, ".*\\.pl$").

%% dired_mark_regex(+Dir:string, +Regex:string) is det.
%	visit the directory Dir in Emacs dired mode,
%	and mark all files  whose name matches the
%	regex Regex.

dired_mark_regex(Dir, Regex) :-
	lisp([progn, 'find-file'(Dir),
		     'dired-mark-files-regexp'(Regex)]).
%	True if J, Len, and X are unified so that X is a symbol of length Len,
%	of the category "prolog_symbol_continue", and  at J of the current line,
list_at_point(X):- line_get(Obj),
	obj_get([line(Line), begin(B), point(P)], Obj, _),
	string_codes(Y, Line),
	I is P - B,
	scan_string_at_point(Y, 0'[, 0'], I, S),
	term_string(X, S).

%%	neighbor_string(+X:string, +B:string, +E:string, +I:integer, -S:string) is det.
%
%	True if S is the minimum substring of X such that sub_string(X, J, K, _, S),
%		J + length(B)   =<  I  =<  K - length(E),
%	and that S has B as a prefix, and E as a suffix.

% ?- elisp:neighbor_string("abd", "a", "d", 1, S).
% ?- elisp:neighbor_string("abd", "a", "bd", 1, S).

neighbor_string(X, Open, Close, I, S):- neighbor_string(X, Open, Close, I, J, K),
				   N is K - J,
				   N >= 0,
				   sub_string(X, J,  N, _, S).

% ?- elisp:neighbor_string("abcd", "a", "d", 2, J, K).
% ?- elisp:neighbor_string("abd", "a", "d", 2, J, K).
%
neighbor_string(X, Open, Close, I, J, K):- string_length(X, LX),
	string_length(Open, LO),
	string_length(Close, LC),
	I0 is I - LO,
	I0 >= 0,
	neighbor_string_left(X, Open, LO, I0, J),
	neighbor_string_right(X, Close, LC, I, K0, LX),
	K is K0 + LC.

%
neighbor_string_left(X, S, N, I, I) :- sub_string(X, I, N, _, S), !.
neighbor_string_left(X, S, N, I, K) :- I> 0,
				  I0 is I-1,
				  neighbor_string_left(X, S, N, I0, K).

neighbor_string_right(X, S, N, I, I, _):-  sub_string(X, I, N, _, S), !.
neighbor_string_right(X, S, N, I, K, L):- I  < L,
				     I0 is I + 1,
				     neighbor_string_right(X, S, N, I0, K, L).

%
neighbor_string(B, E, X) :- line_at_point(L, I),
			  neighbor_string(L, B, E, I, X).


%%	line_at_point(-X:string, -I:int, P:int) is det.
%	True  if X is  the content of the current line,
%	I is the relative position of the cursor.

		/*****************************
		*     Control Emacs-buffer   *
		*****************************/

%% stream_start is det.
%	Send 'start prolog process' to emacs.
stream_start :- wait('stream-start'()).

%% wait(P:list_command, P) is det.
%	Send a lisp Command P, and wait a response.
wait(P, P) :- wait(P).

% ?- elisp:message("hello").
message(M):-  smash(M, M0), lisp(message(M0)).

%%	prolog(F:phrase) is det.
%	Apply F to the buffer region.
:- meta_predicate prolog(:, ?, ?).
prolog(F) --> region, herbrand, phrase(F).

%% line_exec is det.
%	Get the current line, and exec it as a prolog query.
%
line_exec :- line_get(I),
	obj_get([line(Line)], I, _),
	herbrand(_, Line, G),
	!,
	catch(G, _, fail).

%% smash(F:phrase, X, Y) is det.
%	apply F to the current region, and smash it to Y.
smash(F) --> prolog((F, smash)).

%% herb(F:phrase, X, Y) is det.
%	apply F to X,  smash,  convert to a term, and
%	unify Y with it.
herb(F) --> prolog((F, herbrand_opp)).

%% solve(F:phrase, X, Y) is det.
%	Convert X to a term with prolog variables, and apply F to X,
%	convert the result to a term, and unify Y with it.
solve(F) --> herbrand(_),
	phrase(F),
	herbrand_opp.

%%	line(F:phrase, X, Y).
%	Split the region into lines, and apply F to each of them.
line(F) --> line(F, "\n").

%%	line(F:phrase, I:term, X, Y).
%	Split the region into lines, apply F to each of them,
%	and insert I among lines.
line(A, Ins) --> split,
	remove([]),
	maplist(phrase(A)),
	insert(Ins).

%%	paragraph(F:phrase, +X:codes, -Y).
%	Split the region X into paragraphs, and apply F to each of
%	them, and unify Y with the result.

paragraph(A) --> paragraph,
	remove([]),
	maplist(phrase(A)).

%%	lines(X:codes, Y:list) is det.
%	Split the region into a list of lines,  and
%	unify Y with it.
lines --> split, remove([]).


%%	line_edit(+Ps:obj) is det.
%	Assuming Ps = [line_get(G), scan(S), do(D), line_put(P)],
%       the current line L = X + A + Y ('+' means codes concatenaton)
%	is rewritten to X + A0 + Y,	by
%	applying  G to get the line L,
%	applying  S to L to get  X, A, and Y, as above, and
%	applying  D to A to get A0.

line_edit(Ps):- obj_get([line_get(G), scan(S), do(D), line_put(P)], Ps),
	call(G, I),
	obj_get([line(L)], I),
	call(S, X, A, Y, L, _),
	% act(D, A, A0),
	phrase(D, A, A0),
	obj_put([left([X, A0]), right(Y)], I, I0),
	call(P, I0).


%%	for_ascii(Act:phrase, X, Y) is det.
%	Apply Act to an ascii codes block in the the currrent line
%	and rewrite the block with the result.
for_ascii(Act) --> { line_edit([line_get(line_get_forward),
				scan(find_next_ascii_codes),
				do(Act),
				line_put(line_overwrite)
			       ])
		   }.

%%	open_at_point is det.
%	Open the file whose name is at the cursor.
%[2013/09/08]
open_at_point:-  line_get(Obj),
	obj_get([line(File)], Obj),
	sh(open(File)).

%%	open_at_point(X, X) is det.
%	Open the file whose name is at the cursor.
open_at_point --> {open_at_point}.

%%	line_at_point(-X:string, -I:int, P:int) is det.
%	True  if X is  the content of the current line,
%	P is the starting point, I is the relative position of
%	the cursor.
line_at_point(X, I, P):- line_get(Obj),
	obj_get([point(Q), line(Line), begin(P)], Obj),
	string_codes(X, Line),
	I is Q - P.

%%	line_at_point(-X:string, -I:int, P:int) is det.
%	True  if X is  the content of the current line,
%	I is the relative position of the cursor.
line_at_point(X, I):- line_at_point(X, I, _).

		/******************************
		*     basic line operation    *
		******************************/

%%	line_get(-X:obj) is det
%	Unify X with [point(P), begin(B), end(E), line(R)], where
%	P is the current point,  B  the beginning point, E the end point
%	R the content of the line.
line_get([point(P), begin(B), end(E), line(R)]):-
	lisp(list('line-beginning-position'(),
			   'line-end-position'(),
			   point()),
			[B,E,P]),
	get_buffer_region(B, E, R).

%%	line_get_forward(-X:obj) is det
%
%	Unify X with [point(P), begin(P), end(E), line(R)], where
%	P is the current point, E the end point,
%	R the content of the line from P to E.
line_get_forward([point(P), begin(P), end(E), line(R)]):-
	lisp(list(point(), 'line-end-position'()),  [P, E]),
	get_buffer_region(P, E, R).

%%	line_get_backward(-X:obj) is det
%	Unify X with [point(P), begin(B), end(P), line(R)], where
%	P is the current point,  B  the beginning point,
%	R the content of the line from B to P.
line_get_backward([point(P), begin(B), end(P), line(R)]):-
	lisp(list(point(), 'line-beginning-position'()),
			   [P, B]),
	get_buffer_region(B, P, R).

%%	line_overwrite(+I:obj) is det.
%	Provided that L and R are the text in the line informaton I,
%	replace the line I with L plus R, making the point after L
%	a new current point.
line_overwrite(I):- obj_get([left(L), right(R)], I),
	line_overwrite(I, L, R).

%%	line_overwrite(+I:obj, L:text) is det.
%	Replace the line with L.
line_overwrite(I, L) :- line_kill(I), smash_buffer(L).

%%	line_overwrite(+I:obj, +L0:text, +L1:text) is det.
%	Replace the line I with L0 plus L1. The point after L0
%	is the current point.
line_overwrite(I, L0, L1) :- line_kill(I),
	smash_buffer(L0),
	lisp(point(), P),
	smash_buffer(L1),
	lisp( 'goto-char'(P)).

%%	smash_buffer(+X:list-as-quasi-text) is det.
%	True if all elements in X are written
%	to the standard output in order.
smash_buffer(X) :- smash(X, Y), elisp:insert_buffer(Y).

%%	smash_buffer(+X:text, -Y:list) is det.
%	True if all texts in X are written
%	to the standard output in order, and Y is unfied with [].
smash_buffer(X, []) :- smash_buffer(X).


%%	line_kill(+I:obj) is det.
%	Kill the region I,
line_kill(I) :- obj_get([begin(B), end(E)],  I),
	lisp('kill_region'(B, E)),
	lisp('goto-char'(B)).

%%	line_kill(X, X) is det.
%	Kill the current line.
line_kill --> {line_get(I), line_kill(I)}.

%%	line_wrap(+L:text, +R:text, X, X) is det.
%	Put L and R before and after, respectively, the current line.
line_wrap(L, R) --> {line_wrap(L, R)}.

%%	line_wrap(+L:text, +R:text) is det.
%	Put L and R before and after, respectively, the current line.
line_wrap(L, R):- line_get(I),
	obj_get([line(Line)], I),
	line_overwrite(I, [L, Line, R]).

%%	prolog_symbol_at_point(-X:string) is det.
%	True if X is unified with a symbol at the current line.

%?- sub_string("abc", X, Y, Z, U).
% [2014/03/16]
% ?-  elisp:prolog_symbol_at_point(X).
prolog_symbol_at_point(X):- line_at_point(Y, I),
	symbol_at_index(Y, I, X).

%%	prolog_symbol_at_point(-J:int, -Len:int, -X:string) is det.
%	True if J, Len, and X are unified so that X is a symbol of length Len,
%	of the category "prolog_symbol_continue", and  at J of the current line,
prolog_symbol_at_point(J, Len, X):- line_get(Obj),
	obj_get([line(Line), begin(B), point(P)], Obj, _),
	string_codes(Y, Line),
	I is P - B,
	symbol_at_index(Y, I, J, K, X, prolog_symbol_continue_chk),
	Len is K - J.

%%	apply_symbol_at_point(+F:pred/2) is det.
%	Convert and replace the current prolog symbol by applying F.
:- meta_predicate apply_symbol_at_point(2).
apply_symbol_at_point(F):- prolog_symbol_at_point(J, Len, X),
	call(F, X, Y),
	number_string(J, J0),
	number_string(Len, Len0),
	with_output_to(string(Y0), writeq(Y)),
	lisp('replace-in-line'(J0, Len0, Y0)).

%%	prolog_symbol_at_point(X:string, -I:Int, -Y:string) is det.
%	True if Y is unified with a prolog symbol as a substring of X at I.
symbol_at_index(X, I, Y):- symbol_at_index(X, I, Y, prolog_symbol_continue_chk).

%%	symbol_at_index(+X:string, +I:int, -Y:string, +F:pred/1) is det.
%	True if Y is unified with the longest substring of X at I so that
%	every code of Y satisfies F.
symbol_at_index(X, I, Y, F):- symbol_at_index(X, I, _, _, Y, F).

%%	symbol_at_index(+X:string, +I:int, -J:int, -K:int, -Y:string, +F:pred/1) is det.
%	True if Y, J, and K are unified so that every code of X
%	between point J and K satisfies F, and
%	Y is unified with the substring between J and K.

% ?- elisp:symbol_at_index("+abc-", 2, J, K, Y, prolog_symbol_continue_chk).
:- meta_predicate symbol_at_index(?,?,?,?,?,1).
symbol_at_index(X, I, J, K, Y, F):- symbol_at_index(X, I, J, K, F),
	L is K - J,
	sub_string(X, J, L, _, Y).

%%	symbol_at_index(+X:string, +I:int, -J:int, -K:int, +F:pred/1) is det.
%	True if J and K are unified so that there is a longest substring of X at I
%	that starts at J and ends at K, and that every code of the substring
%	between point J and K satisfies F.
:- meta_predicate symbol_at_index(?,?,?,?,1).
symbol_at_index(X, I, J, K, F):-
	symbol_string_before(X, I, J, F),
	symbol_string_after(X, I, K, F).

%%	prolog_symbol_continue_chk(C:code) is det.
%	True if C is of code_type "prolog_identifier_continue".
prolog_symbol_continue_chk(C):- code_type(C, prolog_identifier_continue).

%%	symbol_string_after(+X:string, +I:int, -K:int, +F:pred/1) is det.
%	True if K is unified so that there is a longest substring of X
%	that starts at I and ends at K, and that every code of the substring
%	between point I and K satisfies F.
:- meta_predicate symbol_string_after(?,?,?,1).
symbol_string_after(S, I, K, F):-  succ(I, J),
	string_code(J, S, C),
	call(F, C),
	!,
	symbol_string_after(S, J, K, F).
symbol_string_after(_, I, I, _).

%%	symbol_string_after(+X:string, +I:int, -K:int, +F:pred/1) is det.
%	True if K is unified so that there is a longest substring of X
%	that starts at K and ends at I, and that every code
%	between point K and I on the substring satisfies F.
:- meta_predicate symbol_string_before(?,?,?,1).
symbol_string_before(S, I, K, F):-
	string_code(I, S, C),
	call(F, C),
	!,
	succ(J, I),
	symbol_string_before(S, J, K, F).
symbol_string_before(_, I, I, _).

		/***********************************************
		*	   lisp list vs. prolog list         *
		***********************************************/

%%	term_to_lisp_text(?X:term, ?Y:string) is det.
%	When Y is instantiated, unify X with a prolog term so that X is a translation of a S-exression
%	in the string Y. Otherwise, unify Y with a string so that X is a prolog term translation of
%	Y as an S-expression in string.

% ?- elisp:term_to_lisp_text(X, "(ab)"), elisp:term_to_lisp_text(X, Y).
%@ X = [[ab]],
term_to_lisp_text(X, Y):- nonvar(Y), !,
	( string(Y)
	->	string_codes(Y, Y0)
	;	 Y0 = Y
	),
	parse_lisp_form(Y0, X).
term_to_lisp_text(X, Y):- once(term_to_lisp_string(X, Y)).

%%	parse_lisp_form(+X:codes, -Y:list) is det.
%	True if Y is unified with a list translation of X as
%	an S-expression in string.

parse_lisp_form(X, Y):- once(lisp_list(Y, X, [])).

%%	lisp_list(-X:codes, +Y:list, -Z:list) is det.
%	True if X is unified with a list which is a translation of
%	a prefix of Y as an S-expression in string, and Z is unified with
%	the remainder.

% ?- elisp:lisp_list(X, `()`, []).
% ?- elisp:lisp_list(X, `(a)`, []).
% ?- elisp:lisp_list(X, `(ab)`, []).
% ?- elisp:lisp_list(X, `(a b)`, []).
% ?- elisp:lisp_list(X, `(a(c)b"de")`, []).
% ?- elisp:lisp_list(X, `(a(c)()b"de")`, []).
% ?- elisp:lisp_list(X, `(a(c)()b"de""")`, []).
% ?- elisp:lisp_list(X, `\\((a(c)()b"de""")`, []).
% ?- elisp:lisp_list(X, `\\(a)\\`, []).

lisp_list(X) --> skip_lisp_filler, lisp_list(X, []).

%%	lisp_list(-X:list, -X0:list, +Y:codes, -Z:codes) is det.
%	True if X, X0, and Z are unified with lists so that
%	the difference between X and X0  is a prolog list translation of
%	that of Y and Z as S-expressions in string codes.

lisp_list(X, Y) --> lisp_element(A),
	{X=[A|X0]},
	skip_lisp_filler,
	lisp_list(X0, Y).
lisp_list(X, X) --> [].

%%	lisp_list_rest(-X:list, -Y:list, +Z:codes, -U:codes) is det.
%	Unfify X, Y, U, and Z so that the difference between X and Y
%	is a prolog list translation of the difference between Z and U as a tail
%	of a lisp list in string codes.

lisp_list_rest(X, X)     --> ")".
lisp_list_rest([A|X], Y) --> lisp_element(A),
	skip_lisp_filler,
	lisp_list_rest(X, Y).

%%	lisp_element(A:term, +X:codes, -Y:codes) is det.
%	Unify A and Y with a list so that A is an element translation of the difference
%	between X and Y as an S-expression in  string codes.

lisp_element(A) --> "(",  !,  skip_lisp_filler,
	lisp_list_rest(A, []).
lisp_element(A) --> lisp_string(A0), !, { herbrand(A0, A) }.
lisp_element(A) --> non_string_atom(A0, []), { A0 \== [] }, !,
	{ herbrand(A0, A) }.

%%	skip_lisp_filler(+X:codes, -Y:codes) is det.
%	Skip longest filler block prefix of Y,  and unify Y with
%	the remainder.

skip_lisp_filler --> [A], { memberchk(A, `\s\t\r\n`) },
	skip_lisp_filler.
skip_lisp_filler --> [].

%%	lisp_string(-X:codes, +Y:codes, -Z:codes) is det.
%	True if X is unified a lisp string codes at a prefix of Y,
%	and Z the remainder.
lisp_string([0'\"|X]) --> "\"", lisp_string_rest(X, []).  % '

%%	lisp_string_rest(-X:codes, X0:codes,  +Y:codes, -Z:codes) is det.
%	True if X and X0 are unified so that the difference between X and X0
%	is a lisp string codes at a prefix of Y,  and then Z with the remainder.

% ?- elisp:lisp_string_rest(X, [], `abcd"`, Y).
lisp_string_rest([0'\"|X], X)	--> "\"".				% '
lisp_string_rest([0'\\, X|Y], Z) --> "\\", [X], !,		% '
	lisp_string_rest(Y, Z).
lisp_string_rest([X|Y], Z)	--> [X], lisp_string_rest(Y, Z).

%%	non_string_atom(-X:codes, -Y:codes, +Z:codes, -U:codes) is det.
%	Unify X, Y, and U so that the difference of X and Y is a
%	non string atom translation of the difference between Z and U.
non_string_atom([X|Y], Z) -->[X], {\+ memberchk(X, `()\"\s\t\r\n`)},
	non_string_atom(Y, Z).
non_string_atom(X, X) --> [].

%%	quit(X, X) is det.
%	Send to quit the "(start-process)" process.
quit --> {lisp(print("PAC process has been quitted by user")),
	  lisp('stop-pac'())
	  }.

%%	edit_emacs_P(?X, ?Y) is det.

%	Ask emacs an edit command in codes,
%	If the codes form a prolog phrase H,
%	run H on (X, Y).
%	Otherwise, split the codes by spaces into a list H of atoms,
%	find an handle that H matches its argument, and
%	finally run the body of the handle as a phrase on (X, Y).

edit_emacs_P --> {ask_command(Codes),
		catch(	herbrand(_, Codes, H),
			_ ,
			split_codes_atoms(Codes, H)
		     )
	       },
		run_command(H).

%%	run_command(+A:term, ?X:codes, ?Y:codes) is det.
%	Findd a handle that has A as its argument when A is a list,
%	that has [X]   when A = -(X), respectively, and
%	apply the body phrase of  the handle to (X, Y).
%       Otherwise, apply A itself as a phrase to (X, Y).

run_command([X|Y]) --> !, ejockey:handle([X|Y]).
run_command(-(X)) --> !, ejockey:handle([X]).
run_command(X) --> phrase(X).

%%	split_codes_atoms(+X:codes, -Y:list) is det.
%	Split X with spaces into a list X0 of blocks of codes,
%	ane unify Y with a list of atoms for the blocks in X0.

split_codes_atoms(X, Y):- split(` `, X, S),
	remove([], S, S0),
	maplist(atom_codes, Y, S0).

%
skip_rest(X, X, []).

% ?- elisp:leading_atoms(`   aab bbb ccc & ddd eee`, A).
% ?- elisp:leading_atoms([], X).

leading_atoms(X, A):- tex:filler(X, Y),
					  once(leading_atoms(A, Y, [])).

% ?- elisp:leading_atoms(X, [], R).
% ?- elisp:leading_atoms(X, ` aab bbb ccc & ddd eee`, R).
% ?- elisp:leading_atoms(X, `aab bbb ccc  ddd eee`, R).
% ?- elisp:handle_atom(X, `a`, R).

leading_atoms(X) --> "&", skip_rest(X).

leading_atoms([A|Y]) --> handle_atom(X), !,
		{ atom_codes(A, X) },
		tex:filler,
		leading_atoms(Y).
leading_atoms([]) --> at_end.

at_end([], []).

% ?- elisp:handle_atom(X, `abc def`, R).
handle_atom(X)  --> w("[a-z]", X, Y), wl("[a-zA-Z0-9]*", Y).
handle_atom(X)	--> w("[*/=?#.<>\\^\\-\\!]", X).

% ?- elisp:make_splitter(`[ab]+`, S), call(S, `abcab`, R).
make_splitter(Codes, S) :- string_codes(Regex, Codes),
						   let(S, phrase(wl(Regex))).

%%	edit_emacs_L(+A:pred, ?X, ?Y) is det.
%	Ask emacs for an edit command in codes by calling A
%	Split the codes by spaces into a list H of atoms,
%	find an handle that H matches its argument, and
%	finally run the body of the handle as a phrase on (X, Y).
%	If no handle is found, the codes is parsed as a phrase
%	and the phrase is applied to (X, Y).

edit_emacs_L(Ask, X, Y):-
	call(Ask, Codes),
	leading_atoms(Codes, Com),
	once(find_handle_call(Com, Codes, X, Y)).

% ?- edit_emacs_L('ask-handle'(), Codes).
% ?- elisp:find_handle_call([split, a, b], _,  X, Y).

%%	find_handle_call(H:list, C:codes, ?X, ?Y) is det.
%
%	Find an handle such that H matches the argument of the head of the
%	handle, and let P be the body phrase of the handle.
%	In case no handle being found, let P be the phrase term
%	whose codes is C. Then, in any case, apply the phrase P to [X, Y].

% ?- elisp:find_handle_call([f, c], `a`, Y), smash(Y).

find_handle_call(Com, _Codes, X, Y):- is_list(Com),
	member(Module,   [ejockey,  ejockey2]),
	clause(Module:handle(H, X, Y), Body),
	match_args(Com, H),
	!,
	once(Module:Body).
%  [2020/05/14] The following does not work.
%	Module:call(Body), !.
%	Module:onde(Body), !.
find_handle_call(_, Codes, X, Y):-  % for other type of handles
	herbrand(_, Codes, C),
	callable(C),
	C =.. A,
	append(A, [X, Y], B),
	D =.. B,
	(	clause(snippets:D, Body)
	->	G = snippets:Body
	;	G = phrase(C, X, Y)
	),
	once(G).  % was	"call(G), !."

%%	match_args(+X:list_of_atoms, +Y:list) is det.
%
%	True if for each pair of elements A and B in order in X and Y,
%	respectively, the following condition is true:
%	If B is a variable then B is unified with A.
%	If B is an atom then A is a prefix of B.
%	If B is a number then B is the atom_number of A,
%	If B is a string then B is the atom_string of A.
%
%	If the tail Z of Y=[...|Z] is a variable the Z is unified with
%	the corresponding tail U of X=[...|U].

% ?- elisp:match_args([f, c], [frame, center]).
% ?- elisp:match_args([sol, go, reg], [solve, goal|X]).
% ?- elisp:match_args([mv, (.)], [mv, (.)]).
% ?- elisp:match_args([f, 'c/d'], [frame, File]).
% ?- elisp:match_args([f, b], [frame, File]).
% ?- elisp:match_args([f, b], [frame|File]).

match_args(A, B):- var(B), !, A = B.
match_args([A|B], [C|D]):- once(match_arg_one(A, C)), !,
						   match_args(B, D).
match_args([], []).

%
match_arg_one(A, B) :- var(B), A = B.
match_arg_one(A, B) :- atom(B), sub_atom(B, 0, _, _, A).
match_arg_one(A, B) :- number(B), number_atom(B, A).
match_arg_one(A, B) :- string(B), atom_string(A, B).

%
normal_keys([], _, []).
normal_keys([X|Xs], Ys, [Y|Zs]):- select(Y, Ys, Y0s),
	match_arg_one(X, Y),
	!,
	normal_keys(Xs, Y0s, Zs).

% ?- elisp:apropos_chk([a,b,c], axy).
apropos_chk([Y|_], X):- match_arg_one(Y, X), !.
apropos_chk([_|Ys], X):- apropos_chk(Ys, X).

%%	ask_handle(-C:codes) is det.
%	Ask emacs for a handle codes, and unify C with it.
ask_handle(Codes) :- call_lisp_text('ask-handle'(), Codes).

% [2020/05/14]
% request from emacs as (prolog-query G).
%	shortcut "s-l" (Command + L )
user:prolog_call(G, _, [G, " done.\n"]):- once(G).

% ?- elisp:call_lisp_text(list("1","2","3"),  R).

%%	ask_command(-C:codes) is det.
%	Ask emacs for a command codes, and unify C with it.
ask_command(Codes) :- call_lisp_text('ask-command'(), Codes).

%%	lisp_to_list(+X:term, -Y:list) is det.
%	Ask emacs to eval an meta-exression X,
%	and unify Y with a list returned from emacs.

% ?- elisp:lisp_to_list(+(1, 2, 3), Y).
% ?- elisp:lisp_to_list(sort(list(4, 8, 21, 17, 33,  7,  21, 7), quote(quote(>))), V).
% ?- elisp:lisp_to_list(sort(list(4, 8, 21, 17, 33,  7,  21, 7), quote(quote(<))), V).

lisp_to_list(M, L):- call_lisp(M, [value(V), string(t)]),
	term_to_lisp_text(L, V).

% ?- elisp:lisp(+(1, 2, 3), Y).
% ?- elisp:lisp(sort(quote([4, 8, 21, 17, 33,  7,  21, 7]), quote(>)), V).

%
lisp(X, Y) :- lisp_to_list(X, Y0), Y0 = [Y|_].

%%	lisp_atom(+X:codes, -Y:atom) is det.
%	Ask emacs to eval an S-exression codes  X,
%	and unify Y with an atom for the returned value text codes
%	from emacs.

lisp_atom(X, Y) :- call_lisp(X, [value(Z)]),  atom_codes(Y, Z).

%%	no_reply_progn(E:codes) is det.
%	Ask emacs to eval in progn mode for E and wait for
%	"done."

no_reply_progn(E) :- wait(E).

%%	unquote(+X:codes, -Y:codes) is det.
%	Remove double quote codes if exists at both ends of X,
%	and unify Y with the body of the codes.
%	Otherwise, unify Y with X.

unquote(X, Y):- append([[0'\"], Y0, [0'\"]], X), !,
	unbackquote(Y0, Y).
unquote(X, X).

%%	unbackquote(+X:codes, -Y:codes) is det.
%	Remove backquote codes used for escape character from X,
%	and unify Y with the remainder.

unbackquote([], []).
unbackquote([0'\\, 0'\\|X], [0'\\|Y]):-	!, unbackquote(X, Y). %'
unbackquote([0'\\, C|X], [0'\\, C|Y]):-	!, unbackquote(X, Y).
unbackquote([C|X], [C|Y]):- unbackquote(X, Y).

%%	list(?X:list) is nondet.
%	Unify X with most general lists
%	in the 'shortest one first' order.

list([]).
list([_|L]) :- list(L).

%%	escapeoff(+X:codes, -Y:codes) is det.
%	Remove all escape character code from X,
%	and unify Y with the remainder.

escapeoff --> eh:sed(elisp:escapeoff), flatten.

escapeoff([X])--> "\\",[X].


		/********************
		*     read_codes    *
		********************/

%
% ?- read_codes(X, `yz`).
% |:abcyz

read_codes(X, End):- get_code(C), C \== (-1), !,
	read_codes(C, End, End, [], X, []).
read_codes(_, _):- throw(unexpected_end_of_file(read_codes)).

%
read_codes(X):- read_codes(X, `\000\\n`).

%
read_codes(C, [C|R], End, Q, X, Y):- !, read_codes(R, End, [C|Q], X, Y).
read_codes(C, _, End, [], [C|X], Y):- !, read_codes(End, End, [], X, Y).
read_codes(C, _, End, Q, X, Y):- reverse(Q, Z, X),
	read_codes(C, End, End, [], Z, Y).

% [2016/07/11]
% Knowing end-of-message by use of read_pending_codes/3.
% ?- elisp:lisp("\000\\n\n", R).

read_codes([], End, Q, X, Y):- !,
	read_pending_codes(user_input, P, []),
	( P==[] -> X = Y
	;	reverse(Q, Z, X),
		read_codes_suspended(P, End, Z, Y)
	).
read_codes(R, End, Q, X, Y):- get_code(C),
	read_codes(C, R, End, Q, X, Y).

%
read_codes_suspended(S, End, X, Y):-
	( append(A, End, S) -> append(A, Y, X)
	 ; append(S, Z, X),
	   read_pending_codes(user_input, P, []),
	   read_codes_suspended(P, End, Z, Y)
	).

reverse([A|B], X, Y):- reverse(B, [A|X], Y).
reverse([], X, X).

		/***************************************************
		*           call_lisp/read_mini_buffer             *
		***************************************************/

%%	read_term_from_lisp(-X:term) is det.
%	Read a term converted from emacs-lisp text with nullstop.

read_term_from_lisp(X) :-
	read_codes(Codes),
	term_codes(X, Codes).


%%	call_lisp(+X:term, ?O:options) is det.
%	Ask lisp to eval X as an S expression, with options
%	value(V)	; V is unified with a return value codes
%	string(t)	: require lisp to use 'prin1_to_string'
%	string(nil)	: tell lisp not to use 'prin1_to_string'
%	wait		: require lisp to send 'done' when done.
%	lisp		: tell lisp not to reply.


% ?- elisp:call_lisp([], [value(V), string(t)]).
call_lisp(X, Options):-
	(	memberchk(noreply, Options)
	->	term_to_lisp_string(X, X0),		% just send without waiting.
		send_off(X0)
	;	memberchk(wait, Options)
	-> 	P = progn(X, 'send-done'()),
		term_to_lisp_string(P, X0),
		send_off(X0),
		read_codes(_)
	;	memberchk(value(V), Options),
		( memberchk(string(F), Options);  F = nil )
	->	term_to_lisp_string(respond(X, F), X0),
		send_off(X0),
		read_codes(V)
	;	call_lisp(X, [wait])
	).

% For use as handle([lisp |Options]).
handle_lisp(X, Keys, Out):-
	FullKeySet = [paragraph, region, noreply, wait, value, string],
	normal_keys(Keys, FullKeySet, Options),
	(	memberchk(noreply, Options)
	->	send_off(X),
		Out=[]
	;	memberchk(wait, Options)
	-> 	send_off(["(progn ", X, " (send-done))"]),
		read_codes(_),
		Out=[]
	;	memberchk(value, Options)
	->	(	memberchk(string, Options) ->  F = t
		;	F = nil
		),
		send_off(["(respond ", X, " ", F, ")"]),
		read_codes(V),
		Out=[V]
	;	handle_lisp(X, [wait], Out)
	).

%
send_off(X):- smash(["$SCRIPT$ ", X, "\000\\n"]),
			 flush_output.

%%	call_lisp(+X:term) is det.
%	Shorthand for call_lisp(X, [wait]).

call_lisp(X):- call_lisp(X, [wait]).

%%	lisp(X) is det.
%	Shorthand for call_lisp(X, [noreply]).
lisp(X):-  call_lisp(X, [noreply]).
%
noreply(X):- call_lisp(X, [noreply]).

%%	wait(X) is det.
%	Shorthand for call_lisp_wait(X).

wait(X) :- call_lisp_wait(X).

%%	call_lisp_wait(X) is det.
%	Shorthand for call_lisp(X, [wait]).

call_lisp_wait(X) :- call_lisp(X, [wait]).

%%	call_lisp_text(X, Y) is det.
%	Shorthand for call_lisp(X, [value(Y)]).

call_lisp_text(X, Y) :- call_lisp(X, [value(Y)]).
%%	call_lisp_exp(X, Y) is det.
%	Shorthand for call_lisp(X, [value(Y), string(t)]).

call_lisp_exp(X, Y) :- call_lisp(X, [value(Y), string(t)]).

% ?- elisp:lisp(sort(list(4, 8, 21, 17, 33,  7,  21, 7), quote(quote(>))), V).

% ?- elisp:lisp(+(1, 2, 3), Y).
% ?- elisp:lisp("(+, 1, 2)", Y).
% ?- elisp:lisp("Abc", Y).

%?- elisp:term_to_lisp_string(append(cons(x,y), z, cons(u,z)), R).
%@ R = "(append (cons x y )z (cons u z ))".

%%	term_to_lisp_string(+X:term  -Y:string) is det.
%	Unify Y with a string for X as an S-expression.

% ?- elisp:term_to_lisp_string(f(), X).
term_to_lisp_string(X, Y):- once(flatten_s_exp(X, Y0, [])),
	once(flatten_to_lisp_form(Y0, Y)).

%%	flatten_s_exp(+X:term, -Y:list, -Z:list) is det.
%	Convert the term X  into an S-expression so that
%	the diffirence between Y and Z is a flat list of tokens
%	for the S-expression.

% ?- elisp:flatten_s_exp(f(a), R, []).
flatten_s_exp(X, ['('|Y], Z):- is_list(X), !,
	flatten_s_exp_list(X, Y, Z).
flatten_s_exp(X, [X0|Y], Y):-
	(	atom(X),  atom_string(X, X0)
	;	atomic(X), term_string(X, X0)
	).
flatten_s_exp(X, Y, Z) :-  compound_name_arguments(X, F, Args),
	flatten_s_exp([F|Args], Y, Z).

%%	flatten_s_exp_list(+X:list, -Y:list, -Z:list) is det.
%	Convert each element of X  into an S-expression so that
%	the diffirence between Y and Z is a flat list of tokens
%	for the list of S-expression.

flatten_s_exp_list([A|B],  X, Y):-  flatten_s_exp(A, X, Z),
	flatten_s_exp_list(B, Z, Y).
flatten_s_exp_list([], [')'|Y], Y).

%%	flatten_to_lisp_form(+X:list, -Y:list) is det.
%	Convert a list X of tokens into a string Y by concatenation
%	so that Y is a string form of an S-expression that X represents

% ?- elisp:flatten_to_lisp_form([a], X).
%@ X = "a "
% ?- elisp:flatten_to_lisp_form(['(', '(', a, b, ')', ')'], X).
flatten_to_lisp_form(X, Y):-
	flatten_to_lisp_form(X, Y0, []),
	atomics_to_string(Y0, Y).

%
flatten_to_lisp_form([], X, X).
flatten_to_lisp_form([T|R], X, Y):- flatten_to_lisp_form(T, X, Z),
	flatten_to_lisp_form(R, Z, Y).
flatten_to_lisp_form('(', ["("|X], X).
flatten_to_lisp_form(')', [")"|X], X).
flatten_to_lisp_form(A,   [A, " "|X], X).

%%	read_mini_buffer(+P:sring, -S:string) is det
%	Read a string from the emacs mini-buffer.

% ?- elisp:read_mini_buffer("hello", X).

read_mini_buffer(Prompt, Str):-
	call_lisp('read-string'(Prompt), [value(StrCodes)]),
	string_codes(Str, StrCodes).

%%	read_number(Max:int, X:number) is det
%	Read a number from the emacs mini-buffer.
read_number(Max,  X):- read_mini_buffer("number : ", S),
	number_string(X, S),
	between(1, Max, X).

		/********************************
		*     set/get string globally   *
		********************************/

%%	setq(+X:atom, +Y:term) is det.
%	Shorthand for lisp(setq(X, Y)).
% ?- elisp:setq(working_directory,  "/Users/cantor/file.pdf").
setq(X, Y):-lisp(setq(X, Y)).

%%	set_string(+X:atom, +Y:text) is det.
%	Ask emacs lisp to eval S-expression "(setq X Y)".

set_string(X, Y):-  smash(Y, Y0), setq(X, Y0).

%%	get_string(X:atom, -V:string) is det.
%	Unify V with the string that is bound to a lisp atom X.

% Run from dum-eh state one by one. (C-c p)
% elisp:set_string('some-unbound', "hello").
% elisp:get_string('some-unbound', X).

% Run paragraph from dum-eh state. (C-c P)
% elisp:set_string('some-unbound', "hello"),
% elisp:get_string('some-unbound', X).

get_string(Atom, String):-
	call_lisp_text(if(boundp(quote(Atom)), Atom, ""),
				   Codes),
	string_codes(String, Codes).

line(X) :- elisp:call_lisp(
	'buffer-substring'(	'line-beginning-position'(),
				'line-end-position'()),
	[value(X)]).

%%	mark_whole_buffer(X, X) is det.
%	Mark whole buffer.

mark_whole_buffer --> {lisp('mark-whole-buffer'())}.

%%	set_mark_region is det.
%	Mark paragraph region.

set_mark_region	:- lisp('mark-paragraph-region'()).

%%	set_mark_region(X, X) is det.
%	Mark paragraph region.

set_mark_region --> {set_mark_region}.

%%	line_region(?X, L:codes) is det.
%	Unify the content of the current line.

line_region(_, Line) :- line_get(Obj), obj_get([line(Line)], Obj).

%%	cur_dir(-D:atom) is det.
%	Unify D with the current directory name.

% ?- elisp:cur_dir(D).
cur_dir(D) :- lisp_atom('default-directory', D).

%%	cur_buf(-B:atom) is det.
%	Unify B with the current buffer name.

% ?- elisp:cur_buf(X).
cur_buf(B) :- current_buffer_name(B).

%%	current_buffer_name(-X:atom) is det.
%	Unify X with current buffer name.
current_buffer_name(X) :- current_buffer_name([], X).

%%	current_buffer_name(_, -A:atom) is det.
%	Unify A with the current buffer name.

% ?- elisp:current_buffer_name(_, Y).
current_buffer_name -->
	{	call_lisp('buffer-name'('current-buffer'()), [value(E)]),
		atom_codes(A, E)
	},
	peek(A).

%%	default_directory(-D:text) is det.
%	D is unified with the codes of emacs default-directory.

default_directory(D) :- call_lisp_text('default-directory', D).

%%	mark(-X:int) is det.
%	Unify X with the marked point if exists, otherwise
%	with the current point.

%	?- elisp:mark(X).
mark(X):- lisp(if(integerp(mark(t)), mark(), point()), X).

%%	point(-X:int) is det.
%	Unify X with the current point.
point(X) :- lisp(point(), X).

%%	goto_char(P:int) is det.
%	'Goto-char' to P.
goto_char(P) :- lisp('goto-char'(P)).

%%	goto_char(P:int, X, X) is det.
%	'Goto-char' to P.
goto_char(P) --> {goto_char(P)}.

%%	goto_char(K:string, S:string) is det.
%	'Goto-char' to the marked point plus I in the buffer, where
%	I is a relative position of K in S.

goto_char(Key, String) :- sub_string(String, I, _, _, Key),
	mark(P),
	P0 is P + I,
	goto_char(P0).

%%	goto_line(P:int) is det.
%	'Goto-line' to line P.

goto_line(P) :- lisp( 'goto-line'(P)).

%%	point_max(-X:int) is det.
%	Unify X with the point-max.

point_max(X):- lisp('point-max'(), X).

%%	point_min(-X:int) is det.
%	Unify X with the point-min.

point_min(X):- lisp('point-min'(), X).

%%	buffer_substring(+P:int, +Q:int, -S:codes) is det.
%	Unify S with the content of the buffer between P and Q.

% ?- elisp:buffer_substring(20, 30, X).
buffer_substring(P, Q, S):- call_lisp('buffer-substring'(P, Q), [value(S)]).

%%	edit_command_set(X, X) is det.
%	Set edit-command-set.

% edit_command_set --> {edit_command_set}.

%%	edit_command_set is det.
%	Set edit-command-set.

%% edit_command_set:- lisp(setq('edit-command-set',
%      quote([set_dired, dired,  prolog, lisp, prooftree] ))).

edit_command_set:- lisp(setq('edit-command-set',
     quote(["set_dired", "dired",  "prolog", "lisp", "prooftree"] ))).

%%	edit(X, X) is det.
%	Ask a predicate name, and Go to its definition.

edit --> {	lisp_atom('read-string'("predicate[/N] : "), P),
		setof(X-Y, find_pred:locate(P, X, Y), Pairs),
		show_select_location(Pairs),
		length(Pairs, N),
		N > 0,
		(	N  == 1
		->	J  = 1
		;	call_lisp('ask-term'("number ? "), [value(J0)]),
			number_codes(J, J0)
		),
		nth1(J, Pairs, _ - [file(F), line(G)]),
		atom_string(F, F0),
		lisp(progn('find-file'(F0), 'goto-line'(G)))
	 }.
edit --> { lisp(message("No requested predicate found.")) }.

%
show_select_location(Pairs) :-
	wait('switch-to-buffer'("*scratch*")),
	insert_buffer("** select number **\n"),
	length(Pairs, N),
	numlist(1, N, Ns),
	maplist(pred(([I,  Spec - _, [I0, ".\t- ", Spec0, "\n"]]:-
		     number_string(I, I0),
		     term_string(Spec, Spec0))),
		Ns, Pairs, Specs),
	insert_buffer(Specs).


%%	zip_minus(?X:list, ?Y:list, ?Z:list) is det.
%	Zip lists with '-'.

zip_minus([], [], []).
zip_minus([A|X], [B|Y], [A-B|R]):- zip_minus(X, Y, R).

%%	visit_file_line(File:file, N:int) is det
%	Send "find-file and goto-char" command to emacs.

% ?- elisp:visit_file_line("/Users/cantor/tmp/deldel.pl", 3).
visit_file_line(File, Line):-
	lisp(progn('find-file'(quote(File)), 'goto-line'(Line))).

%%	ask_number(P:string, X) is det.
%	Ask via emacs with prompt P, and Unify X with the input value.

ask_number(P, X):- atomics_to_string(P, P0),
	lisp('ask-term'(P0), X).

%%	ask_which(+Xs:list, -N:int) is det.
%	Display elements of Xs with a number, ask a number,
%	and unify N with the number input. When Xs is
%	a singleton list, N is unified with 1 without
%	asking.

ask_which([_], 1).
ask_which(Xs, J):- maplist([N-I, [N,-,I]], Xs, Ys),
	insert("\n\t", Ys, Zs),
	ask_number(["Select number:\n\t"|Zs], J).


		/*****************
		*     for tex    *
		*****************/

%%	wdf(X:codes, Y:codes) is det.
%	Put a dollor code at both ends.

wdf --> wrap_dollar_forwarding.

%
wrap_dollar_forwarding --> for_ascii(X\[$, X, $]).
wrap_Red_forwarding --> for_ascii(X\['\\Red{', X, '}']).
wrap_Green_forwarding --> for_ascii(X\['\\Green{', X, '}']).
wrap_Blue_forwarding --> for_ascii(X\['\\Blue{', X, '}']).
wrap_with_single_quote --> region,
	line(X\['''', X, ''''], ',\n'), smash, overwrite.

%%	find_next_ascii_codes(X, A, Y, L, L0) is det.
%	True if  A is an ascii code block, X is non ascii code block,
%	and X + A + Y is the  difference of L from L0.

% ?- elisp:find_next_ascii_codes(X, A, R, ` 宇 宙 a b c  `, Y).
find_next_ascii_codes(X, A, Y) --> skip_to_ascii_codes(X),
	ascii_codes(A),
	current(Y).

%%  ascii_codes(-X:codes, +Y:codes, -Z:codes) is det.
%	True if
%	X is unified with the maximum ascii codes prefix of Y,
%	Z with the remainder of Y.

% ?- elisp:ascii_codes(X, `ab cd  `, R).
%@ X = [97, 98, 32, 99, 100],
%@ R = [32, 32] .
ascii_codes(X) --> wl(+(?(char(space)) + char(ascii\space)),  X).

%%  skip_to_ascii_codes(-X:codes, +Y:codes, -Z:codes) is det.
%	True if
%	X is unfied with the maximum non-ascii codes prefix of Y,
%	Z  with the remainder of Y.

% ?- elisp:skip_to_ascii_codes(X, `abc`, Y).

% ?- elisp:skip_to_ascii_codes(X, `向井 国昭 abc`, Y).
skip_to_ascii_codes(X) --> wl( *(char(\+ascii|space)) | char(space)^(>=(2)), X).

first_token_codes --> wl(*char(white)), wl(*(out(white)), X), peek(X).

%%	global_set_kbd(A:string) is det.
%	Set a global default short cut for a rolog script A.

%	?-  elisp:global_set_kbd("ejockey:handle([eit])").
global_set_kbd(A) :- global_set_kbd("C-c C-u", A).

%%	global_set_kbd(Key:string, A:string) is det.
%	Set a global short cut for a prolog script A.

global_set_kbd(Key, A) :-
	lisp('global-set-key'(kbd(Key),
			quote(lambda(nil, interactive(), 'prolog-query'(A))))).

%%	global_unset_kbd(K:string) is det.
%	global-unset key in kbd form.

global_unset_kbd(K):- lisp('global-unset-key'(kbd(K))).

		/******************************************
		*     Convert xml to quasi bibtex form    *
		******************************************/

% ?- scm(elisp).
% ?- abbreviation('ab cd', X).

abbreviation(A, B):- string_code(I, A, 0'\s), !,    % '
					 succ(J, I),
					 sub_string(A, 0, J, _, B0),
					 atom_string(B, B0).
abbreviation(A, A).

% ?- make_key([published='2014', author='Obama'], Key).

make_key(L, Key):- memberchk(published=Year, L),
	memberchk(author=Author, L),
	!,
	abbreviation(Author, Abbr),
	atom_concat(Abbr, Year, Key).
make_key(L, Key):- memberchk(published=Year, L),
	memberchk(publisher=Pub, L),
	!,
	abbreviation(Pub, Abbr),
	atom_concat(Abbr, Year, Key).
make_key(L, Key):- memberchk(published=Year, L),
	memberchk(publisher=Pub, L),
	!,
	abbreviation(Pub, Abbr),
	atom_concat(Abbr, Year, Key).
make_key(_, '??????').

% Use this predicate after exporting onto the Desktop directory.
% run unix command delicious-export  [2016/12/16]
% ?- parse_xml_for_bibtex(Bib), maplist(writeln, Bib), length(Bib, Length).
% ?- parse_xml_for_bibtex(Bib), open('/Users/cantor/Desktop/test', write, X),maplist(writeqln(X),  Bib), close(X).

% Delicious Library 3; incremental export to bibtex
% revised Workflow:
%	(0) => (1) =>(2) => (3)
%
% (0)  initialize
% %  touch ~/TeXLive/texmf-var/bibtex/bib/delicious-db.pl
%
% (1) run unix script [2016/12/16]:
% %  delicious-export
%
% (2) run prolog query:
% ?- elisp:delicious_update.
%@ true .

%
% (3) Open the output at the Desktop, and  edit .bib files manually.

% ?- elisp:delicious_bibtex_terms("~/TeXLive/texmf-var/bibtex/bib/deldelbooks1.pl").
% ?- elisp:delicious_bibtex_terms("~/Desktop/test.pl", "~/Desktop/test.bib").

% ?- trace, elisp:delicious_update.

delicious_update:-
	delicious_bibtex_terms("~/TeXLive/texmf-var/bibtex/bib/delicious-db.pl",
						   "~/Desktop/new-delicious", ".bib").

delicious_bibtex_terms(DB, Bib_basic, Ext) :- parse_xml_for_bibtex(Bib), !,
	   sort(Bib, Bib0),
	   expand_file_name(DB, [DB0|_]),
	   pshell(touch(DB0)),
	   read_file_to_terms(DB0, Old_bib, []),
	   subtract(Bib0, Old_bib, NewBooks),
	   open(DB0, write, Y),
	   maplist(writeq_book_term(Y),  Bib0),
	   close(Y),
	   sort(NewBooks, NewBooks0),
	   maplist(book_bibitem, NewBooks0, Bib1),
	   atomics_to_string([Bib_basic, Ext],Bib_file),
	   expand_file_name(Bib_file, [Out0|_]),
	   string_length(Ext, N),
	   sub_string(Out0, 0, _, N, Out_basic),
	   ejockey:modify_file_name(Out_basic, 0, Ext, NewOut),
	   open(NewOut, write, Z),
	   maplist(writeln(Z),  Bib1),
	   close(Z).

writeq_book_term(X, Book):-
	writeq(X, Book),
	write(X, ".\n").

%
delicious_bibtex(Out) :- parse_xml_for_bibtex(Bib), !,
	   sort(Bib, Bib0),
	   maplist(book_bibitem, Bib0, Bib1),
	   expand_file_name(Out, [Out0|_]),
	   open(Out0, write, X),
	   maplist(writeln(X),  Bib1),
	   close(X).

writeqln(S, X):- writeq(S, X), write(S, '.\n').

% parse_xml_for_bibtex(Bib):-
% 	expand_file_name("~/Desktop/Delicious Library/Library Media Data.xml", [File]),
% 	parse_xml_for_bibtex(File, Dom),
% 	maplist(pred(([book(L), book([key=Key|L])]:- make_key(L, Key))),
% 			Dom, Bib).

% % ?- parse_xml_for_bibtex("/path/to/xml-file", Dom).
% parse_xml_for_bibtex(File, Books):-
% 	Keys = [
% 			published,
% 			publisher,
% 			title,
% 			upc,
% 			author],
% 	parse_xml_to_books(File, Books, Keys).
% %
% parse_xml_to_books(File, Books, Keys):-
% 	load_xml(File, Dom, [encoding('utf-8')]),
% 	flatten_element_list(Dom, Books, [], Keys).

%
flatten_element_list([A|B], X, Y, Keys):-
	once(flatten_element(A, X, X0, Keys)),
	flatten_element_list(B, X0, Y, Keys).
flatten_element_list([], X, X, _).

%
flatten_element(element(library, _, L), X, Y, Keys):-
	flatten_element_list(L, X, Y, Keys).
flatten_element(element(items, _, L), X, Y, Keys):-
	flatten_element_list(L, X, Y, Keys).
flatten_element(element(book, L, _), [book(L0)|X], X, Keys):-
	projection(L, Keys, L0).
flatten_element(_, X, X, _).

%
projection([], _, []).
projection([Eq|L], Keys, [Eq0|L0]):- modify_value(Eq, Eq0), !,
	projection(L, Keys, L0).
projection([Eq|L], Keys, [Eq|L0]):- Eq = (K = _), memberchk(K, Keys), !,
	projection(L, Keys, L0).
projection([_|L], Keys, L0):- projection(L, Keys, L0).

%
modify_value(published = V, published = V0):-
	sub_atom(V, 0, 4, _, V0).
modify_value(K = '－',  K = void).
modify_value('aspect=－',  aspect = undefined).

% Target sample.
% @book{Noro-2003,
% 	author = {野呂 正行 and 横山 和弘},
% 	publisher = {東京大学出版会},
% 	pages = {288},
% 	title = {グレブナー基底の計算基礎篇－計算代数入門},
% 	keywords = {科学・テクノロジー},
% 	ISBN = {4130614045},
% 	language = {日本語 (Published)},
% 	price = {￥ 4,536},
% 	year = {2003}
% }

book_bibitem(book(L), BB):-
	select(key=Keyval, L, L0),
	maplist(pred( [published=V, [year, " = ", "{", V, "}"]]
				& [K=V, [K, " = ", "{", V, "}"]]),
			L0, Items),
	insert(",\n", Items, Items0),
	smash(["@book{", Keyval, ",\n", Items0, "\n}\n"], BB).

%% meta_handle(+G, +M, +Sp, +P)
%	* +G	: goal,
%	* +M	: module prefix
%	* +Sp	: a predicate with arity 4.
%	* +P	: a list, [] or [_,_]
%
%   Execute G almost in the same way as
%   for phrase(G, X, Y) when P = [X,Y],
%	or call(G) when P=[].
%   For each atomic pharse P of G
%
%		meta_handle(B, M, Sp, [])
%
%   is executed when call(Sp, M, P, B) is true.
%
%	This predicate is convenient to test
%   handles, which may have handle calls in the body,
%   and see their outputs on emacs buffer window.

% helpers.

id([X,X]).
id([]).

transitive([], [], []).
transitive([X, Y], [X, Z], [Z, Y]).

:- meta_predicate meta_handle(:,+,4,+).

%
meta_handle(X,Y,Z,U):- meta_handle_plain(X,Y,Z,U).

%
meta_handle_plain((A,B), M, Sp, P):- transitive(P, Q, R),
     meta_handle_plain(A, M, Sp, Q),
	 meta_handle_plain(B, M, Sp, R).
meta_handle_plain(M:A, _, Sp, P):-  meta_handle_plain(A, M, Sp, P).
meta_handle_plain(A;B, M, Sp, P):-
	 (meta_handle_plain(A, M, Sp, P)
	 ; meta_handle_plain(B, M, Sp, P)).
meta_handle_plain(A->B, M, Sp, P):- transitive(P, Q, R),
	 meta_handle_plain(A, M, Sp, Q), !,
	 meta_handle_plain(B, M, Sp, R).
meta_handle_plain((A->B; C), M, Sp, P):- transitive(P, Q, R),
	(meta_handle_plain(A, M, Sp, Q), !, meta_handle_plain(B, M, Sp, R)
	; meta_handle_plain(C, M, Sp, P)).
meta_handle_plain(nopac({A}), M, _, P):- id(P),
	( M==[], !, call(A)
	; call(M:A) ).
meta_handle_plain(A, M, Sp, P)	:- Sp\==[], Sp\= _:[],
	call(Sp, A, M, P, G),
	!,
	meta_handle_plain(G, M, Sp, []).
meta_handle_plain(A, [], _, P):- apply(A, P).
meta_handle_plain(A, M, _, P):- apply(M:A, P).

:- initial_region_obj.
