:- module(prooftree, [prooftree/2, checkproof/2, proof/2, proof_figure/2]).

:- use_module(util(misc)).

:- use_module(pac('expand-pac')).
:- use_module(pac(op)).
:- use_module(pac(basic)).

term_expansion --> pac:expand_pac.

%%% convert text base proof figure  into prooftrees
%%% [2006/07/21]
% platex via lisp.

proof --> region, prooftree, math.

math --> latex_math( (generate_tex, platex, get_size,
			dvipdfmx, cleaning_tmp_file,
			latex_open_file(pdf), clear) ).

proof_figure --> current(T0),
	prooftree,
	current(P),
	math,
	{ comment_out_region(T0, T) },
	peek([T,"\n", P]),
	overwrite.

proof_platex_region --> current(T0),
	prooftree, current(P),
	{ comment_out_region(T0, T) },
	peek([T,"\n", P]),
	overwrite,
	{ elisp:send_off("(command-execute 'latex-region-interactive)") }.

comment_out_region --> split, maplist(comment_out), insert("\n").

comment_out(X, Y) :- append(`%`, _, X) -> Y = X ; append(`% `, X, Y).

% ?- trace, prooftree:prooftree(`a\n--(rule)\nb`, X).
% ?- trace, prooftree:prooftree(`a\n--c\nb\n`, X).
%@ X = "$$\\prooftree\na\n\\justifies\nb\n\\using c\n\\endprooftree$$\n" .

%@ X = "$$\\prooftree\na\n\\justifies\nb\n\\using c\n\\endprooftree$$\n" .

prooftree -->
	paragraph,
	maplist(split),
	maplist(remove([])),
	remove([]),
	maplist(text2proof),
	maplist(prooftree:gentex),
	smash.

checkproof --> paragraph, remove([]), maplist(oneproof_check),
	cons(["**** Checking proof trees figure: \n"]), flatten.

oneproof_check --> split,
	maplist(kleene_star_greedy(`%`)),
	reverse,
	textpos,
	prooftree_check.

text2proof(P, Y):-maplist(kleene_star_greedy(`%`),P,P0),
    reverse(P0, X),
    textpos(X, X1),
    X1=[[M]|L],
    mkproof(M, Y, L).

textpos([],[]).
textpos([X|Y],[X1|Y1]):-
	line(X1,0,_,X,[]),
	!,
	textpos(Y,Y1).

line(X,I,K)-->spaces(I,J),line1(X,J,K).
line(X,I,K)-->line1(X,I,K).

line1([c(X,I,J)|Y],I,K)-->token(X,I,J),line(Y,J,K).
line1([],I,I)-->[].

token([X|Y],I,K)-->[X],{X\==(0'\s),J is I+1},tokenx(Y,J,K).  %'

tokenx([X|Y],I,K)-->[X],{X\==(0'\s),J is I+1},tokenx(Y,J,K). %'
tokenx([(0'\s)|X],I,K)-->[(0'\s)],{J is I+1},token(X,J,K).	 %'
tokenx([],I,I)-->[].

spaces(I,K)-->[(0'\s)],{J is I+1},spaces(J,K). %'
spaces(I,I)-->[].

manyspaces(I,K)-->spaces(I,K),{K-I>=2}.

%%--------------------------------------------------------------
%% create a proof tree:  p(rule-name, coclusion, antecedents)
%%--------------------------------------------------------------

mkproof(F,p(void,F,[]),[]).	% make proof tree
mkproof(F,p(J,F,A),[L,R|S]):-
	mkusing(F,J,L),
	mkantecedents(J,B,R),
	mkprooflist(B,A,S).

mkprooflist([],[],_).   	 % make proof tree list
mkprooflist([A|B],[A1|B1],L):-
	mkproof(A,A1,L),
	mkprooflist(B,B1,L).

mkusing(c(_,S,E),c(M,S1,E1),L):-
	member(c(M,S1,E1),L),
	touch([S,E],[S1,E1]),
	!.
mkusing(_,void,_).

mkantecedents(void,[],_):-!.
mkantecedents(X,A,L):-suclist(X,L,A).

touch([_,B],[X,_]):-B<X,!,fail.
touch([A,_],[_,Y]):-Y<A,!,fail.
touch(_,_).

touchx(c(_,A,B), c(_,X,Y)):-touch([A,B],[X,Y]).

gentex --> displaymath.

displaymath(X, ["$$", X1,  "$$\n"]):- proof2tex(X, X1).

% ?- expand_exp(:flip([Y]:- string_codes(Y, X)), p, V, user, G, P, []).
proof2tex(p(void,c(X,_,_),_),Y):-!,string_codes(Y,X) .
proof2tex(p(c(X,_,_),c(Y,_,_),L0),
		  ["\\prooftree\n",A,"\n","\\",Conclusion,"\n",Y,"\n","\\using ",Rule,"\n","\\endprooftree"]):- (kleene_string_codes("-",X,Rule)->Conclusion="justifies";kleene_string_codes("=",X,Rule)->Conclusion="Justifies";kleene_string_codes(".",X,Rule)->Conclusion="leadsto"),!,(insert(" ",L0,A5),flatten(A5,A6)),maplist(proof2tex,A6,A) .
proof2tex(p(c(X,_,_),c(Y,_,_),L0),["\\mathvbox{",A,"}{",Y,"}"]):-kleene_string_codes("*",X,_),!,(insert(" ",L0,A6),flatten(A6,A7)),maplist(proof2tex,A7,A) .
proof2tex(X,X):-! .

suclist(_,[],[]).
suclist(X,[Y|Z],[Y|U]):- touchx(X,Y),!,suclist(X,Z,U).
suclist(X,[_|Z],U):- suclist(X,Z,U).


%%-------------------------------------------------------
%% checking a proof tree
%%-------------------------------------------------------
prooftree_check(X,[ "main: ", X1,
	 "height: ",  X2,
	 "conclusion: ", X3,
	 "antecedent: ", X4,
	 "using: ", X5]):-checkmain(X,X1),
	checkheight(X,X2),
	checkconcl(X,X3),
	checkante(X,X4),
	checkusing(X,X5).

checkmain([[_]|_], "ok\n").
checkmain([_|_], "multiple main conclusion\n").
checkmain(_,"ok\n").

checkheight([_,_,_|_],"ok\n").
checkheight(_, ["too short proof tree\n"]).

checkconcl([X,Y|Z], A):-function(Y,X),!,checkconcl([Y|Z], A).
checkconcl([_,_|_], ["multiple conclusions", " or ",  "ambiguous antecedents", " or ", "orphant\n"]).
checkconcl(_,"ok\n").

checkante([_|A],B):-checkante1(A,B).

checkante1([X,Y|Z],B):-totalrel(X,Y),!,checkante1(Z,B).
checkante1([_,_|_], ["empty antecedants\n"]).
checkante1(_,"ok\n").

checkusing([X,Y|L],M):-checkusing1(X,Y),!,checkusing(L,M).
checkusing([_,_|_], "multiple usings\n").
checkusing(_,"ok\n").

checkusing1([],_).
checkusing1([X|Y],Z):- suclist(X,Z,A),atmostone(A),checkusing1(Y,Z).

atmostone([]).
atmostone([_]).

function([],_).
function([X|Y],Z):-suclist(X,Z,S),S=[_],function(Y,Z).

totalrel([],_).
totalrel([X|Y],Z):-suclist(X,Z,S),S=[_|_],totalrel(Y,Z).
