:- module(seqcal,[]).
:- use_module(util(misc)).
:- use_module(pac(basic)).
:- use_module(util('emacs-handler')).

%% Sequent Calculus LK (i.e. Classical, but not Intuitionistic !)
%%  for propositional logic.
%% 2002, 2003, 2004   by K. Mukai
%%
%% An important remark on this program from Logic (by Hagiya of Tokyo University
%%  in some occasion of private communication.)
%%
%%   In classical logic,   `A implies  B`
%%   is logically equivalent to `(not A ) or B.`
%%   On the other hand, it is very simple and straightforward to implement
%%   propositional version of LK without `implies` as seen in the code below.
%%   So, having a look into the code below, it should be clear that this program
%%   is a correct implementation of a prover for propositonal LK.
%%

:- current_op(X,_,->),	X0 is X-10, op(X0,xfy,<>).
:- initialization((current_op(X,_,->),	X0 is X-10, op(X0,xfy,<>)), restore).

% :-op(10, fy, (!)).     % negation
% :-op(800, xfy, (<>)).   % equivalence
% :-op(100, xfx, (>>)).   % separater in a sequent  (usually |- )

%     >     for logical IMPLICATION
%     *     for logical AND
%     +     for logical OR

% Sample use.
% ?- trace, seqcal:prove(a+ !(a)).
% ?- seqcal:prove(a > (b > a)).
% ?- seqcal:prove(!((a+ b)*(a+ !(b))*(!(a)+b)*(!(a)+ !(b)))).


prove(X):- seqcal([]>>[X], P), !,
    format("~w is provable in LK.~n~n",[X]),
	p2q(P,Q),
	raster(Q,R),
	raster(R).
prove(X):-format("~w is NOT provable in LK.~n~n",[X]).

% ?- trace, seqcal:seqcal([] >> [(a+ !(a))], R).

seqcal(X >> Y, p(axiom, (X>>Y))):- member(Z,X),  member(Z,Y), !.
seqcal(X >> Y, p(left(R), (X>> Y), P)):-
	removeOne(X,A,B),
	rewrite(A,A1),
	left(A1,C,R),
	subgoals(B >> Y, C, L),
	seqcallist(L, P).
seqcal(X >> Y, p(right(R),X>>Y, P)):-
	removeOne(Y,A,B),
	rewrite(A,A1),
	right(A1,C,R),
	subgoals(X >> B,C,L),
	seqcallist(L,P).

subgoals(_,[],[]).
subgoals((X >> Y),[A >> B|C],[X1 >> Y1|C1]):-
	append(A,X,X1),
	append(B,Y,Y1),
	subgoals(X >> Y,C,C1).

seqcallist([],[]).
seqcallist([X|Y],[P|Q]):-seqcal(X,P),seqcallist(Y,Q).

right(A * B, [[]  >>  [A], []  >>  [B]], (*)).
right(A + B,  [[]  >> [A,B]], (+)).
right(A > B,  [[A] >> [B]], (>)).
right(!(A), [[A] >> []], (!)).

left(A * B, [[A,B] >> []],(*)).
left(A + B,  [([A] >> []), ([B] >> [])], (+) ).
left(A > B,  [([] >> [A]), ([B] >> [])], (>) ).
left(!(A), [[] >> [A]], (!)).

macrodef(A<B, B>A).
macrodef(<>(A,B), (A>B) * (B>A)).

rewrite(A,B):-macrodef(A,B),!.
rewrite(A,A).

removeOne([A|B],A,B).
removeOne([A|B],C,[A|D]):-removeOne(B,C,D).

%%
%% Format of p-term:
%%   p(axiom,
%%     <sequent>
%%   )
%%
%%
%%   p(<rule name>,
%%       <sequent>,
%%       [<sequent>...]
%%   )
%%
%%
%% Format of q-term:
%%   q(<width>,
%%     goal(<root>,[<R1>,<R2>,<R3>]),
%%     rule(<name>,[<N1>,<N2>,<N3>,<N4>]),
%%     subgoal([<box>...],[<P1>,<P2>,<P3>])
%%   )
%% where
%%  <width> = <P1> + <P2> + <P3>
%%          = <R1> + <R2> + <R3>
%%          = <M1> + <M2> + <M3>
%%          = <N1> + <N2> + <N3> + <N4>
%% length of  <root> = <R2>
%% length of  <name> = <N3>
%%
%%    <M1> +  <M2>/2 = <N1> + <N2>/2
%%                   = <R1> + <R2>/2
%%    <M2>  =  <width> of <box_1>
%%            + <width> of <box_2>
%% 	           ...
%%            + <width> of <box_n>
%%            - <R1>  of <box_1>
%%            - <R3>  of <box_n>
%%


p2q(p(axiom, A),
    q(W,
      goal(A1,[0,W,0]),
      rule([],[]),
      subgoal([],[])
     )
)         :- !,sequent_codes(A, A1), length(A1, W).
p2q(p(J,A,S),
    q(W,
      goal(A1,[R1,R2,R3]),
      rule(J1,[N1,N2,N3,N4]),
      subgoal(S1,[P1,P2,P3])
    )
)   :- sequent_codes(A,A1), length(A1,R2),
    rule_codes(J,J1),length(J1,N3),
    p2q_list(S,S1),
    subgoal_width(S1,M1,M2,M3),
    P2 is M1 + M2 + M3,
    M2r is  M2//2,
    M2l is M2 - M2r,
    R2r is R2//2,
    R2l is R2-R2r,

   ( M1+M2l >= R2l
     ->
    P1 = 0
    ;
     P1 is  R2l - (M1 + M2l)
  ),

  ( M2 >= R2
    ->
    N2 = M2
    ;
    N2 = R2
  ),

   N2r is N2//2,
   N2l is N2 - N2r,

  (N2r + N3  =< M2r + M3
   ->
   P3 = 0,
   N4 is  M2r + M3 - (N2r + N3)
   ;
   P3 is N2r + N3 - (M2r + M3),
   N4 = 0
  ),

   N1 is (P1 + M1 + M2l) - N2l,
   R1 is (P1 + M1 + M2l) - R2l,
   R3 is (M2r + M3 + P3) - R2r,
   W  is P1 + P2 + P3.

sequent_codes((X >> Y), Z) :-
    insert(`, `, X, X1),
    insert(`, `, Y, Y1),
    M = [` `, X1 ,  ` |- `, Y1 , ` `],
    smash_codes(M, Z).


subgoal_width(X,M1,M2,M3):-
	subgoal_width1(X,M1),
	subgoal_width2(X,L),
	subgoal_width3(X,M3),
	M2 is L - (M1 + M3).

subgoal_width1([q(_, goal(_,[M|_]), _, _)|_], M).

subgoal_width2([],0).
subgoal_width2([q(W,_,_,_)|R],M):- subgoal_width2(R, M1), M is W + M1.

subgoal_width3(X,M):- last(X, q(_, goal(_, [_, _, M]), _, _)).

p2q_list([],[]).
p2q_list([X|Y],[X1|Y1]):- p2q(X, X1), p2q_list(Y, Y1).

%%%%%%%%% Print Proof %%%%%%

spaces(N,X):-repeat_code(0'\s , N, X).

line(N,X):-repeat_code(0'-, N, X).

repeat_code(_,0,[]):-!.
repeat_code(C,N,[C|R]):- N > 0,  N1 is N-1, repeat_code(C,N1,R).

rule_codes(right(X),Y):-!, smash_codes(X,X1), append([`(|-`,X1,`)`],Y).
rule_codes(left(X),Y):-!,  smash_codes(X,X1), append([`(`,X1,`|-)`],Y).
rule_codes(X,Y):-smash_codes(X,Y).

raster([]).
raster([X|Y]):-raster(Y), format("~s~n",[X]).

raster(q(_,
      goal(A1, [L, _, R]),
      rule(_, _),
      subgoal([], _)
     ),
     [X]
    ):- !, spaces(L, SL), spaces(R, SR), append([SL, A1, SR], X).
raster(q(_,
      goal(A1, [R1, _, R3]),
      rule(J1, [N1, N2, _, N4]),
      subgoal(S1, [P1, _, P3])
     ),
     [X, Y|Z]):-
   spaces(R1,SR1), spaces(R3,SR3), append([SR1,A1,SR3],X),
   spaces(N1,SN1), spaces(N4,SN4), line(N2,Hline),
   append([SN1,Hline,J1,SN4],Y),
   raster_list(S1,B),
   spaces(P1,SP1),
   spaces(P3,SP3),
   map_append3(SP1,B,SP3,Z).

map_append3(_,[],_,[]).
map_append3(A,[X|Y],B,[X1|Y1]):- append([A,X,B],X1), map_append3(A,Y,B,Y1).

height_adjust_append([],[],_,_,[]).
height_adjust_append([X|Y],[],A,B,[Z|U]):-!,append(X,B,Z), height_adjust_append(Y,[],A,B,U).
height_adjust_append([],[X|Y],A,B,[Z|U]):-!, append(A,X,Z), height_adjust_append([],Y,A,B,U).
height_adjust_append([X|Y],[P|Q],A,B,[Z|U]):- append(X,P,Z), height_adjust_append(Y,Q,A,B,U).

raster_list([],[[]]).
raster_list([X|Y],Z):- raster(X,X1), raster_list(Y,Y1),
	X1=[A|_],
	Y1=[B|_],
	length(A,L1),
	length(B,L2),
	spaces(L1,S1),
	spaces(L2,S2),
	height_adjust_append(X1,Y1,S1,S2,Z).

gentzen --> peek(X, [[]>>[X],X]),
	eh:phrase_on_car(seqcal:seqcal),
	!,
	eh:phrase_on_car(seqcal:(p2q, raster, rasterx)),
	peek([A, B], [B, A]),
	eh:phrase_on_car((smash_codes, seqcal:convert_special)),
	peek([C, D], [`<p> `, C, ` is provable in LK.<br><br></p>`,
	`<p><pre>`, D, `</pre></p>`]),
	smash.
gentzen --> smash_codes,
	convert_special,
	peek(A, [`<p> `, A, ` is NOT provable in LK.<br><br></p>`]),
	smash.

rasterx([],[]).
rasterx([X|Y],Z):-rasterx(Y,Y1),
	convert_special(X,X1),
	append([X1, `<br>`, Y1],Z).

convert_special(X,Y):-convert_special(X,Y,[]).

convert_special([],X,X).
convert_special([0'<|R],[0'&,0'l,0't,0';|X],Y):-!, convert_special(R,X,Y).
convert_special([0'>|R],[0'&,0'g,0't,0';|X],Y):-!, convert_special(R,X,Y).
convert_special([A|R],[A|X],Y):- convert_special(R,X,Y).
