:- module(count_cudg, []).
:- use_module(util(math)).
:- use_module(util(meta2)).
:- use_module(pac(basic)).
:- use_module(pac(meta)).
:- use_module(util(misc)).
:- use_module(pac('expand-pac')).
:- use_module(zdd('zdd-array')).
:- use_module(zdd(zdd)).
:- use_module(zdd('zdd-misc')).
:- use_module(zdd('zdd-plain')).
:- use_module(pac(op)).

% ?- mk_union_find_command(a-b, [c], F).
% ?- mk_union_find_command(a-b, [b,c], F).

		/***********************************************
		*     count connected udg with pow and memo    *
		***********************************************/

% ?- count_cudg([a,b], C).
%@ C = 1.
% ?- count_cudg([a,b,c], C).
%@ C = 4.
% ?- count_cudg([a,b,c,d], C).
%@ C = 38.

% ?- count_cudg([a,b,c,d,e], C).
% ?- count_cudg([a,b,c,d,e,f], C).
% ?- count_cudg([a,b,c,d,e,f,g], C).
% ?- time(count_cudg([a,b,c,d,e,f,g,h], C)).
%@ % 26,904,200 inferences, 2.565 CPU in 2.623 seconds (98% CPU, 10486976 Lips)
%@ C = 251548592.

% ?- N=10, time((numlist(1, 10, Ns), count_cudg(Ns, C))).
%@ % 1,439,423,037 inferences, 142.657 CPU in 145.688 seconds (98% CPU, 10090105 Lips)
%@ N = 10,
%@ Ns = [1, 2, 3, 4, 5, 6, 7, 8, 9|...],
%@ C = 34496488594816.

% ?- N=11, numlist(1, N, Ns), call_with_time_limit(18000, time(count_cudg(Ns, C))).
%@ % 9,519,980,664 inferences, 1300.998 CPU in 1317.768 seconds (99% CPU, 7317443 Lips)
%@ N = 11,
%@ Ns = [1, 2, 3, 4, 5, 6, 7, 8, 9|...],
%@ C = 35641657548953344.

count_cudg(Ns, C):- count_connected_udg(Ns, C, _, _).
%
count_connected_udg(Ns, C, X, S):-
	open_state(S),
	zdd_sort(Ns, Ns0, S),
	all_links(Ns0, Links),
	findall([A], member(A, Ns0), Singletons),
	zdd_ord_insert(Singletons, 1, Init, S),
	memo(add_link(1)-Init, S),
	zdd_power(Links, Pow, S),
	add_links_to_udg(Pow, U, S),
	collect_finals(U, X, S),
	card(X, C, S).

%
add_links_to_udg(0, 0, _):- !.
add_links_to_udg(1, Y, S):- !, memo(add_link(1)-Y, S).
add_links_to_udg(X, Y, S):- memo(add_link(X)-Y, S),
	(	nonvar(Y) -> true		%,  write(.) % many hits.
	;	cofact(X, t(A, L, R), S),
		add_links_to_udg(L, L0, S),
		add_links_to_udg(R, R1, S),
		add_link(A, R1, R0, S),
		zdd_join(L0, R0, Y, S)
	).

		/***************************
		*     common predicates    *
		***************************/

%! mk_union_find_command(+,+,?) semidet.
mk_union_find_command(A-B, D, F):-	successive_select(D, [A, B], R),
	(	R = []  -> F = zdd_ord_insert([D, A->B])
	;	R = [A] -> F = union_find_insert(A->B, A, D)
	;	R = [B] -> F = union_find_insert(A->B, B, D)
	).

% ?- successive_select([b, a, c], [d], R).
% ?- successive_select([b, a, b, c], [b, a], R).
successive_select([], Y, Y):-!.
successive_select([A|X], Y, Z):-select(A, Y, Y0), !,
	successive_select(X, Y0, Z).
successive_select([_|X], Y, Z):-successive_select(X, Y, Z).

% ?- all_links([a,b,c], Links).
all_links(Nodes, Links):-
	findall(A-B, (append(_, [A|R], Nodes),
				  member(B, R),
				  A@<B
				 ),
			Links).

% ?- zdd((X<<set([[b]]), @(union_find_insert(a->b, b, [a], X, Y)), psa(X), psa(Y))).
union_find_insert(_E, _A, _G, X, X, _):- X < 2, !.
union_find_insert(E, A, G, X, Y, S):-
	cofact(X, t(U, L, R), S),
	(	U= (_->_) ->	Y = X
	;	union_find_insert(E, A, G, L, L0, S),
		(	memberchk(A, U) ->
			ord_union(G, U, H),
			zdd_ord_insert([H,E], R, R0, S)
		;	union_find_insert(E, A, G, R, R1, S),
			zdd_insert(U, R1, R0, S)
		),
		zdd_join(L0, R0, Y, S)
	).

% ?- zdd((X<<{[[a], [b], [c]]},
%	@(add_links([a-b, b-c], X, Y)), psa(Y))).
% ?- zdd((X<<{[[a],[b],[c]]},
%	@(add_links([a-b, a-c], X, Y)), @(collect_finals(Y, Y0)), psa(Y0))).
% ?- zdd((X<<{[[a], [b], [c], [d]]}, psa(X),
%	@(add_link(a-b, X, Y)), psa(Y), @(add_link(b-c, Y, Z)),
%	@(add_link(a-d, Z, U)), psa(U))).

add_link(_, X, 0, _):- X<2, !.
add_link(U, X, Y, S):- % memo is useless here.
	cofact(X, t(A, L, R), S),
	(	A = (_->_) -> Y = 0
	;	add_link(U, L, L0, S),
		(	mk_union_find_command(U, A, F) ->
			call(F, R, R0, S)
		;	add_link(U, R, R1, S),
			zdd_insert(A, R1, R0, S)
		),
		zdd_join(L0, R0, Y, S)
	).

% X<<{[[a,b], a->b]}, psa(X), @(collect_finals(X, Y)), psa(Y))).
collect_finals(X, 0, _):- X<2, !.
collect_finals(X, Y, S):- cofact(X, t(_, L, R), S),
	collect_finals(L, L0, S),
	collect_pure_links(R, R0, S),
	zdd_join(L0, R0, Y, S).

%
collect_pure_links(X, X, _):-X<2, !.
collect_pure_links(X, Y, S):- cofact(X, t(A, L, R), S),
	collect_pure_links(L, L0, S),
	(	A = [_|_] -> R0 = 0
	; 	zdd_insert(A, R, R0, S)
	),
	zdd_join(L0, R0, Y, S).

		/***********************************
		*     naive count connected udg    *
		***********************************/

% ?- naive_count_cudg([a, b], C).
% ?- naive_count_cudg([a, b, c], C).
% ?- N=2, numlist(1, N, Ns), naive_count_cudg(Ns, C).
% ?- N=3, numlist(1, N, Ns), naive_count_cudg(Ns, C).
% ?- N=4, numlist(1, N, Ns), naive_count_cudg(Ns, C).
% ?- N=5, numlist(1, N, Ns), naive_count_cudg(Ns, C).
% ?- N=6, numlist(1, N, Ns), naive_count_cudg(Ns, C).
% ?- N=7, numlist(1, N, Ns), naive_count_cudg(Ns, C).
% ?- N=8, numlist(1, N, Ns), time(naive_count_cudg(Ns, C)).
% ?- N=9, numlist(1, N, Ns), time(naive_count_cudg(Ns, C)).
%@ % 583,687,085 inferences, 70.093 CPU in 71.077 seconds (99% CPU, 8327282 Lips)
%@ N = 9,
%@ Ns = [1, 2, 3, 4, 5, 6, 7, 8, 9],
%@ C = 66296291072.

%  Smart version  is faster than naive one.
% ?- N=9, numlist(1, N, Ns), time(count_cudg(Ns, C)).
%@ % 195,807,591 inferences, 18.776 CPU in 19.663 seconds (95% CPU, 10428519 Lips)
%@ N = 9,
%@ Ns = [1, 2, 3, 4, 5, 6, 7, 8, 9],
%@ C = 66296291072.


naive_count_cudg(Ns, C):- naive_count_connected_udg(Ns, C, _, _).
%
naive_count_connected_udg(Ns, C, X, S):-
	open_state(S),
	zdd_sort(Ns, Ns0, S),
	all_links(Ns0, Links),
	findall([A], member(A, Ns0), Singletons),
	zdd_ord_insert(Singletons, 1, Init, S),
	add_links(Links, Init, U, S),
	collect_finals(U, X, S),
	card(X, C, S).

%
add_links([], X, X, _).
add_links([L|Ls], X, X0, S):- % memo is useless here.
	add_link(L, X, X1, S),
	zdd_join(X, X1, X2, S),
	add_links(Ls, X2, X0, S).
