:- module(zdd_array, [
				open_state/1, open_state/2, close_state/1
			,	memo/1, memo/2, memochk/1, memochk/2, memoq/1, memoq/2
			,	memo_index/2
 			,	open_memo/0, open_memo/1
			,	close_memo/0, close_memo/1
			,	set_memo/1, set_memo/2, update_memo/2, update_memo/3
			,	memo_add_new/1, memo_add_new/1, insert_memo/3
			,	unify_args/3
			,	add_child/2, add_child/3
			,	pred_memo_update/2, pred_memo_update/3
			,	cofact/2, cofact/3, index/3
			,	zdd_slim/2, zdd_slim/3
			,	zdd_copy/4, zdd_ord_copy/4, zdd_pred_copy/5
			,	inherit_compare/1, inherit_compare/2
			,	inherit_compare_opp/1, inherit_compare_opp/2
			  ]).

:- use_module(zdd(zdd)).

bucket_aware_rehash.		% #(_, _, #(_,_))
% \+ bucket_aware_rehash.	% #(_, #(_,_))

% suppress details of hash/vector.
user:portray(X):- compound(X),
	(	functor(X, s, N), (N=3; N=4) -> write('..')
	;	functor(X, ctrl, N), N>2 -> write('<ctrl>')
	;	compound_name_arguments(X, F, _),
		( F = (#) -> write(#(..))
		; F = call_continuation -> write(cont(..))
		)
	).

% ?- hash_initial(X), writeln(X).
% ?- hash_initial(100, X).
:- if(bucket_aware_rehash).
hash_initial(N, #(0, 0, V)):-!, functor(V, #, N).
hash_initial(#(0, 0, #([]))).
:- else.
hash_initial(N, #(0, V)):-!, functor(V, #, N).
hash_initial(#(0, #([]))).
:- endif.

% hash table access.
:- if(bucket_aware_rehash).
goal_expansion(hash_initial(X),  X = #(0, 0, #([]))).
goal_expansion(hash_initial(N, X),
			   (	X = #(0, 0, V),
					functor(V, #, N))).
goal_expansion(hash_initial(X),  X = #(0, 0, #([]))).
goal_expansion(hash_initial(N, X),
			   (	X = #(0, 0, V),
					functor(V, #, N))).
goal_expansion(hash_key_count(H, C), ( arg(1, H, C) )).
goal_expansion(hash_key_count_set(H, C), ( setarg(1, H, C) )).
goal_expansion(hash_bucket_count(H, C), ( arg(2, H, C) )).
goal_expansion(hash_bucket_count_set(H, C), ( setarg(2, H, C) )).
goal_expansion(hash_vector(H, V), ( arg(3, H, V) )).
goal_expansion(hash_vector_set(H, V), ( setarg(3, H, V) )).

:- else.

goal_expansion(hash_initial(X),  X = #(0, #([]))).
goal_expansion(hash_initial(N, X),
			   (	X = #(0, V),
					functor(V, #, N))).
goal_expansion(hash_initial(X),  X = #(0, #([]))).
goal_expansion(hash_initial(N, X),
			   (	X = #(0, V),
					functor(V, #, N))).
goal_expansion(hash_key_count(H, C), ( arg(1, H, C) )).
goal_expansion(hash_key_count_set(H, C), ( setarg(1, H, C) )).
goal_expansion(hash_vector(H, V), ( arg(2, H, V) )).
goal_expansion(hash_vector_set(H, V), ( setarg(2, H, V) )).
:- endif.

%
goal_expansion(hash_of_state(X, H),  ( arg(1, X, A), arg(2, A, H))).
goal_expansion(array_hash_of_state(S, A, H), ( arg(1, S, A), arg(2, A, H))).
%
goal_expansion(vector_of_state(S, Array), ( arg(1, S, Z), arg(3, Z, Array))).
goal_expansion(vector_of_state(S, Array, Max),
			   ( arg(1, S, Z),
				 arg(3, Z, Array),
				 arg(1, Z, Max))).
goal_expansion(child_state(S, S0),
	( arg(4, S, Child),
		(	Child == [] -> S0 = S
		;	S0 = Child
		)
	)).

%%	Note on state.
%	Assuming a totally ordered set of atoms, a state is a collection of
%	families of (finite) sets of (finite) these atoms. In a state,
%	each family of sets (FOS) is given a unique nonnegative integer.
%	0 is reserved for the empty family {}, and 1 for {{}} = {0} = 1.
%	If a FOS F is neither 0 nor 1, there must
%	be the maximum atom A among those are an element of a set of F.
%	A is called the maximum atom of F. The maximum atom is defined neither for
%	0 nor 1. Throuhout successive updating process of a state, once the unique
%	id  of a FOS is created, this id is kept for the FOS henceforth.
%	`cofact/3` is the main bidirectional interface predicate on states to get
%	a FOS id (fact), and to store new FOS (cofact) in a state.
%	A nonnegative integer k is used for the FOS whose id is k, and vice versa.

% ?- open_state(S), write(S).
%!	open_state(-S) is det.
%	S is unified with a new state.
open_state(S):- open_state(S, []).
%
open_state(s(#(1, Hash, Vector), Extra, Compare, []), Args):-
	default_compare(Compare),
	( memberchk(hash_size(HsizeExp), Args); 	HsizeExp = 1 ), !,
	( memberchk(vector_size(VsizeExp), Args); 	VsizeExp = 1 ), !,
	( memberchk(extra(Extra), Args); 	Extra = [varnum-0, varzip-[]] ), !,
	Hsize is HsizeExp,
	Vsize is VsizeExp,
	functor(Vector, #, Vsize),
	setarg(1, Vector, 0),
	open_hash(Hsize, Hash).

%!  memo_index(?:X-V, +S) is det.
%   E must be ground, whose main functor is used as a counter name.
%	A unique integer for E is unified with V.
%	This is handy to give a series of integer index to members of a family
%	of ground terms.

% ?- open_state(S), memo(a-1, S), memo(a-X, S).
% ?- open_state(S), memo_index(x-I, S), memo_index(y-J, S), memo_index(x-K, S).
% ?- open_state(S), memo_index(x(1)-I, S), memo_index(x(2)-J, S), memo_index(x(1)-K, S).
% ?- open_state(S), memo_index(a(1)-V, S), memo_index(a(1)-U, S).
% ?- open_state(S), memo_index(a(1)-V, S), get_key(a, C, S).
memo_index(E-V, S):- memo(E-V, S), !,
	(	nonvar(V) -> true
	;	functor(E, CounterName, _),
		memo_index(CounterName, V, S)
	).
%
memo_index(Name, V0, S):- arg(2, S, Extra),
	(	select(Name-V, Extra, Extra0) ->
		V0 is V+1
	;	Extra0 = Extra,
		V0 = 1
	),
	Extra1 = [Name-V0|Extra0],
	setarg(2, S, Extra1).

% Structure Sharing.
setarg_state_extra(X, Y):- arg(3, X, Extra), setarg(3, Y, Extra).
%
setarg_state_core(X, Y):- arg(1, X, Core), setarg(1, Y, Core).
%
clear_child(S) :- set_child([], S).
%
set_child(Child, S):- setarg(4, S, Child).

%
default_compare(compare).

%!	 dump_array(+S) is det.
%	 Print all triples stored in the state S.

% ?- open_state(S), zdd_eval(pow([a,b,c]), _, S), dump_array(S).
% ?- open_state(S), zdd_eval(family([[a],[b],[c]]), X, S), dump_array(S).
dump_array(S):- vector_of_state(S, Array, Max),
	  forall( between(2, Max, J),
			  ( arg(J, Array, T), writeln(J=T))).

%!	close_state(+S) is det.
%	Close the state S. Unused memory is expected to be freed (not sure).

% Note: States are nested recursively in a linear way.
% ?- open_state(S), close_state(S).
close_state(S):- clear_state(S).

%
clear_state(S):-
	nb_setarg(1, S, []),
	nb_setarg(2, S, []),
	nb_setarg(3, S, []),
	nb_setarg(4, S, []).

%!	open_memo(+S) is det.
%	Open a fresh state M, and  M is attatched to S.
%	M is called the child state of S.
%	The child state is mainly used as a hash table.

% ?- zdd((open_memo, memo(a-1), memo(a-V), close_memo)).
% ?- zdd((open_memo, memo(a-1), memo(a-V), close_memo, memo(a-U), var(U))).
open_memo :- @(open_memo).
%
open_memo(S):- nonvar(S),!,
	open_state(S0),
	set_child(S0, S).
open_memo(S):- open_state(S).

%!	close_memo(+S) is det.
%	Close the chiled state of S.

close_memo :- @(close_memo).    % memo and state are alomost the same.

% Close nested memos recursively.
close_memo(S):- must_be(nonvar, S),
	child_state(S, S0),
	clear_child(S),
	(	S0 == [] -> true
	; 	close_state(S0)		% Note that NOT close_memo.
	).

%!	inherit_compare(+S) is det.
%	Set the same compare predicate to the child state of S.

inherit_compare(S):- get_compare(CMP_pred, S),
	child_state(S, Child),
	set_compare(CMP_pred, Child).

%!  inherit_compare(+S, +T) is det.
%   Set the same compare predicate of S to the state T.

inherit_compare(S, T):- get_compare(CMP_pred, S),
	set_compare(CMP_pred, T).

%!	inherit_compare_opp(+S) is det.
%   Set the oppoiste compare predicate of S to the child state of S.

inherit_compare_opp(S):- get_compare(Pred, S),
	child_state(S, Sub),
	set_compare(pred_compare_opp(Pred), Sub).

%!	inherit_compare_opp(+S, +T) is det.
%   Set the oppoiste compare predicate of S to the state T.

inherit_compare_opp(S, T):- get_compare(Pred, S),
	set_compare(pred_compare_opp(Pred), T).

% ?- zdd((memo(a-1), memoq(a-1))).
% ?- zdd((memo(a-1), memochk(a-X))).
% ?- zdd((memochk(a-X))).  %@ false.
% ?- zdd((memo(a-1), memoq(a-X))).	% false
% ?- zdd((memo(a-1), memo(a-X))).
% ?- zdd((X<<pow(:numlist(1, 1000)), card(X, C))). % in zdd module.
% ?- open_state(S), memo(a-1, S), memo(a-V, S).
% ?- open_state(S), memo(a-1, S), memo(a-V, S), set_memo(a-2, S),
%	memo(a-U, S).
% ?- zdd:zdd((set_memo(a-1), memo(a-X))).

memo(P):- @(memo(P)).

%!	memo(+:X-V, +S) is det.
%	The input pair X-V is unified with with a member of a bucket of the hash
%	table of the state S. Otherwise, create a new entry for X-V.

memo(X-V, S):- child_state(S, S0),
	hash_of_state(S0, H),
	hash_update(X, H, V).


% ?- zdd((set_memo(a-1), memo(a-Y), set_memo(a-2), memo(a-R))).
% ?- zdd((set_memo(a-1), memo(a-Y), set_memo(a-V), memo(a-R), V=2)).
set_memo(X-V)	:- @(=(S)), hash_fresh_entry(X, S, V, _).

%!	set_memo(+:X-V, S) is det.
%   Replace old X entry with X-V when X entry exists,
%	otherwise, simply create X-V entry.

set_memo(X-V, S)	:- 	hash_fresh_entry(X, S, V, _).

% ?- zdd((memo(a-1), update_memo(a-X, Y), X=2, memo(a-U))).
update_memo(X-FreshVar, OldVal):- @(=(S)),
	hash_fresh_entry(X, S, FreshVar, OldVal).

%!	update_memo(+:X-FreshVar, +Old, +S) is det.
%	Replace X-Old entry with X-FreshVar entry when
%	the former exists, otherwise simply create X-FreshVar entry with
%	Old left uninstantiated.

update_memo(X-FreshVar, OldVal, S):-
	hash_fresh_entry(X, S, FreshVar, OldVal).

%!	insert_memo(+Key, +X, +S) is det.
%	Insert X in the zdd associated with the Key when the Key entry exists,
%	otherwise the zdd is assumed to be 1.

% ?- zdd((@(insert_memo(abc(5), p(0,0)-p(1,0))), @(insert_memo(abc(5), p(1,1)-p(0,1))), memo(abc(5)-X), psa(X))).
insert_memo(Key, X, S):-
	update_memo(Key-New, Old, S),
	( var(Old) -> Old=1		% empty set
	; true
	),
	zdd_insert(X, Old, New, S).
%
:- meta_predicate pred_memo_update(3, ?).
pred_memo_update(Pred, X):- @(pred_memo_update(Pred, X)).

%!	pred_memo_update(+Pred, +:K-V, S) is det.
%	Replace the pair K-L0 with K-L, where L is obtained by
%	applying Pred to V, L0, by calling Pred(V, L0, L).

:- meta_predicate pred_memo_update(3, ?, ?).
pred_memo_update(Pred, K-V, S):- update_memo(K-L, L0, S),
	call(Pred, V, L0, L).

% ?- zdd((memo_add_new(a-1), memo(a-X))).
% ?- zdd((memo_add_new(a-1), memo_add_new(a-2), memo(a-X))).
% ?- zdd((memo_add_new(a-1), memo_add_new(a-2), memo_add_new(a-1), memo(a-X))).
memo_add_new(X):-  pred_memo_update(add_new, X).

%!	memo_add_new(+:K-V, S) is det.
%	Insert V in the list associated with the key K only when V is new
%   in the list.

memo_add_new(X, S):- pred_memo_update(add_new, X, S).

% ?- zdd((pred_memo_update(add_new, a-1), memo(a-X))).
add_new(V, [], [V]):-!.
add_new(V, L0, L0):- memberchk(V, L0), !.
add_new(V, L0, [V|L0]).

% ?- zdd((add_child(suc(a), 1), memo(suc(a)-X))).
% ?- zdd((add_child(a, 1), add_child(a, 2), memo(a-X))).
% ?- zdd((add_child(a, 1), add_child(a, 2), add_child(a, 1), memo(a-X))).
% ?- numlist(1, 100000, Ns),
%	time(zdd(( maplist(pred([Child]:- add_child(a, Child)), Ns), memo(a-X)))).

add_child(X, Child):-  pred_memo_update(add_new, X-Child).
%
add_child(X, Child, S):- pred_memo_update(add_new, X-Child, S).

%
memochk(X):- @(memochk(X)).

%!	memochk(+:X-V, +S) is det.
%	Unify V with the value of key X only when X entry exists in S,
%	otherwise, fails.

memochk(X-V, S):- child_state(S, S0),
	hash_of_state(S0, H),
	term_bucket_index(X, H, B, _),
	memberchk(X-V, B).
%
memoq(X):- @(memoq(X)).

%!	memoq(+Q:X-V, +S) is det.
%	Check V with the value of key X compared by `==`
%	stored in the hash table of S.
memoq(X-V, Z):- memochk(X-U, Z), !,  U == V.

		/**********************************************
		*         index/get_index/set_index           *
		**********************************************/

% ?- Array = a(#(1,2)), index(2, Array, X), writeln(Array).
% ?- Array = a(#(1,2)), index(5, Array, X), writeln(Array).

%!	index(+I, +A, ?X) is det.
%	Unify X with I-th element of the Array when I is less than equal to
%	the current size of Array, otherwise, extend Array enough
%	for I to satisfy the above condition.

index(I, Array, X):- arg(1, Array, Vec),
		functor(Vec, _, N),
		(	I =< N -> arg(I, Vec, X)
		;   enlarge_vector(Vec, I, Vec0), !,
			nb_setargs(N, Vec, Vec0),
			setarg(1, Array, Vec0),
			arg(I, Vec0, X)
		).

% ?- enlarge_vector(#(1,2,3), 5, R), writeln(R).
% ?- enlarge_vector(#(1,2,3), 3, R), writeln(R).

%!	enlarge_vector(+A, +I, -B) is det.
%	When vector A has I-th (I>= 1) element then B is unified with A,
%	otherwise, extend A so as to be the minimum array such that
%	new size = 2^n x old size >= I for an integer n, and B is unified
%	with this new A.  In the latter case, all entries of of the old A
%	are unified with that of new A in order.

enlarge_vector(A, I, B) :-
		functor(A, G, N),
		(	I =< N -> B = A
		;	iterate_double_number(N, I, N0),
			functor(B, G, N0),
			nb_setargs(N, A, B)
		).

%!	get_index(+I, +Array, -X) is det.
%	Equivalent to index(I, Array, X).
get_index(I, Array, X):- index(I, Array, X), !.

%!	set_index(+I, +Array, -X) is det.
%	Unify X with I_th element of the Array when I is less than equal to
%	the current size of Array, otherwise, extend Array enough
%	for Array to have I-th emlement, then unify X with the I-th element of
%	extended Array, which is an uninstatiated variable.

set_index(I, Array, X):- arg(1, Array, Vec),
		functor(Vec, _, N),
		(	I =< N -> setarg(I, Vec, X)
		;   enlarge_vector(Vec, I, Vec0),
			nb_setargs(N, Vec, Vec0),
			setarg(1, Array, Vec0),
			setarg(I, Vec0, X)
		).

%
iterate_double_number(N, I, N):- I=<N, !.
iterate_double_number(N, I, M):- N0 is 2*N,
	iterate_double_number(N0, I, M).

nb_setargs(0, _, _):-!.
nb_setargs(I, X, Y):- arg(I, X, A),
		nb_setarg(I, X, 0),				% [2021/10/30]
		arg(I, Y, A),
		I0 is I - 1,
		nb_setargs(I0, X, Y).

% NEW style: assuming simple failure for exceeding index.
% ?- unify_args(1, f(A, B, C), a).
unify_args(I, X, A):- arg(I, X, A), !,
	J is I + 1,
	unify_args(J, X, A).
unify_args(_, _, _).

% OLD style of iteration on args.
% ?- old_unify_args(3, f(A, B, C), a).
old_unify_args(0, _, _):-!.
old_unify_args(I, X, A):- arg(I, X, A), !,
	J is I - 1,
	old_unify_args(J, X, A).

%
getid_via_global(G, X, I):- b_getval(G, S), cofact(I, X, S).

		/***********************************************
		*         new_array_elem/get_elem/set_elem     *
		***********************************************/
%!	open_hash(+N, -H) is det.
%   Create a new hash table with N entries for buckets, and unify with H.
open_hash(N, H):- hash_initial(N, H),
	hash_vector(H, V),
	unify_args(1, V, []).

%! hash_memo(+X, +R, ?V) is det.
%	Unify X-V with an element in the hash table R. Rehash may
%	be applied to R by `check_rehash/1`.

hash_memo(X, R, V):- arg(2, R, H),
	hash_update(X, H, V), !.

%!	term_bucket_index(+X, +H, -B, -I) is det.
%	I is unified with the value for X by `term_hash/4`,
%	and B is unified with I-th bucket of H.

% ?- term_bucket_index(b, #(2, 1, #([],[],[])),  B, I).
term_bucket_index(X, H, B, I):-
	hash_vector(H, Vec),
	functor(Vec, _, N),
	term_hash(X, 3, N, I0),
	I is I0 + 1,
	arg(I, Vec, B).

%!	hash_update(+X, +H, ?E) is det.
%	Put a key-value term X-E on the hash table H.

hash_update(X, H, Val):-
	check_rehash(H),
	term_bucket_index(X, H, U, I),
	(	memberchk(X-V, U) -> Val = V
	;	hash_vector(H, Vec),
		setarg(I, Vec, [X-Val | U]),
		hash_key_count(H, C),
		C0 is C + 1,
		hash_key_count_set(H, C0)
	).

%!	hash_fresh_entry(+X, +S, -Var, -OldVal) is det.
%
%	If X-OldVal exists in the hash table then the entry is removed,
%	otherwise, OldVal is left uninstantiated. Then X-var is put as a fresh pair
%	on the hash table.  Rehash may be performed by `check_rehash/1`.
%	This predicate is used to calculate value using old value
%	after preparing the entry for the new value.

hash_fresh_entry(X, S, FreshVar, OldVal):- hash_of_state(S, H),
	check_rehash(H),
	term_bucket_index(X, H, U, I),
	hash_vector(H, Vec),
	(	select(X-OldVal, U, U0)
	->	setarg(I, Vec, [X-FreshVar | U0])
	;	setarg(I, Vec, [X-FreshVar | U])
	).
% ?- hash_initial(H), check_rehash(H), writeln(H).
% ?- hash_initial(H), check_rehash(H), check_rehash(H),
%  check_rehash(H), writeln(H).
% ?- listing(check_rehash).
:- if(bucket_aware_rehash).
check_rehash(H):- H = #(C, B, V),
		functor(V, _, N),
		(	C > B + 1.7 * N -> once(rehash(H))
		;	true
		).

:- else.

check_rehash(H):- H = #(C, V), !,
		functor(V, _, N),
		(	C > 1.7 * N -> once(rehash(H))
		;	true
		).
:- endif.


%!	new_array_elem(+X, +Z, -I) is det.
%	Unify I with an integer index for entry X

% ?- numlist(1,3, Ns), zdd((X<<pow(Ns), card(X, C))).
new_array_elem(X, Z, I):-
		arg(1, Z, J),
		arg(3, Z, Vec),
		I is J + 1,
		functor(Vec, _, N),
		(	I > N
		->	once(double_array(Z)),
			arg(3, Z, Vec0)
		;	Vec0 = Vec
		),
		setarg(1, Z, I),
		arg(I, Vec0, X).

%!	double_array(+A) is det.
%	Update A by extending array component of A by double.
%	(Destructive update).

double_array(A) :- arg(3, A, Z),
		functor(Z, G, N),
		(	N == 0 -> N0 = 2
		; 	N0 is 2*N
		),
		functor(Z0, G, N0),
		nb_setargs(N, Z, Z0),
		setarg(3, A, Z0).

%!	get_elem(+X, +Z, -Y) is det.
%	Unify Y with X-th element of ZDD stored in Z.

get_elem(X, Z, Y):- arg(3, Z, Vec),
	   arg(X, Vec, Y).

%!	set_elem(X, Z, Y) is det.
%	Set X-th element of the vector component of Z to Y.
set_elem(X, Z, Y):- arg(3, Z, Vec), setarg(X, Vec, Y).

% ?- hash_initial(H), rehash(H), writeln(H), rehash(H), rehash(H), rehash(H), writeln(H).

%!	rehash(H, H0) is det.
%	Extend the hash table H to double-sized H0, and rehash H to H0.
rehash(H):-
	hash_vector(H, Vec),
	functor(Vec, F, N),
	N0 is N + N,
	functor(Vec0, F, N0),
	unify_args(1, Vec0, []),
	!,
	(	functor(H, _, 2)->
		migrate_hash(N, Vec, Vec0),
		hash_vector_set(H, Vec0)
	;	migrate_hash(N, Vec, Vec0, 0, C),
		hash_vector_set(H, Vec0),
		hash_bucket_count_set(H, C)
	).
%!	migrate_hash(I, V, V0) is det.
%	Rehash all elements of I-th bucket of V into V0.
migrate_hash(0, _, _):-!.
migrate_hash(I, V, V0):-
	arg(I, V, B),
	once(migrate_bucket(B, V0)),
	I0 is I - 1,
	migrate_hash(I0, V, V0).
%
migrate_bucket([], _):-!.
migrate_bucket([Q|U], H):-
	(	Q = (X-_) -> E = Q
	;   X = Q,
		E = Q
	),
	functor(H, _, S),
	term_hash(X, 3, S, K),
	K0 is K + 1,
	arg(K0, H, D),
	setarg(K0, H, [E|D]),
	migrate_bucket(U, H).

% For buckets number aware rehashing.
migrate_hash(0, _, _, C, C):-!.
migrate_hash(I, V, V0, C, C0):-
	arg(I, V, B),
	once(migrate_bucket(B, V0, C, C1)),
	I0 is I - 1,
	migrate_hash(I0, V, V0, C1, C0).

%
migrate_bucket([], _, C, C):-!.
migrate_bucket([Q|U], H, C, C0):-
	(	Q = (X-_) -> E = Q
	;   X = Q,
		E = Q
	),
	functor(H, _, S),
	term_hash(X, 3, S, K),
	K0 is K + 1,
	arg(K0, H, D),
	setarg(K0, H, [E|D]),
	(	D=[] -> C1=C
	;	C1 is C+1
	),
	migrate_bucket(U, H, C1, C0).


		/****************
		*     cofact    *
		****************/

% ?- zdd((cofact(X, t(a, 1, 0)))).
% ?- zdd((X<<set([a]), psa(X))).
cofact(X, Y):- @(cofact(X, Y)).

%!	cofact(?X, ?C:t(A, L, R), +State) is det.
%	Bidirectional.
%	X is unified with the index of a triple C, or
%	C is unified with the triple t/3 stored at index X of the array.
%
%	It is explained in terms of famiy of sets as follows.
%	If X is given then
%	Y is a triple t(A, L, R) such that
%	A is the minimum atom in X w.r.t specified compare predicate,
%	L = { U in X | not ( A in U ) },
%	R = { V \ {A}  | V in X, A in V }.
%	If Y is given then
%	X = union of L and { unionf of U and {A} | U in R }.
%
%   Non standard use of cofact/3 is possible keeping the structure sharing,
%	but withoug zero_suppress rule. IMO the rule is only meaningful
%	under family of sets semantics for the empty family {} of sets.
%
% ?- #S, cofact(X, [a,b,d], S), cofact(X, L, S).
% ?- #S, cofact(X, [a], S), cofact(X, [A|B], S).
% ?- #S, cofact(X, f(a,b,d), S), cofact(X, L, S).

cofact(X, C, S):- nonvar(X), !,   % X>1 assumed.
	vector_of_state(S, Vec),
    arg(X, Vec, C).
cofact(X, t(_, X, 0), _):-!.	 % the famous Minato's zero-suppress rule.
cofact(X, C, S):- array_hash_of_state(S, A, H),
	hash_update(C, H, X),		% check C-X entry in H (hash)
	(	nonvar(X) -> true		% exists.
	;	new_array_elem(C, A, X)	% new.
	).

		/*****************************************
		*     copy, slim, ord_copy, pred_copy    *
		*****************************************/

%!	zdd_ord_copy(+X, +S, -Y, +T) is det.
%   Copy zdds in X on S to Y on T. X is a nested lists of integers,
%   which is a unique index of a zdd.

zdd_ord_copy(X, S, Y, T):- integer(X), !,
	zdd_ord_copy_basic(X, S, Y, T).
zdd_ord_copy([], _, [], _):-!.
zdd_ord_copy([X|Xs], S, [Y|Ys], T):-!,
	zdd_ord_copy(X, S, Y, T),
	zdd_ord_copy(Xs, S, Ys, T).

%!	zdd_ord_copy_basic(+X, +S, -Y, +T) is det.
%   Assuming both states S, T shares a same ordering,
%	copy zdds in X on S to Y on T.

zdd_ord_copy_basic(X, _, X, _):- X< 2,!.
zdd_ord_copy_basic(X, S, Y, S0):- memo(ord_copy(X)-Y, S0),
	(	nonvar(Y) -> true
	;	cofact(X, t(A, L, R), S),
		zdd_ord_copy_basic(L, S, L0, S0),
		zdd_ord_copy_basic(R, S, R0, S0),
		cofact(Y, t(A, L0, R0), S0)		% Ordering is the same.
	).

%!	zdd_copy(+X, +S, -Y, +T) is det.
%   Copy zdds in X on S to Y on T.  Sharing orders of S and T is not
%	assumed.

zdd_copy(X, S, Y, T):- integer(X), !, zdd_copy_basic(X, S, Y, T).
zdd_copy([], _, [], _):-!.
zdd_copy([X|Xs], S, [Y|Ys], T):-
	zdd_copy(X, S, Y, T),
	zdd_copy(Xs, S, Ys, T).
%
zdd_copy_basic(X, _, X, _):- X< 2,!.
zdd_copy_basic(X, S, Y, S0):- memo(copy(X)-Y, S0),
	(	nonvar(Y) -> true
	;	cofact(X, t(A, L, R), S),
		zdd_copy_basic(L, S, L0, S0),
		zdd_copy_basic(R, S, R1, S0),
		zdd_insert(A, R1, R0, S0),	% Ordering may be different.
		zdd_join(L0, R0, Y, S0)
	).

%!	zdd_pred_copy(+Pred, +X, +S, -Y, +T) is det.
%	Predicate Pred/2 is to map atoms of S to that of T.
%	Mapping atoms by Pred, copy zdds in X of S to those in Y of T.

% ?- open_state(S), open_state(S0), zdd_eval(pow([a,b]), X, S),
%	zdd_pred_copy(pred([A, A]), X, S, Y, S0), card(X, C, S), card(Y, D, S0),
%	psa(X, S), psa(Y, S0).
% ?- open_state(S), zdd_eval(pow([a, b, c]), X, S),
% zdd_pred_copy(pred([A,A-A]), X, S, Y, S), psa(X, S), psa(Y, S),
% card(X, C, S), card(Y, D, S).

:- meta_predicate zdd_pred_copy(2, ?, ?, ?, ?).

zdd_pred_copy(Pred, X, S, Y, S0):- integer(X), !,
	zdd_pred_copy_basic(Pred, X, S, Y, S0).
zdd_pred_copy(_, [], _, [], _):-!.
zdd_pred_copy(Pred, [X|Xs], S, [Y|Ys], S0):- zdd_pred_copy(Pred, X, S, Y, S0),
	zdd_pred_copy(Pred, Xs, S, Ys, S0).
%
zdd_pred_copy_basic(_, X, _, X, _):- X< 2,!.
zdd_pred_copy_basic(Pred, X, S, Y, S0):- memo(pred_copy(X)-Y, S0),
	(	nonvar(Y) -> true
	;	cofact(X, t(A, L, R), S),
		zdd_pred_copy_basic(Pred, L, S, L0, S0),
		zdd_pred_copy_basic(Pred, R, S, R1, S0),
		call(Pred, A, B),
		zdd_insert(B, R1, R0, S0),
		zdd_join(L0, R0, Y, S0)
	).

%%  Be careful because of destructive structure sharing !.
% ?- open_state(S), open_memo(S), zdd_eval(pow([1,2]), X, S),
%	zdd_eval(pow([a,b]), Y, S),
%	zdd_slim(Y, U, S),  writeln(S).
zdd_slim(X, Y):- @(zdd_slim(X, Y)).

%!	zdd_slim(+X, -Y, +S) is det.
%	Remove all redundant zdds that are irrelevant to those specified
%	in X. In fact, perform zdd_ord_copy(X, S, Y, T) for fresh state T,
%	then the content of the orginal S is destructively replaced
%	with that of T by `setarg/3`. T is closed at exit.

zdd_slim(X, Y, S):- open_state(S0),
	open_memo(S0),
	inherit_compare(S, S0),
	setarg_state_extra(S, S0),
	clear_child(S),		% clear substate (old memo) of the original state.
	zdd_slim(X, S, Y, S0),
	setarg_state_core(S0, S),
	clear_state(S0).	% Only detaching args for possible sharing.
%
zdd_slim(X, S, Y, T):- integer(X), !, zdd_ord_copy_basic(X, S, Y, T).
zdd_slim([], _, [], _):-!.
zdd_slim([X|Xs], S, [Y|Ys], T):- zdd_slim(X, S, Y, T),
	zdd_slim(Xs, S, Ys, T).

%@ true.
